###
####
####
print(`Package "Solving third-order holonomic differential equations", Maple 16`):
print(`Copyright 2017, Mouafo Wouodjie Merlin, University of Kassel`):
###

with(DETools):
bind(FormalPowerSeries):
 _Envdiffopdomain:=[Dx,x]:
read "hsum17.mpl":
read("Hom.mpl");
read("MinOp.mpl");



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################

# We compute here the operator L' comming from L by the change of variables f. Note that L has to be an third order
# linear irreducible differential operator.
ChangeOfVariables:=proc(L,f)
                   local Dx,x,l,f1,f2,f3,A,B,C,D,E,F,G;
                   Dx,x:=op(_Envdiffopdomain);
                   l:=lcoeff(L,Dx);
                   A:=normal(coeff(L,Dx,2)/l);
                   B:=normal(coeff(L,Dx,1)/l);
                   C:=normal(coeff(L,Dx,0)/l);
                   f1:=normal(diff(f,x));
                   f2:=normal(diff(f1,x));
                   f3:=normal(diff(f2,x));
                   D:=normal(subs(x=f,A)*f1-3*f2/f1);
                   E:=normal(subs(x=f,-A)*f2+3*(f2/f1)^2+subs(x=f,B)*f1^2-f3/f1);
                   F:=normal(subs(x=f,C)*f1^3);
                   G:= ReplirrRoot(evala(collect(numer(Dx^3+D*Dx^2+E*Dx+F),Dx,factor)),{});
                   return collect(G,Dx,factor);
end:

ChangeOfVariablesGeneral:=proc(L,f)
                   local Dx,x,l,f1,f2,f3,A,B,C,D,E,F,G;
                   Dx,x:=op(_Envdiffopdomain);
                   l:=lcoeff(L,Dx);
                   A:=normal(coeff(L,Dx,2)/l);
                   B:=normal(coeff(L,Dx,1)/l);
                   C:=normal(coeff(L,Dx,0)/l);
                   f1:=normal(diff(f,x));
                   f2:=normal(diff(f1,x));
                   f3:=normal(diff(f2,x));
                   D:=normal(subs(x=f,A)*f1-3*f2/f1);
                   E:=normal(subs(x=f,-A)*f2+3*(f2/f1)^2+subs(x=f,B)*f1^2-f3/f1);
                   F:=normal(subs(x=f,C)*f1^3);
                   G:= collect(Dx^3+D*Dx^2+E*Dx+F,Dx);
                   return collect(G,Dx,factor);
end:

############################################################################################################
############################################################################################################

# We compute here the operator L' comming from L by the exp_product transformation with parameter r. Note that L has to
# be an third order linear irreducible differential operator.
ExpProduct:=proc(L,r)
            local Dx,x,l,r1,r2,A,B,C,K,E,F,G;
            Dx,x:=op(_Envdiffopdomain);
            l:=lcoeff(L,Dx);
            A:=normal(coeff(L,Dx,2)/l);
            B:=normal(coeff(L,Dx,1)/l);
            C:=normal(coeff(L,Dx,0)/l);
            r1:=normal(diff(r,x));
            r2:=normal(diff(r1,x));
            K:=normal(A-3*r);
            E:=normal(B-2*r*A+3*r^2-3*r1);
            F:=normal((r^2-r1)*A-r*B-r2+3*r*r1-r^3+C);
            G:=ReplirrRoot(evala(collect(numer(Dx^3+K*Dx^2+E*Dx+F),Dx,factor)),{});
            return collect(G,Dx,factor);
end:

ExpProductGeneral:=proc(L,r)
            local Dx,x,l,r1,r2,A,B,C,K,E,F,G;
            Dx,x:=op(_Envdiffopdomain);
            l:=lcoeff(L,Dx);
            A:=normal(coeff(L,Dx,2)/l);
            B:=normal(coeff(L,Dx,1)/l);
            C:=normal(coeff(L,Dx,0)/l);
            r1:=normal(diff(r,x));
            r2:=normal(diff(r1,x));
            K:=normal(A-3*r);
            E:=normal(B-2*r*A+3*r^2-3*r1);
            F:=normal((r^2-r1)*A-r*B-r2+3*r*r1-r^3+C);
            G:=collect(Dx^3+K*Dx^2+E*Dx+F,Dx);
            return collect(G,Dx,factor);
end:

############################################################################################################
############################################################################################################

# We compute here the operator L' comming from L by the gauge transformation with parameter r0,r1 and r2. Note that L
# has to be an third order linear irreducible differential operator.
GaugeTransf:=proc(L,r0,r1,r2)
            local Dx,x,y,Y,a0,a1,a2,a3,DE,z,S,D1,D2,D3,b0,b1,b2,SUM,SUM2,EQS,sol,solution,G;
            Dx,x:=op(_Envdiffopdomain);
            a3:=lcoeff(L,Dx);
            a2:=normal(coeff(L,Dx,2));
            a1:=normal(coeff(L,Dx,1));
            a0:=normal(coeff(L,Dx,0));
            DE:=a3*diff(y(x),x$3)+a2*diff(y(x),x$2)+a1*diff(y(x),x)+a0*y(x);
            S:=diff(y(x),x$3)=solve(DE,diff(y(x),x$3));
            z:=x->r0*y(x)+r1*diff(y(x),x)+r2*diff(y(x),x$2);
            D1:=collect(subs(S,diff(z(x),x)),{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            D2:=collect(subs(S,diff(D1,x)),{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            D3:=collect(subs(S,diff(D2,x)),{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            SUM:=collect(b0*z(x)+b1*D1+b2*D2+D3,{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            SUM2:=collect(subs({diff(y(x),x$2)=Y^2,diff(y(x),x)=Y,y(x)=1},SUM),Y);
            EQS:={coeffs(collect(SUM2,Y),Y)};
            solution:=solve(EQS,{b0,b1,b2});
            sol:=eval([b0,b1,b2],solution);
            G:= ReplirrRoot(evala(collect(numer(Dx^3+sol[3]*Dx^2+sol[2]*Dx+sol[1]),Dx,factor)),{});
            return collect(G,Dx,factor);
end:

GaugeTransfGeneral:=proc(L,r0,r1,r2)
            local Dx,x,y,Y,a0,a1,a2,a3,DE,z,S,D1,D2,D3,b0,b1,b2,SUM,SUM2,EQS,sol,solution,G;
            Dx,x:=op(_Envdiffopdomain);
            a3:=lcoeff(L,Dx);
            a2:=normal(coeff(L,Dx,2));
            a1:=normal(coeff(L,Dx,1));
            a0:=normal(coeff(L,Dx,0));
            DE:=a3*diff(y(x),x$3)+a2*diff(y(x),x$2)+a1*diff(y(x),x)+a0*y(x);
            S:=diff(y(x),x$3)=solve(DE,diff(y(x),x$3));
            z:=x->r0*y(x)+r1*diff(y(x),x)+r2*diff(y(x),x$2);
            D1:=collect(subs(S,diff(z(x),x)),{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            D2:=collect(subs(S,diff(D1,x)),{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            D3:=collect(subs(S,diff(D2,x)),{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            SUM:=collect(b0*z(x)+b1*D1+b2*D2+D3,{diff(y(x),x$2),diff(y(x),x),y(x)},factor);
            SUM2:=collect(subs({diff(y(x),x$2)=Y^2,diff(y(x),x)=Y,y(x)=1},SUM),Y);
            EQS:={coeffs(collect(SUM2,Y),Y)};
            solution:=solve(EQS,{b0,b1,b2});
            sol:=eval([b0,b1,b2],solution);
            G:= collect(Dx^3+sol[3]*Dx^2+sol[2]*Dx+sol[1],Dx);
            return collect(G,Dx,factor);
end:

############################################################################################################
############################################################################################################

# We check if two third order linear irreducible differential operators L1 and L2 are related just with exp_product
# transformation, no other transformations are involved. If that is true then we return the parameter r of this
# exp_product transformation.
ExpProdEquiv:=proc(L1,L2)
            local Dx,x,r1,r2,l1,l2,r,a,b;
            Dx,x:=op(_Envdiffopdomain);
      # Specify the coefficients of L1 and L2.
            l1:=lcoeff(L1,Dx);
            l2:=lcoeff(L2,Dx);
            a[2]:=normal(coeff(L1,Dx,2)/l1);
            a[1]:=normal(coeff(L1,Dx,1)/l1);
            a[0]:=normal(coeff(L1,Dx,0)/l1);
            b[2]:=normal(coeff(L2,Dx,2)/l2);
            b[1]:=normal(coeff(L2,Dx,1)/l2);
            b[0]:=normal(coeff(L2,Dx,0)/l2);
      # Give the value that r will take if L1 and L2 are ExpProduct equivalent
            r:=normal((a[2]-b[2])/3);
      # check the ExpProduct equivalence of L1 and L2
            r1:=normal(diff(r,x));
            r2:=normal(diff(r1,x));
            if  expand(b[1]) <> expand(normal(a[1]-2*r*a[2]+3*r^2-3*r1)) or expand(b[0]) <> expand(normal((r^2-r1)*a[2]-r*a[1]-r2+3*r*r1-r^3+a[0]))                    then return 0; end if:
            return [factor(r)];
end:

############################################################################################################
############################################################################################################

# We want to see if the element i or -i appears in a list A modulo an integer.
MatchInt := proc(A, i)
            local j;
            for j in A do
                 if type(i-j,integer) or type(i+j,integer) then
                      return true
                 fi;
            od;
            false
end:

############################################################################################################
############################################################################################################

# We want to see if the element i or -i appears in a list A modulo an integer, and if it is true then we take the
# minimum between i and j = a correspondant of i in A modulo an integer.
MatchInt1 := proc(A, i)
            local j,E;
            for j in A do
               if type(i-j,integer) then
                  if i-j > 0 then
                     return j
                  else
                     return i
                  fi;
               else
                  if type(i+j,integer) then
                     E:= indets({i,j},{RootOf,name,function});
                     if E={} then
                        return evalr(min(i,j))
                     else
                        return j
                     fi;
                  fi;
               fi;
            od;
            false
end:

############################################################################################################
############################################################################################################

# We want to see if the element i appears in a list A modulo an integer.
MatchInt2 := proc(A, i)
            local j;
            for j in A do
                 if type(i-j,integer) then
                      return true
                 fi;
            od;
            false
end:

############################################################################################################
############################################################################################################

# We want to see if the element i or -i appears in a list A modulo an integer. It is the same as the code
#                                       MatchInt := proc(A, i)
MatchInt3 := proc(A, i)
            local j;
            for j in A do
                 if type(i-j,integer) or type(-i-j,integer)  then
                      return true
                 fi;
            od;
            false
end:

############################################################################################################
############################################################################################################

# We have a list A which elements are also lists of two elements. We have also a list j of two elements. We take an
# element i of A. We want to see if the first element of i appears in j modulo an integer and the second element
# of i is the other element of j modulo an integer.
MatchInt4 := proc(A, i)
            local j;
            for j in A do
                 if (type(i[1]-j[1],integer) and type(i[2]-j[2],integer)) or (type(i[1]-j[2],integer)
                     and type(i[2]-j[1],integer)) then
                      return true
                 fi;
            od;
            false
end:

############################################################################################################
############################################################################################################

# We have two list A and B. We want to take all the elements in A which appear in B modulo an integer. p is
# its order of appearence in B. The output is a list and the elements of this list are also the lists of two elements.
MatchInt5 :=proc(A, B)
            local i,j,C,p;
            C:=table();
            for j in A do
               p:=0;
               for i in B do
                  p:=p+1;
                  if type(i-j,integer) then
                     C[j,i]:=[j,p]
                  fi;
               od;
            od;
            C:= convert(C,'list');
            return C
end:

############################################################################################################
############################################################################################################

# We have two lists i and j of two elements. We want to see if the first element of i appears in j modulo
# an integer and the second element of i is the other element of j modulo an integer.
MatchInt7 := proc(i, j)
            if (type(i[1]-j[1],integer) and type(i[2]-j[2],integer)) or (type(i[1]-j[2],integer) and
               type(i[2]-j[1],integer)) then
                   return true
            fi;
            false
end:

############################################################################################################
############################################################################################################

MatchInt8 :=proc(A, B)
            local i,j,C;
            C:=table();
            for j in A do
                 for i in B do
                     if MatchInt7(i[2], j[2]) then
                         C[j,i]:= [j[1],i[1],i[2]];
                     fi;
                 od;
            od;
            C:= convert(C,'list');
            return C
end:

############################################################################################################
############################################################################################################

Matchfunc :=proc(A, i)
            local j,a,k,C;
            C:={};
            k:=0;
            for j in A do
               if j[2]=i[2] then
                  k:=k+1;
                  a:=[op({op(j[1]),op(i[1])})];
                  C:=subs(j[1]=a,A);
                  break
               fi;
            od;
            if k=1 then
               return C
            else
               return {op(A),i}
            fi;
end:

############################################################################################################
############################################################################################################

# We want to do the division of a modulo b. Our code give us the remainder of this division.
moduloint:=proc(a,b)
           local Dx,x,c,d;
           Dx,x:=op(_Envdiffopdomain);
           c:=1;
           if a < 0 then
               c:=-1;
           fi;
           d:=a;
           while (abs(d) > abs(b)) or (abs(d)=abs(b)) do
               d:=abs(d)-abs(b);
           od;
           return d*c;
end:

############################################################################################################
############################################################################################################

# Let G be a list of elements in the form [a,b]. If [a,b] appear two time in G, we remove one of them. If [e,c]
# and [c,e] appear in G then we remove one of them.
Matchlist:= proc(G,a)
            local n,j,H,i;
            H:= [];
            j:=0;
            n:= nops(G);
            for i in G do
               if (i[1]=a[2] and i[2]=a[1]) or (i[1]=a[1] and i[2]=a[2])  then
                  break
               fi;
               j:=j+1;
            od;
            if j < n then
               return true
            fi;
            return false
end:

############################################################################################################
############################################################################################################

# We take two list [a1] and [b1]. We try to see if the list [b1] is in the form:
#       [a1] or [a1+1] or [a1-1]
Matchlistmod1:= proc(a,b)
               if (a[1]=b[1] or a[1]=b[1]+1 or a[1]=b[1]-1) then
                  return true
               else
                  return false
               fi;
end:

############################################################################################################
############################################################################################################

# We take two list [a1,a2] and [b1,b2]. We try to see if the list [b1,b2] is in the form:
#       [a1,a2] or [a1,a2+1] or [a1,a2-1] or [a1+1,a2] or [a1+1,a2+1] or [a1+1,a2-1] or
#       [a1-1,a2] or [a1-1,a2+1] or [a1-1,a2-1] or [a2,a1] or [a2,a1+1] or [a2,a1-1] or
#       [a2+1,a1] or [a2+1,a1+1] or [a2+1,a1-1] or [a2-1,a1] or [a2-1,a1+1] or [a2-1,a1-1].
Matchlistmod2:= proc(a,b)
               if (a[1]=b[1] or a[1]=b[1]+1 or a[1]=b[1]-1) and
                  (a[2]=b[2] or a[2]=b[2]+1 or a[2]=b[2]-1) then
                  return true
               elif (a[1]=b[2] or a[1]=b[2]+1 or a[1]=b[2]-1) and
                    (a[2]=b[1] or a[2]=b[1]+1 or a[2]=b[1]-1) then
                  return true
               else
                  return false
               fi;
end:

############################################################################################################
############################################################################################################

# For a rational number a, we compute its representation modulo an integer in [-1,1].
# We have two representations: the first one is in [-1,0] and the other in [0,1].
divimod1:=proc(a)
          local a1,a2,b,n;
          n:= 0;
          if a < 0 then
             a1:= -abs(numer(a));
             a2:= abs(denom(a));
             b:= a+n;
             while abs(a1/a2) > 1 do
                n:= n+1;
                b:= a+n;
                a1:= -abs(numer(b));
                a2:= abs(denom(b));
             od;
             return [b,b+1]
          else
             a1:= numer(a);
             a2:= denom(a);
             b:= a-n;
             while abs(a1/a2) > 1 do
                n:= n+1;
                b:= a-n;
                a1:= numer(b);
                a2:= denom(b);
             od;
             return [b,b-1]
          fi;
end:

############################################################################################################
############################################################################################################

# For a rational number a, we compute its representation modulo an integer in [0,1[.
divimod2:=proc(a)
          local a1,a2,b,n;
          n:= 0;
          if type(a,integer) then
             return [0]
          elif a < 0 then
             a1:= -abs(numer(a));
             a2:= abs(denom(a));
             b:= a+n;
             while abs(a1/a2) > 1 do
                n:= n+1;
                b:= a+n;
                a1:= -abs(numer(b));
                a2:= abs(denom(b));
             od;
             return [b+1]
          else
             a1:= numer(a);
             a2:= denom(a);
             b:= a-n;
             while abs(a1/a2) > 1 do
                n:= n+1;
                b:= a-n;
                a1:= numer(b);
                a2:= denom(b);
             od;
             return [b]
          fi;
end:

############################################################################################################
############################################################################################################

# For a non-rational number b1, we compute its representation with the rational part modulo an integer in ]0,1].
# We assume the non-rational part of b1 in the set ext1.
Ratiopartmodint:=proc(b1,ext1)
              local G1,G11,G111,ira,ira1;
              G1:= ext1;
              ira1:={};
              G11:= b1;
              while G1<>{} do
                 ira:= G1[1];
                 ira1:= ira1 union {ira};
                 G11:= collect(convert(G11,polynom),ira,evala);
                 G11:= coeff(G11,ira,0);
                 G1:= indets(G11,{RootOf,name}) minus ira1;
              od;
              ####### We make the rational part of b1 belong to ]0,1] (using divimod2)
              G111:= divimod2(G11);
              return [b1-G11+op(G111),op(G111),b1-G11,G11];
end:

############################################################################################################
############################################################################################################

# We choose all the good RooOfs (no the RootOfs inside RootOfs) in b which are not in the set extp.
# We return a list of three elements:
#   - the RootOfs (no the RootOfs inside RootOfs) in b which are not in extp;
#   - all the RootOfs in b which are not in extp;
#   - all the RootOfs in b which are in extp.
ChooseRootOflistminus:=proc(b,extp)
           local e,e1,e2,e3,k,s,v1,v2,A1,A2,k1,k2,extpp,H,ss;
           e1:= map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1]/ConvirrRoot(s)[3],s),indets(b,{RootOf}));
           extpp:= map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1]/ConvirrRoot(s)[3],s),extp);
           e:= e1 minus extpp;
           if e={} then
              return [{},{},indets(b,{RootOf})]
           elif nops(e)=1 then
              e2:=map(s-> `if`(member(ConvirrRoot(s)[1]/ConvirrRoot(s)[3],e),s,NULL),indets(b,{RootOf}));
              e3:= indets(b,{RootOf}) minus e2;
              return [e2,e2,e3]
           else
              H:= table();
              e2:=map(s-> `if`(member(ConvirrRoot(s)[1]/ConvirrRoot(s)[3],e),s,NULL),indets(b,{RootOf}));
              e3:= indets(b,{RootOf}) minus e2;
              for k in e2 do
                 v1:=e2 minus {k};
                 v2:=subs({seq(s=0,s in v1)},b);
                 if member(k,indets(v2,{RootOf})) then
                    H[k]:= k
                 else
                    A1:=map(s->`if`(member(s,indets(k,{RootOf})),s,NULL),v1);
                    if A1<>{} then
                       A2:=v1 minus A1;
                       k1:=subs({seq(A1[i]=m[i],i=1..nops(A1))},k);
                       k2:=subs({seq(A1[i]=m[i],i=1..nops(A1)),seq(A2[i]=0,i=1..nops(A2))},b);
                       if member(k1,indets(k2,{RootOf})) then
                          H[k]:= k
                       fi;
                    fi;
                 fi;
              od;
              return [convert(H,'set'),e2,e3]
           fi;
end:

############################################################################################################
############################################################################################################

# We choose all the good RooOfs (no the RootOfs inside RootOfs) in b which  belongs to the set extp.
# We return a list of three elements:
#   - the RootOfs (no the RootOfs inside RootOfs) in b which are in extp;
#   - all the RootOfs in b which are not in extp;
#   - all the RootOfs in b which are in extp.
ChooseRootOflistinter:=proc(b,extp)
           local e,e1,e2,e3,k,s,v1,v2,A1,A2,k1,k2,extpp,H,ss;
           e1:= map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1]/ConvirrRoot(s)[3],s),indets(b,{RootOf}));
           extpp:= map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1]/ConvirrRoot(s)[3],s),extp);
           e:= e1 intersect extpp;
           if e={} then
              return [{},indets(b,{RootOf}),{}]
           elif nops(e)=1 then
              e2:=map(s-> `if`(member(ConvirrRoot(s)[1]/ConvirrRoot(s)[3],e),s,NULL),indets(b,{RootOf}));
              e3:= indets(b,{RootOf}) minus e2;
              return [e2,e3,e2]
           else
              H:= table();
              e2:=map(s-> `if`(member(ConvirrRoot(s)[1]/ConvirrRoot(s)[3],e),s,NULL),indets(b,{RootOf}));
              e3:= indets(b,{RootOf}) minus e2;
              for k in e2 do
                 v1:=e2 minus {k};
                 v2:=subs({seq(s=0,s in v1)},b);
                 if member(k,indets(v2,{RootOf})) then
                    H[k]:= k
                 else
                    A1:=map(s->`if`(member(s,indets(k,{RootOf})),s,NULL),v1);
                    if A1<>{} then
                       A2:=v1 minus A1;
                       k1:=subs({seq(A1[i]=m[i],i=1..nops(A1))},k);
                       k2:=subs({seq(A1[i]=m[i],i=1..nops(A1)),seq(A2[i]=0,i=1..nops(A2))},b);
                       if member(k1,indets(k2,{RootOf})) then
                          H[k]:= k
                       fi;
                    fi;
                 fi;
              od;
              return [convert(H,'set'),e3,e2]
           fi;
end:

############################################################################################################
############################################################################################################

# Inverse the side of an equation E
inverequal:=proc(E)
     local a1, a2;
     a1:=lhs(E);
     a1:=RootOf(op([1,1],indets(a1,`RootOf`)));
     a2:=rhs(E);
     return a2=a1
end:

############################################################################################################
############################################################################################################

# Convert an expression into RootOf notation.
ConvRoot:=proc(E)
          local s1,s2,s3;
          s1 := indets(E, {radical, nonreal});
          s2:=radfield(s1);
          s3:=map(s->inverequal(s),s2[2]);
          return subs({op(s3)},E);
end:

############################################################################################################
############################################################################################################

# Replace  C = RootOf(x^n + a^n*b) or RootOf(x^n + a^n*b,index=i) by a*RootOf(x^n + b,index=i), i=1..n.
ReplRoot3:=proc(C)
          local Dx,x,c1,c2,c3,c4,i;
          Dx,x:=op(_Envdiffopdomain);
          c1:=indets(C,RootOf);
          if c1={} then
             return [C]
          else
             c2:= ConvirrRoot(C);
             c3:=op([1,1],{c2[1]/c2[3]});
             return [[seq(c2[3]*RootOf(c3,indets(c3)[1],index=i),i=1..c2[2])],c2[2],c2[3]]
          fi;
end:

############################################################################################################
############################################################################################################

Arange:=proc(A3,A2,k)
  local H2,s;
  H2:= {seq(`if`(member(rhs(s),indets(k,RootOf)) and rhs(s)<>k,rhs(s)=lhs(s),NULL),s in A3)};
  return subs(H2,k);
end:

############################################################################################################
############################################################################################################

# Replace in an expression the RootOf(x^n + a^n*b) or RootOf(x^n + a^n*b,index=i) which are not in the set
# extp by a*RootOf(x^n + b,index=i), i=1..n
ReplRoot:=proc(C,extp)
          local Dx,x,A1,A2,A3,A4,A5,A6,A7,A8,A9,A10,A11,A12,A13,A14,A15,A16,A17,H1,d,e1,e2,j,m,n,B1,C1,i;
          Dx,x:=op(_Envdiffopdomain);
          A1:= ChooseRootOflistminus(C,extp);
          if A1[1]={} then
             return [C]
          else
             C1:=subs({seq(A1[3][i]=m[i],i=1..nops(A1[3]))},C);
             A2:= subs({seq(A1[3][i]=m[i],i=1..nops(A1[3]))},A1[1]);
             A3:={seq(n[i]=A2[i],i=1..nops(A2))};
             H1:=map(s-> Arange(A3,A2,s),A2);
             A4:= map(s-> ReplRoot3(s),[op(H1)]);
             A5:=[seq(map(s->d[i]=s,A4[i][1]),i=1..nops(A4))];
             A6:={{seq(e1[i],i=1..nops(A5))}};
             j:=1;
             while j <= nops(A5) do
                # A7:=seq(A6,e1[j] in A5[j]);
                A7:={seq(subs(e1[j]=i,A6),i in A5[j])};
                A7:=map(s->op(s),A7);
                A6:= A7;
                j:=j+1;
             od;
             A8:= convert([op(A7)],'list');
             A9:=[seq({seq(A2[i]=rhs(B1[i]),B1 in A8)},i=1..nops(A2))];
             A10:={{seq(e2[i],i=1..nops(A9))}};
             j:=1;
             while j <= nops(A9) do
                A11:={seq(subs(e2[j]=i,A10),i in A9[j])};
                A11:=map(s->op(s),A11);
                A10:= A11;
                j:=j+1;
             od;
             A12:= convert([op(A11)],'list');
             A13:= seq(subs(A12[i],C1),i=1..nops(A12));
             A14:= convert([A13],'list');
             A15:= seq(subs(A3,A14[i]),i=1..nops(A14));
             A16:= convert([A15],'list');
             A17:= seq(subs({seq(m[j]=A1[3][j],j=1..nops(A1[3]))},A16[i]),i=1..nops(A16));
             return convert([A17],'list');
          fi;
end:

############################################################################################################
############################################################################################################

####### We replace C = RootOf(x^n + a^n*b) or RootOf(x^n + a^n*b,index=i) by a*RootOf(x^n + b).
ReplRoot1:=proc(C)
          local c1,c2,c3,x,Dx;
          Dx,x:=op(_Envdiffopdomain);
          c1:=indets(C,RootOf);
          if c1={} then
             return C
          else
             c2:=ConvirrRoot(C);
             c3:=op([1,1],{c2[1]/c2[3]});
             return c2[3]*RootOf(c3,indets(c3)[1]);
          fi;
end:

############################################################################################################
############################################################################################################

####### We replace in an expression the RootOf(x^n + a^n*b) or RootOf(x^n + a^n*b,index=i) which are not in the set
####### extp by a*RootOf(x^n + b).
ReplRoot2:=proc(C,extp)
          local Dx,x,A1,A2,A3,A4,A5,A6,s,n,m,H1,C1,i;
          Dx,x:=op(_Envdiffopdomain);
          A1:= ChooseRootOflistminus(C,extp);
          if A1[1]={} then
             return C
          else
             C1:=subs({seq(A1[3][i]=m[i],i=1..nops(A1[3]))},C);
             A2:= subs({seq(A1[3][i]=m[i],i=1..nops(A1[3]))},A1[1]);
             A3:={seq(n[i]=A2[i],i=1..nops(A2))};
             H1:=map(s-> Arange(A3,A2,s),A2);
             A4:= map(s-> ReplRoot1(s),[op(H1)]);
             A5:=subs({seq(A2[i]=A4[i],i=1..nops(A2))},C1);
             A6:=subs(A3,A5);
             return subs({seq(m[i]=A1[3][i],i=1..nops(A1[3]))},A6);
          fi;
end:

############################################################################################################
############################################################################################################

# We decompose an expression a^n*b^m*c in the form [[a,n],[b,m],[c,1]].
factdecomp:=proc(A)
          local C,E;
          C:= factors(A);
          E:=[op(ifactors(C[1])[2]),op(C[2])];
          return E
end:

############################################################################################################
############################################################################################################

# A = RootOf(x^n + a^n*b) or RootOf(x^n + a^n*b,index=i) where b cannot be written as something with power n times
# something. We convert A to a*RootOf(x^n + b) or a* RootOf(x^n + b,index=i).
ConvirrRoot:=proc(A)
             local q,d,i,a,b,n,e,c,c1,c2,c3,c4,c5,R;
             if indets(A,RootOf)={} then
                return [A,1,1]
             fi;
             c1:=convert(op([1,1],{A}),polynom);
             if indets(c1) = {} then
                return [A,1,1]
             fi;
             c2:=indets(c1,{RootOf,name,function,sqrt}) union indets(c1) minus {indets(c1)[1]};
             if c2 = {} then
                c3:= c1
             else
                c3:=subs({seq(c2[i]=m[i],i=1..nops(c2))},c1);               
             fi;
             d:= collect(c3/lcoeff(c3,indets(c1)[1]),indets(c1)[1],Normalizer);
             n:= degree(d,indets(c1)[1]);
             c4:=indets(d-indets(c1)[1]^n);
             if member(indets(c1)[1],c4) then
                return [A,n,1]
             fi:
             e:= factors(coeff(d,indets(c1)[1],0));
             c:= factdecomp(coeff(d,indets(c1)[1],0));
             a:=1;
             b:= sign(e[1]);
             for i in c do
                b:=b*i[1]^(irem(i[2],n,'q'));
                a:=a*(i[1])^q;
             od;
             R:=a*subs(c1=b+indets(c1)[1]^n,A);
             return [subs({seq(m[i]=c2[i],i=1..nops(c2))},R),n,a];
end:

############################################################################################################
############################################################################################################

# Replace in an expression the RootOf(x^n + a^n*b) or RootOf(x^n + a^n*b,index=i) which are not in the set extp
# by a*RootOf(x^n + b) or  a* RootOf(x^n + b,index=i).
ReplirrRoot:=proc(C,extp)
          local Dx,x,A1,A2,A3,A4,A5,A6,s,n,m,H1,C1,i;
          Dx,x:=op(_Envdiffopdomain);
          A1:= ChooseRootOflistminus(C,extp);
          if A1[1]={} then
             return C
          else
             C1:=subs({seq(A1[3][i]=m[i],i=1..nops(A1[3]))},C);
             A2:= subs({seq(A1[3][i]=m[i],i=1..nops(A1[3]))},A1[1]);
             A3:={seq(n[i]=A2[i],i=1..nops(A2))};
             H1:=map(s-> Arange(A3,A2,s),A2);
             A4:= map(s-> ConvirrRoot(s)[1],[op(H1)]);
             A5:=subs({seq(A2[i]=A4[i],i=1..nops(A2))},C1);
             A6:=subs(A3,A5);
             return subs({seq(m[i]=A1[3][i],i=1..nops(A1[3]))},A6);
          fi;
end:

############################################################################################################
############################################################################################################

# When ext contain a singularity of L or a multiplicity by a rational number of a singularity of L.
Singext:=proc(C,x)
         local Dx,A,s,i,H;
         Dx,x:=op(_Envdiffopdomain);
         A:=table();
         H:=table();
         i:=0;
         for s in C do
            i:=i+1;
            if i=1 then
               if indets(s[1],`RootOf`) <> {} then
                  s[1]:= sqrfree(evala(:-Norm(s[1])),x)[2,1,1];
                  A[i]:= s;
                  H[i]:=s[1];
               else
                  A[i]:= s;
               fi;
            else
               if indets(s[1],`RootOf`) <> {} then
                  s[1]:= sqrfree(evala(:-Norm(s[1])),x)[2,1,1];
                  if member(s[1],convert(H,'set'))=false then
                     A[i]:= s;
                     H[i]:=s[1];
                  fi;
               else
                  A[i]:= s;
               fi;
            fi;
         od;
         A:= convert(A,'list');
         H:= convert(H,'set');
         return A
end:

############################################################################################################
############################################################################################################

# Express singularities of an operator L in terms of monic irreducible polynomials. If infinity is a singularity of L
#              then we take 'infinity' as its irreducible polynomial (just a notation).
#
# extppp = {};
Singular:=proc(L,extppp)
            local Dx,x,A,v,i,T,L1,s,H,C,a,b;
            Dx,x:=op(_Envdiffopdomain);
            # We make the coefficients of L in such a way that they didn't have common factor.
            L1 :=evala(Primpart(L,Dx));
            # A give us the list of minimal polynomial associated to the zeros of the leading coefficient of L1. In
            # other word we express by A the singularities of the operator L1 in terms of monic irreducible
            # polynomials.
            A := PolynomialTools:-Sort(map(i -> i[1],`DEtools/kovacicsols/factors`(lcoeff(L1,Dx), x, extppp)), x);
            # We check if p=infinity is also a singularity of L1. If it is true then we take 'infinity' as its
            # irreducible polynomial (just a notation).
            v:=DEtools['gen_exp'](L1,T,x=infinity,'groundfield'=extppp);
            if nops(v)=1 and (v[1][1]=0 and v[1][2]=1 and v[1][3]=2) then
                    A:=A;
            else
                    A:=[infinity,op(A)];
            fi;
            C:=table();
            for s in A do
               if s = infinity then
                  C[s]:= [s,s];
               else
                  C[s]:= [s,ConvirrRoot(RootOf(s,x))[1]]
               fi;
            od;
            C:= convert(C,'list');
            return Singext(C,x)
end:

############################################################################################################
############################################################################################################

# Express singularities of an third order linear differential operator L which are not apparents in terms of
# [s,pol] where s is the non-apparent singularity and pol its monic irreducible polynomials. If infinity is a
# non-apparent singularity of L then we take 'infinity' as its irreducible polynomial (just a notation).
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# B = Singular(L,ext);
NotAppSing:=proc(L,B,ext)
              local Dx,x,A,T,v,i;
              A:=table();
              Dx,x:=op(_Envdiffopdomain);
              # We check over all the singular points of L
              for i in B do
                 # We compute the generalized exponents of L at p.
                 v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                 # For a point to be an apparent singularity, the necessary condition is that all the generalized
                 # exponents belong to the same module and the ramification index is 1; that means nops(v[i[1]])=1 or
                 # nops(v[i[1]][1])=4 and degree(lhs(v[i[1]][1][-1]),T) = 1.
                 if degree(lhs(v[i[1]][1][-1]),T) <> 1 then
                    A[i]:= i;
                 else
                    if nops(v[i[1]][1])=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # exponents differences have to be positive integers.
                       if (v[i[1]][1][1]>=0 and v[i[1]][1][2]>=0 and v[i[1]][1][3]>=0) and
                          (type(v[i[1]][1][1],integer) and type(v[i[1]][1][2],integer) and
                          type(v[i[1]][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          next
                       else
                          A[i]:= i;
                       fi;
                    else
                       A[i]:= i;
                    fi;
                 fi;
              od;
              A:= convert(A,'list');
              return A
end:

############################################################################################################
############################################################################################################

# We take a third order irreducible linear differential operator and a minimal polynomial "i" related to one of its
# singularities. We return a list of list [e1,e2,e3],[Delta1,Delta2,Delta3],[de1,de2,de3] where for j=1,2,3,
# ej is the generalized exponent, Deltaj is the exponent differences and dej is the ramification index of this
# singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
genexpdiff:=proc(L,i,t,ext)
       local Dx,x,c1,c2,d,extp,extpp,v,e1,e2,e3,de1,de2,de3,f,extp1,H1,H,j,C,T;            
       Dx,x:=op(_Envdiffopdomain);
       v:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
       if i[1]=infinity then 
          v:=subs(1/x=t,v);
       else
          if not has(i[1],x) then
             return NULL
          fi;
          v:=subs(x-i[2]=t,v);
       fi;
       extp := indets(ReplirrRoot(i[2],{}),RootOf) union ext;
       # we see if the singularity is an element of ext
       extp1:= indets(ReplirrRoot(i[2],{}),RootOf) minus ext;
       if nops(v)=3 then
          v[1][1]:=subs(T=t,v[1][1]);
          v[2][1]:=subs(T=t,v[2][1]);
          v[3][1]:=subs(T=t,v[3][1]);
          de1:=1;
          de2:=1;
          de3:=1;
          e1:=convert(series(v[1][1],t=0,100),polynom);
          e2:=convert(series(v[2][1],t=0,100),polynom);
          e3:=convert(series(v[3][1],t=0,100),polynom);
          d:=[e2-e1,e3-e1,e3-e2];
          C:=[[e2,e1],[e3,e1],[e3,e2]];
       elif nops(v)=2 then
          if nops(v[1])=3 then
             v[1]:=subs(T=t,v[1]);
             v[2]:=subs(T=t,v[2]);
             de1:=1;
             de2:=1;
             de3:=1; 
             e1:=convert(series(v[1][1],t=0,100),polynom);
             e2:=convert(series(v[1][2],t=0,100),polynom);
             e3:=convert(series(v[2][1],t=0,100),polynom);
             d:=[e2-e1,e3-e1,e3-e2];
             C:=[[e2,e1],[e3,e1],[e3,e2]];
          else # nops(v[1])=2
             if nops(v[2])=3 then;
                v[1]:=subs(T=t,v[1]); 
                v[2]:=subs(T=t,v[2]); 
                de1:=1;
                de2:=1;
                de3:=1; 
                e1:=convert(series(v[1][1],t=0,100),polynom);
                e2:=convert(series(v[2][1],t=0,100),polynom);
                e3:=convert(series(v[2][2],t=0,100),polynom);
                d:=[e2-e1,e3-e1,e3-e2];
                C:=[[e2,e1],[e3,e1],[e3,e2]];
             else # nops(v[2])=2
                if degree(lhs(v[1][-1]),T)= 2 then ### p=i[2] is an irregular singularity, that 
                                                   ### means p=i[2] is in Sirr.
                   de1:=2;
                   de2:=2;
                   de3:=1; 
                   f:= lcoeff(rhs(v[1][-1]),t)/lcoeff(lhs(v[1][-1]),T);
                   extpp:= indets(f,{RootOf,name});
                   f:=[solve(x^(degree(lhs(v[1][-1]),T))-f,x)];
                   f:=map(s-> op(ReplRoot(s,extpp)),f);
                   f:=map(s->simplify(s),f);
                   f:=map(s->s*t^(1/(degree(lhs(v[1][-1]),T))),f);
                   e1:=convert(series(subs(T=f[1],v[1][1]),t=0,100),polynom);
                   e2:=convert(series(subs(T=f[2],v[1][1]),t=0,100),polynom);
                   e3:=convert(series(subs(T=t,v[2][1]),t=0,100),polynom);
                   d:=[e2-e1,e3-e1,e3-e2];
                   C:=[[e2,e1],[e3,e1],[e3,e2]];
                elif indets(v[1][1],`RootOf`) <> {} and nops(indets(ReplirrRoot(v[1][1],{}),RootOf) minus extp )>0 then
                   de1:=1;
                   de2:=1;
                   de3:=1; 
                   v[1][1]:=subs(T=t,v[1][1]);
                   v[2][1]:=subs(T=t,v[2][1]);
                   c1:=ReplRoot(v[1][1],extp);
                   c2:=map(s-> s-v[2][1],c1);
                   d:=[op(c2),c1[2]-c1[1]];
                   C:=[op(map(s->[s,v[2][1]],c1)),[c1[2],c1[1]]];
                   e1:= v[2][1];
                   e2:= c1[1];
                   e3:=c1[2];
                elif degree(lhs(v[2][-1]),T)= 2 then ### p=i[2] is an irregular singularity, that 
                                                     ### means p=i[2] is in Sirr.
                   de1:=1;
                   de2:=2;
                   de3:=2; 
                   f:= lcoeff(rhs(v[2][-1]),t)/lcoeff(lhs(v[2][-1]),T);
                   extpp:= indets(f,{RootOf,name});
                   f:=[solve(x^(degree(lhs(v[2][-1]),T))-f,x)];
                   f:=map(s-> op(ReplRoot(s,extpp)),f);
                   f:=map(s->simplify(s),f);
                   f:=map(s->s*t^(1/(degree(lhs(v[2][-1]),T))),f);
                   e1:=convert(series(subs(T=t,v[1][1]),t=0,100),polynom);
                   e2:=convert(series(subs(T=f[1],v[2][1]),t=0,100),polynom);
                   e3:=convert(series(subs(T=f[2],v[2][1]),t=0,100),polynom);
                   d:=[e2-e1,e3-e1,e3-e2];
                   C:=[[e2,e1],[e3,e1],[e3,e2]];
                else # indets(v[2],`RootOf`) <> {} and nops(indets(ReplirrRoot(v[2][1],{}),RootOf) minus extp)>0
                   de1:=1;
                   de2:=1;
                   de3:=1;
                   v[1][1]:=subs(T=t,v[1][1]);
                   v[2][1]:=subs(T=t,v[2][1]);
                   c1:=ReplRoot(v[2][1],extp);
                   c2:=map(s->s-v[1][1],c1);
                   d:=[op(c2),c1[2]-c1[1]]; 
                   C:=[op(map(s->[s,v[1][1]],c1)),[c1[2],c1[1]]]; 
                   e1:= v[1][1];
                   e2:= c1[1];
                   e3:=c1[2];
                fi;
             fi;
           fi;
        else # nops(v)=1
           if degree(lhs(v[1,-1]),T)=3 then ### p=i[2] is an irregular singularity, that means p is in Sirr.
              de1:=3;
              de2:=3;
              de3:=3; 
              f:= lcoeff(rhs(v[1][-1]),t)/lcoeff(lhs(v[1][-1]),T);
              extpp:= indets(f,{RootOf,name});
              f:=[solve(x^(degree(lhs(v[1][-1]),T))-f,x)];
              f:=map(s-> op(ReplRoot(s,extpp)),f);
              f:=map(s->simplify(s),f);
              f:=map(s->s*t^(1/(degree(lhs(v[1][-1]),T))),f);
              e1:=convert(series(subs(T=f[1],v[1][1]),t=0,100),polynom);
              e2:=convert(series(subs(T=f[2],v[1][1]),t=0,100),polynom);
              e3:=convert(series(subs(T=f[3],v[1][1]),t=0,100),polynom); 
              d:=[e2-e1,e3-e1,e3-e2];
              C:=[[e2,e1],[e3,e1],[e3,e2]];
           else ### degree(lhs(v[1,-1]),T)=1
              de1:=1;
              de2:=1;
              de3:=1;
              if nops(v[1])=4 then
                 v[1]:=subs(T=t,v[1]);
                 e1:=convert(series(v[1][1],t=0,100),polynom);
                 e2:=convert(series(v[1][2],t=0,100),polynom);
                 e3:=convert(series(v[1][3],t=0,100),polynom);
                 d:=[e2-e1,e3-e1,e3-e2];
                 C:=[[e2,e1],[e3,e1],[e3,e2]];
              elif nops(v[1])=3 then
                 if indets(v[1][1],`RootOf`) <> {} and nops(indets(ReplirrRoot(v[1][1],{}),RootOf) minus extp )>0 then
                    v[1][2]:=subs(T=t,v[1][2]);
                    v[1][1]:=subs(T=t,v[1][1]);
                    c1:=ReplRoot(v[1][1],extp);
                    c2:=map(s-> s-v[1][2],c1);
                    d:=[op(c2),c1[2]-c1[1]];
                    C:=[op(map(s->[s,v[1][2]],c1)),[c1[2],c1[1]]];
                    e1:= v[1][2];
                    e2:= c1[1];
                    e3:=c1[2];
                 else # indets(v[1][2],`RootOf`) <> {} and nops(indets(ReplirrRoot(v[1][2],{}),RootOf) 
                      # minus extp )>0 then
                    v[1][1]:=subs(T=t,v[1][1]);
                    v[1][2]:=subs(T=t,v[1][2]);
                    c1:=ReplRoot(v[1][2],extp);
                    c2:=map(s-> s-v[1][1],c1);
                    d:=[op(c2),c1[2]-c1[1]];
                    C:=[op(map(s->[s,v[1][1]],c1)),[c1[2],c1[1]]];
                    e1:= v[1][1];
                    e2:= c1[1];
                    e3:=c1[2];
                 fi;
              else ### indets(v[1][1],`RootOf`) <> {} the degree of RootOf is 3
                 c1:=ReplRoot(v[1][1],extp);
                 c1:=subs(T=t,c1);
                 d:= [c1[2]-c1[1],c1[3]-c1[1],c1[3]-c1[2]];
                 C:=[[c1[2],c1[1]],[c1[3],c1[1]],[c1[3],c1[2]]];
                 e1:= c1[1];
                 e2:= c1[2];
                 e3:=c1[3];
              fi;
           fi;
        fi;
        H1:= [e1,e2,e3],d,[de1,de2,de3],C,nops(v[1]);
        H:= table();
        for j from 1 to nops([H1]) do
           H[j]:=H1[j]
        od;
        H:= convert(H,'list');
        return H;
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
######################################## BESSEL SQUARE SOLUTIONS ###########################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSq(L2,t,G,ext);       ####### L2 can be our L operator.
# C1 = ExptestBesSqequiv1(L1,C2,x,t,ext)     ####### L1 can be our M operator.
# i = index of an element in C2[5][1].       ####### C2[5][1] = set of non-apparent singularities of L2
towopegenexpdiffBesSq:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equivaBesSq2(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equivaBesSq1(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsingBesSq(L2,t,G,ext);       ####### L2 can be our L operator.
ExptestBesSqequiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= CondiBesSqirr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqequiv2(L1,C2,x,t,ext)       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
ExptestBesSqequiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqequiv2(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equivaBesSq1:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               if extp0 subset ext then
                  return extp3 union ext
               fi;
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equivaBesSq2:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqequiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiffBesSq(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgaugeBesSq1:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,G,ext);
EquivExpgaugeBesSq1:= proc(M,L,R,x,t,_T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=ExptestBesSqequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiffBesSq(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgaugeBesSq1(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqequiv1(L1,C2,x,t,ext)     ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiffBesSq(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgaugeBesSq:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=ExptestBesSqequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,G,ext);
EquivExpgaugeBesSq:= proc(M,L,R,x,t,_T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=ExptestBesSqequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiffBesSq(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgaugeBesSq(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the Besselsquare case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol.
# E = genexpdiff(L,i,t,ext)
CondiBesSqirr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l;
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 3 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              if l <> 1 then
                 return NULL
              fi;
              C:= map(s-> convert(series(subs(t=1/t,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              if (evala(A[1]-A[2])=0 or evala(A[1]+A[2])=0) and (evala(A[3]-2*A[1])=0 or evala(A[3]+2*A[1])=0) then
                 return [E[2],E[4],A,B]
              elif (evala(A[1]-A[3])=0 or evala(A[1]+A[3])=0) and (evala(A[2]-2*A[1])=0 or evala(A[2]+2*A[1])=0) then
                    return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]],[A[1],A[3],A[2]],[B[1],B[3],B[2]]]
              elif (evala(A[2]-A[3])=0 or evala(A[2]+A[3])=0) and (evala(A[1]-2*A[2])=0 or evala(A[1]+2*A[2])=0) then
                    return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]],[A[2],A[3],A[1]],[B[2],B[3],B[1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm "irrsingBesSq"
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext)
irrsingBesSq:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or
                    member(t,indets(A[i[1]][2][3])) then
                    # We are here in the Bessel square case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d:= CondiBesSqirr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the Besselsquare case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol.
# E = genexpdiff(L,i,t,ext)
CondiBesSqreg:=proc(E,t,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if evala(E[2][1]+E[2][2])=0 and evala(E[2][3]+2*E[2][1])=0 and evala(E[2][3]-2*E[2][2])=0 and
                 evala(E[2][1]-E[2][2]+E[2][3])=0 then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif evala(E[2][1]-E[2][3])=0 and evala(E[2][2]-2*E[2][1])=0 and evala(E[2][2]-2*E[2][3])=0 and
                 evala(E[2][1]+E[2][3]-E[2][2])=0 then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif evala(E[2][2]+E[2][3])=0 and evala(E[2][1]-2*E[2][2])=0 and evala(E[2][1]+2*E[2][3])=0 and
                 evala(E[2][2]-E[2][3]-E[2][1])=0 then
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              elif type(evala(E[2][1]+E[2][2]),integer)  and type(evala(E[2][3]+2*E[2][1]),integer) and
                 type(evala(E[2][3]-2*E[2][2]),integer) and type(evala(E[2][1]-E[2][2]+E[2][3]),integer) then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif type(evala(E[2][1]-E[2][3]),integer) and type(evala(E[2][2]-2*E[2][1]),integer) and
                 type(evala(E[2][2]-2*E[2][3]),integer) and type(evala(E[2][1]+E[2][3]-E[2][2]),integer) then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif type(evala(E[2][2]+E[2][3]),integer) and type(evala(E[2][1]-2*E[2][2]),integer) and
                 type(evala(E[2][1]+2*E[2][3]),integer) and type(evala(E[2][2]-E[2][3]-E[2][1]),integer) then
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H1 = NotAppSing(L,G,ext)
# H = irrsingBesSq(L,t,H1,ext)[-1]
regsingtrueBesSq:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                 # must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= CondiBesSqreg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1];
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in two parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer)
#       - removable singularities with integer exponent differences.
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingBesSq(L,t,F,ext)[-1]
# H = regsingtrueBesSq(L,t,R,ext);
SregseptrueBesSq:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingBesSq(L,t,F,ext)[-1]
# Sreg = regsingtrueBesSq(L,t,R,ext);
# RSreg = SregseptrueBesSq(L,Sreg,ext);
UnionSregseptrueBesSq:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################

# Delta = exponent differences of the singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# G = Singular(L,ext);
# F = NotAppSing(L,G,ext);
# R = regsingstruc(L,t,F,ext)
# Sreg = regsingtrue(L,t,R,ext);
# Sirr = irrsingBesSq(L,t,F,ext);
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp);
IrrRegAppsingBesSq:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,
                    H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or
                    member(t,indets(A[i][2][3])) then
                    # We are here in the Bessel square case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d1:= CondiBesSqirr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                    # must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # generalized exponents have to be positive integers and the formal solution must not contain logarithm.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= CondiBesSqreg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1];
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31];
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31];
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= CondiBesSqreg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1];
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31];
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31];
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsingstruc(L,t,ext)[1][i],x),
#######                      with i=position of the minimal polynom of p) is 1.
############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g.
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Bespossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:=table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     if member(-h,convert(C,'set'))=false and member(h,convert(C,'set'))=false then
                           C[k]:= h;
                     fi;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# G = Singular(L,ext);
# Sirr = IrrRegAppsingBesSq(L,t,G,ext)[1];
BesSqSubst:=proc(L,x,t,Sirr,ext)
               local Dx,m,A,ff,B,extp0,extp1,extp2,extp3,extp4,extp5,v,v1,i,j,s;
               Dx,x:=op(_Envdiffopdomain);
               A:=table();
               i:=0;
               for s in Sirr[1] do
                  i:=i+1;
                  m:= ldegree(ReplirrRoot(Sirr[3][i][1],{}),t);
                  if m=FAIL then
                     return NULL
                  fi;
                  ff := add( (coeff(convert(eval(subs(csgn=1,simplify(ReplirrRoot(Sirr[3][i][1],{})) assuming(t>0))),                                                   polynom),t,j)/(2*j))*(`if`(s[1]=infinity,1/x,x-s[2]))^j,j=m .. -1);
                  ff:=evala(Reduce(ff));
                  if s[1]=infinity then
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All the names or RootOf
                                           # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp5:= ext union extp1
                  else
                     extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                     extp2:= extp1 minus extp0; # All the names or RootOf
                                                 # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp3:= ext minus extp0;
                     extp4:= extp2 union extp3;
                     v1:={};
                     for v in extp4 do
                        if extp0 intersect indets(v,{RootOf,name}) <> {} then
                           v1:= v1 union {v}
                        fi;
                     od:
                     extp5:= extp4 minus v1;
                  fi;
                  A[s]:= evala(Trace(convert(evala(ff),polynom),extp5,extp5),x)
               od;
               A:= convert(A,'list');
               B := Bespossibility(A);
               B:= map(j-> `if`(member(x,indets(j,{name})),j,NULL),B);
               if B=[] then
                  return NULL
               fi;
               return map(s-> Normalizer(s),B)
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
ZeroesmultordpolBesSq:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsingBesSq(L,t,H,ext)[2];
# B = ZeroesmultordpolBesSq(f,x);
# n = nops(B)
newregsingBesSq:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],CondiBesSqreg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],CondiBesSqreg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################

# Input: a function f(x) and a minimal polynomial i. The minimal polynomial with Root infinity is infinity.
# Output: f(x)+c such that f(p)+c=0 with p=RootOf(i,x). If i=infinity then f(x) is returned i.e c=0.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,H,ext);
# Sirr = R[1];
# f = element of HypBesSqSubst(L,x,Sirr,ext);
# i = element of R[2][1].
changeconstantBesSq:= proc(f,x,i,ext)
                 local Dx,c,A,ff,C;
                 Dx,x:=op(_Envdiffopdomain);
                 if i[1]=infinity then
                    A:= f;
                 else
                    ff:=numer(Normalizer(f+C));
                    c:=SolveTools:-Linear({coeffs(convert(collect(eval(subs(csgn=1,simplify(evala(Rem(ff,i[1],x)))
                                                               assuming(x>0))),x),polynom),x)},{C});
                    if c=NULL or nops(c) <> 1 then
                       return NULL;
                    else
                       A:=Normalizer(f+subs(c[1],C));
                    fi;
                 fi;
                 return factor(A);
end:

############################################################################################################
############################################################################################################

# Give the good candidates for f by looking at those candidates which vanish at elements of regsingtrue (all
# elements of regsingtrue should give the same constant c). Note that regsingtrue <> {} in this case.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,H,ext);
# Sirr = R[1];
# F = BesSqSubst(L,x,Sirr,ext).
CandichangvarBesSq:=proc(F,R,ext) ### For all the case except the integer case.
                local Dx,x,i,f,A,j,s,B1,n1,SR,n;
               Dx,x:=op(_Envdiffopdomain);
               A:= table();
               n:=nops(R[2][1]);
               for s in F do
                  j:=0;
                  f:= changeconstantBesSq(s,x,R[2][1][1],ext);
                  if f=NULL then
                     next
                  fi;
                  j:=j+1;
                  if n>1 then
                     for i from 2 to n do
                        if R[2][1][i][1]=infinity then
                           if not degree(numer(f),x)<degree(denom(f),x) then
                              break
                           fi;
                           j:=j+1;
                        else
                           if not evala(Rem(numer(f),R[2][1][i][1],x))=0 then
                              break
                           fi;
                           j:=j+1;
                        fi;
                     od;
                  fi;
                  if j <> n then
                     next
                  fi;
                  B1:= ZeroesmultordpolBesSq(f,x);
                  if B1=NULL then
                     next
                  fi;
                  n1:= nops(B1);
                  SR:= newregsingBesSq(L,t,x,B1,R[2],n1,ext);
                  A[s] := [f,SR[2],SR[1]];
               od;
               A:= convert(A,'set');
               if A={} then
                  return NULL
               else
                  return A
               fi;
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate v for the Bessel parameter must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = ZeroesmultordpolBesSq(f,x);
# n = nops(B1);
# RS = newregsingBesSq(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# v = candidate for the Bessel parameter.
CondinuexpBesSq:= proc(E1,E2,m,v,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(-2*m*v-E2[1]),integer) and type(evala(2*m*v-E2[2]),integer) and
                 type(evala(4*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]+2*m*v),integer) and
                 type(evala(E1[2][1]-2*m*v),integer)) then
                 return true
              elif (type(evala(2*m*v-E2[1]),integer) and type(evala(-2*m*v-E2[2]),integer) and
                 type(evala(4*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][1]-E1[3][2]),integer) and type(evala(E1[1][1]-2*m*v),integer) and
                 type(evala(E1[2][1]+2*m*v),integer)) then
                 return true
              elif (type(evala(2*m*v-E2[1]),integer) and type(evala(-2*m*v-E2[2]),integer) and
                 type(evala(-4*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-2*m*v),integer) and
                 type(evala(E1[2][1]+2*m*v),integer)) then
                 return true
              elif (type(evala(-2*m*v-E2[1]),integer) and type(evala(2*m*v-E2[2]),integer) and
                 type(evala(-4*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][1]-E1[3][2]),integer) and type(evala(E1[1][1]+2*m*v),integer) and
                 type(evala(E1[2][1]-2*m*v),integer)) then
                 return true
              elif (type(evala(2*m*v-E2[1]),integer) and type(evala(2*m*v-E2[2]),integer) and
                 type(evala(4*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]+2*m*v),integer) and
                 type(evala(E1[2][1]-2*m*v),integer)) then
                 return true
              elif (type(evala(2*m*v-E2[1]),integer) and type(evala(2*m*v-E2[2]),integer) and
                 type(evala(4*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][1]-2*m*v),integer) and
                 type(evala(E1[2][2]+2*m*v),integer)) then
                 return true
              elif (type(evala(2*m*v-E2[1]),integer) and type(evala(-2*m*v-E2[2]),integer) and
                 type(evala(4*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][2]+2*m*v),integer) and
                 type(evala(E1[2][2]-2*m*v),integer)) then
                 return true
              elif (type(evala(-2*m*v-E2[1]),integer) and type(evala(2*m*v-E2[2]),integer) and
                 type(evala(4*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][2]-2*m*v),integer) and
                 type(evala(E1[2][2]+2*m*v),integer)) then
                 return true
              elif (type(evala(-2*m*v-E2[1]),integer) and type(evala(-2*m*v-E2[2]),integer) and
                 type(evala(-4*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-2*m*v),integer) and
                 type(evala(E1[2][1]+2*m*v),integer)) then
                 return true
              elif (type(evala(-2*m*v-E2[1]),integer) and type(evala(-2*m*v-E2[2]),integer) and
                 type(evala(-4*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][1]+2*m*v),integer) and
                 type(evala(E1[2][2]-2*m*v),integer)) then
                 return true
              elif (type(evala(-2*m*v-E2[1]),integer) and type(evala(2*m*v-E2[2]),integer) and
                 type(evala(-4*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][2]-2*m*v),integer) and
                 type(evala(E1[2][2]+2*m*v),integer)) then
                 return true
              elif (type(evala(2*m*v-E2[1]),integer) and type(evala(-2*m*v-E2[2]),integer) and
                 type(evala(-4*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][2]+2*m*v),integer) and
                 type(evala(E1[2][2]-2*m*v),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

# Input: A,B in k
# Output: r0,r1 such that A=r0+r1*B.
compare :=proc(A,B)
          local iszero,E,sol,a,b;
          iszero := A- (a+b*B);
          iszero := evala(Expand(numer(Normalizer(iszero))));
          iszero := convert(iszero,RootOf);
          do
              E := indets(iszero,RootOf);
              if E = {} then break fi;
              iszero := subs(E[1]=newvar(),iszero);
          od;
          sol := SolveTools:-Linear({coeffs(expand(iszero),indets(iszero,{name,symbol}) minus {a,b})},{a,b});
          if not type(sol,set) or not type(map(rhs,sol),set(rational)) then ####### error NO_SOL
             return NULL
          fi;
          op(subs(sol, [a,b]))
end:

############################################################################################################
############################################################################################################

# Input: A monic polynomial f\in K[x] and a positive integer p\in\N
# Output: g\in K[x], such that if there exists a solution to y^p=f, then y=g is such a solution.
ispower := proc(f,x,p)
           local d,n,A,Ap,a,i,b;
           if p=1 or degree(f,x)=0 then
                return f
           fi;
           d:=degree(f,x);
           n:=d/p;
           if not type(n,integer) then
                return NULL
           fi;
           A:=add(a[i]*x^i,i=0..n-1)+x^n;
           Ap:=collect(A^p,x);
           for i from 1 to n do
                # Solve a linear equation in one unknown:
                b[n-i]:=solve(subs({seq(a[n-j]=b[n-j],j=1..i-1)},coeff(Ap,x,d-i))-coeff(f,x,d-i),a[n-i] );
           od;
           Ap:=subs({seq(a[n-j]=b[n-j],j=1..n)},Ap);
           subs({seq(a[n-j]=b[n-j],j=1..i-1)},A)
end:

############################################################################################################
############################################################################################################

# Output: a new variable
newvar := proc()
          local t;
          t
end:

############################################################################################################
############################################################################################################
##########################################   LOGARITHMIC CASE   ############################################
############################################################################################################
############################################################################################################

# We compute the Bessel square parameter nu related to a candidate of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# Sreg = R[2];
# g = candidate for the change of variable parameter f.
findnuLogBesSq := proc(L,g,Sreg,x,t)
             local Dx,i,v,n,m,a;
             Dx,x:=op(_Envdiffopdomain);
             m:= nops(Sreg[1]);
             n := 2*max(degree(numer(g),x),degree(denom(g),x));
             ####### Here all the exponent differences at the zeroes of g are integers
             a:=0;
             for i from 1 to m do
                if Sreg[1][i][1]=infinity then
                   a:= a+abs(evala(ReplirrRoot(Sreg[3][i][1],{})))
                else
                   a:= a+abs(evala(ReplirrRoot(Sreg[3][i][1],{}))*degree(Sreg[1][i][1],x))
                fi;
             od;
             v:= floor(a/n);
             return {[[v],g]}
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# F = BesSqSubst(L,x,t,R[1],ext);
findBesSqln:=proc(L,R,F,x,t,ext)
                    local Dx,H,G1,g;
                    Dx,x:=op(_Envdiffopdomain);
                    G1:= CandichangvarBesSq(F,R,ext);
                    if G1=NULL then
                       return NULL
                    fi;
                    H:=table();
                    for g in G1 do
                       H[g]:= op(findnuLogBesSq(L,g[1],R[2],x,t))
                    od;
                    H:= convert(H,'set');
                    if H={} then
                       return NULL
                    else
                       return H
                    fi;
end:

############################################################################################################
############################################################################################################
#############################################   IRRATIONAL CASE   ##########################################
############################################################################################################
############################################################################################################

leadcoeffRootnameinter:=proc(A,x,ext1)
                    local Dx,l,l1,ira,ira1,ira2,sum1,sum2;
                    Dx,x:=op(_Envdiffopdomain);
                    l:=A;
                    l1:=ext1;
                    while l1<>{} and (l1 intersect ext1 <> {}) do
                       ira1:=ChooseRootOflistinter(l,l1)[1];
                       if ira1<>{} then
                          ira:=ira1[1];
                       else
                          ira2:= l1 intersect ext1:
                          ira:=ira2[1];
                       fi;
                       sum1:=convert(collect(l,ira,evala),polynom);
                       sum2:= `if`(ldegree(sum1,ira)<0,convert(subs(ira=1/ira,sum1),polynom),sum1);
                       l:= lcoeff(sum2,ira);
                       l1:= indets(l,{RootOf,name});
                    od;
                    return l
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,ext);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# F = BesSqSubst(L,x,t,R[1],ext);
findBesSqIrr:=proc(L,R,F,x,t,ext)
                    local Dx,vf,ira,ira1,ira2,ext1,ext2,i,f,ff,h,hh,m,a,b,l,l1,G1,V,vi,n,r,sum1,j,H,k,e,v1,v2,v3;
                    Dx,x:=op(_Envdiffopdomain);
                    G1:= CandichangvarBesSq(F,R,ext);
                    if G1=NULL then
                       return NULL
                    fi;
                    vf:=table();
                    n:=nops(R[2][1]);
                    ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                    ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                    if n > 1 then
                       for i from 2 to n do
                          ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                          ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                          ext1:= ext1 intersect ext2;
                       od;
                    fi;
                    ####### Let k be the field of constants that we want to work on such that L is
                    ####### in k(x)[\partial].
                    ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
                    ####### irrational case.
                    H:=table();
                    r[1]:=1;
                    for i from 1 to n do
                       H[i]:= ReplirrRoot(R[2][3][i][1],{});
                       l1:= leadcoeffRootnameinter(H[i],x,ext1);
                       if type(l1,rational)=false or l1=0 then
                          return NULL
                       fi;
                       b[i]:= l1*(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));
                       if i>1 then
                          r[i] := Normalizer(abs(b[i]/b[1])); #  r[i]=(m_i*degree(p_i,x))/(m_1*degree(p_1,x))
                                                         #  m_i is the multiplicity order as zero of f
                                                         #  associated to i.
                          if not type(r[i],rational) then
                             return NULL
                          fi;
                       fi;
                    od;
                    H:= convert(H,'list');
                    l:= ilcm( seq(denom(r[i]),i=1..n));
                    h:= mul(`if`(R[2][1][i][1]=infinity,1,(R[2][1][i][1])^(r[i]*l/degree(R[2][1][i][1],x))),
                        i=1..n);
                    for f in G1 do
                       ff:=numer(f[1]);
                       if degree(h,x)=0 then
                          a:=1;
                       else
                          a:=degree(ff,x)/degree(h,x);
                       fi;
                       if not type(a,integer) then next fi;
                       ff:=collect(ff/lcoeff(ff,x),x,Normalizer);
                       hh:=collect(h/lcoeff(h,x),x,Normalizer);
                       if Normalizer(ff-hh^a) <> 0 then
                          next;
                       fi;
                       # now h^a=ff
                       for i from 1 to n do
                          m[i]:= `if`(R[2][1][i][1]=infinity,degree(denom(f[1]),x) - degree(ff,x),
                                            r[i]*a*l/degree(R[2][1][i][1],x));
                          vi:= [seq(H[i]+j,j=0..2*m[i]-1)]/(2*m[i]);
                          if vi=[] then
                             return NULL
                          fi;
                          vi:= map(s-> `if`(CondinuexpBesSq(R[2][4][i],R[2][3][i],m[i],s,ext)=false,NULL,s), vi);
                          if vi=[] then
                             return NULL
                          fi;
                          vi:=map(s-> ReplRoot2(s,{}),vi);
                          if i = 1 then
                             V:=vi
                          else
                             V:=[seq(`if`(MatchInt1(V,vi[i])<>false, MatchInt1(V,vi[i]), NULL),i=1..nops(vi))];
                          fi;
                          if V=[] then
                             return NULL
                          fi
                       od;
                       vf[f]:= [V,factor(f[1])];
                    od;
                    vf:= convert(vf,'list');
                    if vf=[] then
                       return NULL
                    else
                       return vf
                    fi;
end:

############################################################################################################
############################################################################################################
############################################   BASEFIELD CASE   ############################################
############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# F = BesSquarSubst(L,x,t,R[1],ext);
findBesSqvfBasfield:= proc(L,R,F,x,t,ext)
              local Dx,A,i,f,ff,h,hh,r0,r1,m,a,l,G1,vf,V,vi,j,G,H1,n;
              Dx,x := op(_Envdiffopdomain);
              G1:= CandichangvarBesSq(F,R,ext);
              if G1=NULL then
                 return NULL
              fi;
              vf:= table();
              n:=nops(R[2][1]);
              # all exponent differences are a rational multiple of d[1]
              # compute these rational coefficients
              r0[1]:= 0;
              r1[1]:= 1;
              if n>1 then
                 for i from 2 to n do
                    r0[i],r1[i]:= compare(ReplirrRoot(R[2][3][i][1]),ReplirrRoot(R[2][3][1][1],{}));
                    if r0[i]=NULL or r1[i]=NULL then
                       return NULL
                    fi;
                 od;
              fi;
              l:= ilcm( seq(denom(r1[i]),i=1..n));
              h:= mul((`if`(R[2][1][i][1]=infinity,1,R[2][1][i][1]))^(r1[i]*l),i=1..n);
              for f in G1 do
                 ff:=numer(f[1]);
                 a:= `if`(degree(h,x)=0,1,degree(ff,x)/degree(h,x));
                 if not type(a,integer) then next fi;
                 ff:=collect(ff/lcoeff(ff,x),x,Normalizer);
                 hh:=collect(h/lcoeff(h,x),x,Normalizer);
                 if Normalizer(ff-hh^a) <> 0 then
                    next;
                 fi;
                 # now h^a=ff
                 for i from 1 to n do
                    m[i]:= `if`(R[2][1][i][1]=infinity,degree(denom(f[1]),x) - degree(ff,x),
                    r1[i]*a*l);
                    vi:= [seq(ReplirrRoot(R[2][3][i][1],{})+j,j=0..2*m[i]-1)]/(2*m[i]);
                    if vi=[] then
                       return NULL
                    fi;
                    vi:= map(s-> `if`(CondinuexpBesSq(R[2][4][i],R[2][3][i],m[i],s,ext)=false,NULL,s), vi);
                    if vi=[] then
                       return NULL
                    fi;
                    vi:=map(s-> ReplRoot2(s,{}),vi);
                    if i = 1 then
                       V:=vi
                    else
                       V:=[seq(`if`(MatchInt1(V,vi[i])<>false, MatchInt1(V,vi[i]), NULL),i=1..nops(vi))];
                    fi;
                    if V=[] then
                       return NULL
                    fi
                 od;
                 vf[f]:= [V,factor(f[1])];
              od;
              vf:= convert(vf,'list');
              if vf=[] then
                 return NULL
              else
                 return vf
              fi;
end:

############################################################################################################
############################################################################################################
###############################################   RATIONAL CASE   ##########################################
############################################################################################################
############################################################################################################

####### We test if the numerator of f is a denom(2*v)^th power function.
powerfunc:=proc(f,v,x)
           local Dx,newh,c,h;
           Dx,x := op(_Envdiffopdomain);
           h:=numer(f);
           c:=lcoeff(h,x);
           newh:=ispower(h/c,x,denom(2*v));
           if newh=NULL then
              return NULL
           fi;
           if Normalizer(h/c-newh^denom(2*v)) <> 0 then
              return NULL
           fi;
           return true
end:

############################################################################################################
############################################################################################################

# Let f be a candidate for the change of variables parameter. We compute here the candidates for nu modulo
# an integer related to f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# F = BesSqSubst(L,x,t,R[1],ext);
# B1 = CandichangvarBesSq(F,R,ext) ;
# f = element of B1;
# n0 = nops(R[2][1]);
helpBesSqvfrat:=proc(f,x,n0,ext)
                  local Dx,i,j,vi,V,n1,n,s;
                  Dx,x := op(_Envdiffopdomain);
                  n:= nops(f[3][1]);
                  n1:= 0;
                  for i from 1 to n do
                     if f[2][i]=0 then
                        next
                     fi;
                     vi := [seq(ReplirrRoot(f[3][3][i][1],{})+j,j=0..2*f[2][i]-1)]/(2*f[2][i]);
                     vi:=[seq(`if`(MatchInt([0,1/2,op(vi[1..j-1])],vi[j]), NULL, vi[j]),j=1..nops(vi))];
                     if vi=[] then
                        return NULL
                     fi;
                     if i>n0 then
                        vi:= map(s-> `if`(powerfunc(f[1],s,x)=false or
                                       CondinuexpBesSq(f[3][4][i],f[3][3][i],f[2][i],s,ext)=false,NULL,s),vi);
                     else
                        vi:= map(s-> `if`(CondinuexpBesSq(f[3][4][i],f[3][3][i],f[2][i],s,ext)=false,NULL,s),vi);
                     fi;
                     if vi=[] then
                        return NULL
                     fi;
                     n1:= n1+1;
                     if n1=1 then
                        V:= vi
                     else
                        V:=[seq(`if`(MatchInt1(V,vi[j])<>false, MatchInt1(V,vi[j]), NULL),j=1..nops(vi))];
                     fi;
                     if V=[] then
                        return NULL
                     fi;
                  od;
                  V:=map(j->abs(j),V);
                  return [V,f[1]];
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# F = BesSqSubst(L,x,t,R[1],ext);
findBesSqvfRat:=proc(L,R,F,x,t,ext)
                  local Dx,G1,f,vf,H,n;
                  Dx,x := op(_Envdiffopdomain);
                  G1:= CandichangvarBesSq(F,R,ext);
                  if G1=NULL then
                     return NULL
                  fi;
                  vf:= table();
                  n:= nops(R[2][1]);
                  for f in G1 do
                     H:= helpBesSqvfrat(f,x,n,ext);
                     if H=NULL then
                        next
                     fi;
                     vf[f] := H;
                  od;
                  vf:= convert(vf,'list');
                  if vf=[] then
                     return NULL
                  else
                     return vf
                  fi;
end:

############################################################################################################
############################################################################################################
############################################   INTEGER CASE   ##############################################
############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSq(L,t,E,ext);
# F = BesSqSubst(L,x,t,R[1],ext);
findBesSqvfInt:= proc(L,R,F,x,t,ext)
           local Dx,ff,f,n,i,g,lc,a,k,v,C,c,cc,fctrs,vf,s,inff,V,B1,n1,RS,v1,j,res,V1;
           Dx,x:=op(_Envdiffopdomain);
           vf:=table();
           for f in F do
               ff:=Normalizer(f+C);
               inff:=map(s->s[1],R[1][1]);
               if member(infinity,{op(inff)}) then
                   n:=degree(numer(ff));
               else
                   n:=degree(denom(ff));
               fi;
               # denom(2*nu) divises n. So if n=1 then denom(2*nu)=1 and that means nu belong to 1/2+Z or
               # nu belong to Z which is not good. nu also have to be different from 0 (see the proof of
               # one of corollary in the thesis related to this part).
               if n<=1 then
                  next
               fi;
               g:=numer(ff);
               lc:=lcoeff(g,x);
               res:=table();
               for i from 2 to n do
                  # i must be a divisor of n
                  V1:= table();
                  if irem(n,i)=0 then
                     # f must be in the form g^i.
                     a:=ispower(collect(g/lc,x,Normalizer),x,i);
                     if a=NULL then
                        next
                     fi;
                     s:=numer(Normalizer(evala(Content(g-lc*a^i,x))));
                     s:=evala(Factors(s,ext));
                     cc:={seq(`if`(degree(k[1],C)=1,-coeff(k[1],C,0)/coeff(k[1],C,1),NULL),k=s[2])};
                     if member(infinity,{op(inff)})=false then
                        fctrs:=sqrfree(subs(C=0,g),x);
                        # the multiplicity of each factor of g is a multiple of i
                        if add(modp(k[2],i),k=fctrs[2])=0 then
                           cc:=cc union {0};
                        fi;
                     fi;
                     if cc <> {} then
                        for c in cc do
                           g:= factor(subs(C=c,ff));
                           B1:= ZeroesmultordpolBesSq(g,x);
                           if B1=NULL then
                              next
                           fi;
                           n1:= nops(B1);
                           RS:= newregsingBesSq(L,t,x,B1,R[2],n1,ext);
                           V:=table();
                           for k from 1 to i-1 do
                              v := k/(2*i);
                              if denom(2*v)<=i then
                                 v1:=[seq(`if`(CondinuexpBesSq(RS[1][4][j],RS[1][3][j],RS[2][j],v,ext)=false,NULL,v),
                                        j=1..n1)];
                                 if nops(v1)<>n1 then
                                    next
                                 else
                                    V[v]:= v
                                 fi;
                              fi;
                           od;
                           V:= convert(V,'list');
                           if V <> [] then
                              V1[c] := [V,factor(g)]
                           fi;
                        od;
                     fi;
                  fi;
                  V1:= convert(V1,'list');
                  if V1<>[] then
                     res[i]:= op(V1);
                  fi;
               od;
               res:= convert(res,'list');
               if res<>[] then
                  vf[f]:= op(res);
               fi;
            od;
            vf:= convert(vf,'list');
            if vf=[] then
               return NULL
            else
               return vf
            fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature 1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1, l2, p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name,sqrt}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,E,ext);
# F = Hyp2F2Subst(L,x,t,R[1],ext);
helpBesSqSolutions1:= proc(L,R,F,x,t,T,ext)
           local Dx,ext1,ext2,n1,i,B,s,H;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H:= findBesSqvfInt(L,R,F,x,t,ext);
           elif R[3][3]<>[] then
              H:= findBesSqln(L,R,F,x,t,ext);
           elif R[3][2]<>[] then
              H:= findBesSqvfRat(L,R,F,x,t,ext);
           else
              ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
              ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
              n1:=nops(R[2][1]);
              if n1 > 1 then
                 for i from 2 to n1 do
                    ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                    ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                    ext1:= ext1 intersect ext2;
                 od;
              fi;
              if ext1 <> {} then
                 H:= findBesSqIrr(L,R,F,x,t,ext);
              else
                 H:= findBesSqvfRat(L,R,F,x,t,ext);
              fi;
           fi;
           if H=NULL then
              return NULL
           else
              return H
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name,,sqrt}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsingBesSq(L,t,G,ext);
# a = element of helpBesSqSolutions1(L,r,RSreg,info,ext,x);
helpBesSqSolutions2:= proc(L,a,A,x,t,ext)
           local Dx,LBB,LBB1,eq,Y,J,n1,M,L22,e1,R,i,T;
           Dx,x:=op(_Envdiffopdomain);
           n1:= nops(a[1]);
           eq:=HolonomicDE(BesselI(e1, x)^2,Y(x));
           LBB1:=de2diffop(eq, Y(x));
           for i from 1 to n1 do
              LBB:=subs(e1=a[1][i],LBB1);
              M:= ChangeOfVariables(LBB,a[2]);
              R:=EquivExpgaugeBesSq(M,L,A,x,t,T,ext);
              if R<>NULL then
                 return [{[a[1][i],R]},a[2]]
              fi;
           od;
           return NULL
end:

############################################################################################################
############################################################################################################

helpBesSqSolutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

FindBesSqSolutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsingBesSq(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= BesSqSubst(L,x,t,R[1],ext);
           if F=NULL then
              return NULL
           fi;
           C:=table();
           H:= helpBesSqSolutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= helpBesSqSolutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

BesSqSolutions:= proc(L)
           local Dx,x,s,k,F,E,H1,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:=table();
           H1:= FindBesSqSolutions(L);
           if H1=NULL then
              return "No Bessel square type solutions"
           fi;
           F:= map(s-> [op(s[1]),s[2]],H1);
           for C in F do
              E:= subs(C[1][2]=map(k->helpBesSqSolutions3(k),C[1][2]),C[1]);
              E:= seq([E[1],op(E[2][i]),C[2]],i=1..nops(E[2]));
              H[C]:= E
           od;
           H:= convert(H,'set');
           return H
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
####################################### BESSEL SQUARE ROOT SOLUTIONS #######################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSqRoot(L2,t,G,ext);       ####### L2 can be our L operator.
# C1 = ExptestBesSqRootequiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator.
# i = index of an element in C2[5][1].     ####### C2[5][1] = set of non-apparent singularities of L2
towopegenexpdiffBesSqRoot:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;          
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do   
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then 
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equivaBesSqRoot2(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;                     
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);                  
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;                          
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equivaBesSqRoot1(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;          
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsingBesSqRoot(L2,t,G,ext);       ####### L2 can be our L operator.
ExptestBesSqRootequiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= CondiBesSqRootirr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSqRoot(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqRootequiv2(L1,C2,x,t,ext)       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
ExptestBesSqRootequiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSqRoot(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqRootequiv2(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equivaBesSqRoot1:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               if extp0 subset ext then
                  return extp3 union ext
               fi;
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equivaBesSqRoot2:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSqRoot(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqRootequiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiffBesSqRoot(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgaugeBesSqRoot1:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,G,ext);
EquivExpgaugeBesSqRoot1:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=ExptestBesSqRootequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiffBesSqRoot(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgaugeBesSqRoot1(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingBesSqRoot(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = ExptestBesSqRootequiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiffBesSqRoot(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgaugeBesSqRoot:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,G,ext);
EquivExpgaugeBesSqRoot:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=ExptestBesSqRootequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiffBesSqRoot(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgaugeBesSqRoot(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the Bessel square Root case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol.
# E = genexpdiff(L,i,t,ext)
CondiBesSqRootirr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l; 
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 3 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              if l > 2 then
                 return NULL
              fi;  
              C:= map(s-> convert(series(subs(t=1/t^l,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              if type(evala(B[1]*2),integer)=false or type(evala(B[2]*2),integer)=false or type(evala(B[3]*2),integer)=false then
                 return NULL
              fi;
              if (evala(A[1]-A[2])=0 or evala(A[1]+A[2])=0) and (evala(A[3]-2*A[1])=0 or evala(A[3]+2*A[1])=0) then
                 return [E[2],E[4],A,B]
              elif (evala(A[1]-A[3])=0 or evala(A[1]+A[3])=0) and (evala(A[2]-2*A[1])=0 or evala(A[2]+2*A[1])=0) then
                   return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]],[A[1],A[3],A[2]],[B[1],B[3],B[2]]]
              elif (evala(A[2]-A[3])=0 or evala(A[2]+A[3])=0) and (evala(A[1]-2*A[2])=0 or evala(A[1]+2*A[2])=0) then
                   return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]],[A[2],A[3],A[1]],[B[2],B[3],B[1]]]
              else
                 return NULL
              fi;             
end: 

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm "irrsingBesSqRoot"
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext)
irrsingBesSqRoot:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do 
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular 
                 # singular point is that: at least one of its exponent differences                      
                 # contains the parameter t.  
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or member(t,indets(A[i[1]][2][3])) then
                    # We are here in the Bessel square root case which has more other                       
                    # conditions for a point to be an irregular singularty of L 
                    d:= CondiBesSqRootirr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of                     
                    # the irregular singular point p:=RootOf(i[1],x) using the exponent                     
                    # differences and the ramification index at this point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then 
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the BesselsquareRoot case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol.
# E = genexpdiff(L,i,t,ext)
CondiBesSqRootreg:=proc(E,t,ext)
              local Dx,x; 
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(E[2][1]-E[2][2]),integer) or type(evala(E[2][1]+E[2][2]),integer)) and 
                 (type(evala(E[2][3]-2*E[2][1]),integer) or type(evala(E[2][3]+2*E[2][1]),integer)) then  
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif (type(evala(E[2][1]-E[2][3]),integer) or type(evala(E[2][1]+E[2][3]),integer)) and 
                   (type(evala(E[2][2]-2*E[2][1]),integer) or type(evala(E[2][2]+2*E[2][1]),integer)) then  
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif (type(evala(E[2][2]-E[2][3]),integer) or type(evala(E[2][2]+E[2][3]),integer)) and 
                   (type(evala(E[2][1]-2*E[2][2]),integer) or type(evala(E[2][1]+2*E[2][2]),integer)) then 
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H1 = NotAppSing(L,G,ext)
# H = irrsingBesSqRoot(L,t,H1,ext)[-1]
regsingtrueBesSqRoot:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G; 
              Dx,x:=op(_Envdiffopdomain);
              B:=table();                                       
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do 
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point,                     
                 # its exponent differences must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or 
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and 
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= CondiBesSqRootreg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1]; 
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then 
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in two parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer)
#       - removable singularities with integer exponent differences.
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingBesSqRoot(L,t,F,ext)[-1]
# H = regsingtrueBesSqRoot(L,t,R,ext);
SregseptrueBesSqRoot:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingBesSqRoot(L,t,F,ext)[-1]
# Sreg = regsingtrueBesSqRoot(L,t,R,ext);
# RSreg = SregseptrueBesSqRoot(L,Sreg,ext);
UnionSregseptrueBesSqRoot:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################

# Delta = exponent differences of the singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingBesSqRoot(L,t,F,ext)[-1]
# Sreg = regsingtrueBesSqRoot(L,t,R,ext);
# Sirr = irrsingBesSqRoot(L,t,F,ext);
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# R = Singular(L,extppp);
IrrRegAppsingBesSqRoot:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,
                    n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do 
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular                         
                 # singular point is that: at least one of its exponent differences                      
                 # contains the parameter t.  
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or 
                    member(t,indets(A[i][2][3])) then
                    # We are here in the Bessel square root case which has more other                       
                    # conditions for a point to be an irregular singularty of L 
                    d1:= CondiBesSqRootirr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of                     
                    # the irregular singular point p:=RootOf(i[1],x) using the exponent                     
                    # differences and the ramification index at this point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point,                     
                    # its exponent differences must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent                          
                       # singularity: all the generalized exponents have to be positive                        
                       # integers and the formal solution must not contain logarithm.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and 
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and 
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then 
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or 
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and 
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= CondiBesSqRootreg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1]; 
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31]; 
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and 
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31]; 
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or 
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and 
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= CondiBesSqRootreg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1]; 
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31]; 
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and 
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31]; 
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then 
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then 
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then 
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then 
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsingstruc(L,t,ext)[1][i],x),
####### with i=position of the minimal polynom of p) is either 1 or 2 since we square the parameter of change
####### of variables.
#######
#######                          g=f^2=(cA1.A2^d)/B where B is the denominator of f^2.
#######                                We use "Delta1" for regsingtrue(L,t,ext).
############################################################################################################
############################################################################################################

# We compute the product of (x-p)^(2*m) where p is an irregular singularity (pole) of f and m is its multiplicative
# order. That will give us the denominator of g=f^2.
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext);
# C = irrsingBesSqRoot(L,t,H,ext) or IrrRegAppsingBesSqRoot(L,t,G,ext)[1];
PolirrProdBesSqRoot:=proc(L,C,x,ext)
            local Dx,b,i,s;
            Dx,x:=op(_Envdiffopdomain);
            b:=1;
            i:=0;
            for s in C[1] do
               i:=i+1;
               if s[1] <> infinity then
                  b:=b*(s[1])^(2*C[4][i])
               fi;
            od;
            return b
end:

############################################################################################################
############################################################################################################

degpol:=proc(a,x,t,ext)
          local Dx,l1,d;
          Dx,x:=op(_Envdiffopdomain);
          l1:=series(subs(t=1/x,a),x=0,100);
          l1:=eval(subs(csgn=1,simplify(l1) assuming(x>0)));
          d:= degree(convert(l1,polynom), x);
          if d=FAIL then
             return false
          else
             return d
          fi;
end:

############################################################################################################
############################################################################################################

# We take as input A = irrsingBesSq(L,t,ext) and l the minimal integer m such that f belong to k(x^(1/m)). We return a
# list of list in the form [[pol,s,d,acc],dA,neq,B,nonpole] where [pol,s,d,acc] is related to s with:
# s = an irregular singularity of L (pole of f)
# pol = polynomial over K with zero s.
# d = truncated series of g=f^2 at an irregular singularity s of L.
# acc = smallest integer greater than or equal to m_p/l.
# dA = degree bound for the numerator of f^2   (A = numer(f^2)).
# neq = number of linear equations for the coefficients of A = numer(f^2).
# B = denominator of f^2.
# nonpole = denomintor of f but with integer power (the smallest integer greater than or equal to m_p/l).
# Here n_e=1 and l=2.
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# A = irrsingBesSqRoot(L,t,H,ext) or IrrRegAppsingBesSqRoot(L,t,H,ext)[1];
# G = regsingtrueBesSqRoot(L,t,A[-1],ext) or IrrRegAppsingBesSqRoot(L,t,H,ext)[2];
#
# We want to use here Delta1. Since the generalized exponents are not orderned then we will take Delta1 to be
# the first, after the second and after the third exponent differences.
SirrBesSqRootinfo1:=proc(L,A,G,x,t,ext) 
          local Dx,m,a1,a2,b1,b2,b3,R,e1,e2,e3,acc2,acc1,j,s,B,d,i,e,acc,a,b,l1,k,d1,v,v1,E,neq,dA,nonpole,
                extp1,extp2,extp3,extp4,extp5,extp6,extp0,extp7,H1;
          Dx,x:=op(_Envdiffopdomain);
          # We find the coefficients e1 and e2 of (x-p)^(1/2) in the two solutions                
          # (with t the unknown) of b*t^2 = x-p. Those coefficients are the two first             
          # solutions of the equation x^2-1/b = 0. Here b=1. So the equation is now 
          # x^2-1=0. So e1,e2 belong to {-1,1}. That means e1-e2=-+2.
          E:= [];
          nonpole:= 1;
          if G=[] then
             neq:= 0
          else
             neq := add(`if`(s[1]=infinity,1,degree(s[1],x)), s=G[1]);
          fi; 
          B:=1;
          dA:=0;
          j:=0;
          for s in A[1] do
             j:=j+1;
             m:= 2*A[4][j];
             if type(m,integer)=false then 
                return NULL
             fi; 
             if s[1] <> infinity then
                B:= B*(s[1])^m;
             else
                dA := m;
             fi;
             acc:= m; # degree minimal, en valeur absolue, du polynome de 
                      # la partie polaire de f^2 en p.
             acc2:= ceil(A[4][j]); # nombre d'equations venant de p pour les                                               
                                   # coefficients de A=(f^2)*B. Dans
                                   # les normes, la formule est: 
                                   # ceil((1/n_e)*m_j)=ceil(A[4][j]). Here n_e = 2.
             acc1:= floor(A[4][j]); # nombre de termes dans la partie polaire de f^2                                        
                                    # qu'on n'a pas. Dans les normes, la formule est:                                       
                                    # floor(((n_e-1)/n_e)*m_j)= floor((n_e-1)*A[4][j]).
                                    # Here n_e = 2.
             neq:= neq + acc2*op(`if`(s[1]=infinity,1,degree(s[1],x)));
             d:= convert(series(A[3][j][1],t=0,100),polynom);
             if type(d,`+`) then
                d1:= [op(d)] 
             else
                d1:=[d] 
             fi;
             # We remove the constant term in d
             e:=[]; 
             for i in d1 do
                if member(t,indets(i)) then 
                   e:=[op(e),i]
                fi;
             od; 
             l1:=map(s->subs(t=t^2,s),e);
             l1:=map(s->series(subs(t=1/x,s),x=0,100),l1);
             l1:=map(s->eval(subs(csgn=1,simplify(s) assuming(x>0))),l1);
             d1:=[e,l1];
             d:= add(d1[1][i]/(ldegree(convert(d1[2][i],polynom), x)), i=1..nops(e));
             d:=evala(expand(d^2)); 
             d:=convert(series(d,t=0,100),polynom);  
             if type(d,`+`) then
                d1:= [op(d)] 
             else 
                d1:=[d] 
             fi;
             d:=0;
             for i in d1 do
                l1:= degpol(i,x,t,ext);
                if l1<>false and l1<=m then 
                   d:=d+i
                fi; 
             od;
             if member(t,indets(d,{name}))=false then
                return NULL
             fi;
             if s[1]=infinity then
                b:=subs(t=1/x,d);
                a1:=convert(series(numer(b),x=0,100),polynom);
                a2:=expand(convert(series(denom(b),x=0,100),polynom));
                b:= factor(a1/a2);
                b:=convert(series(b,x=0,100),polynom);
                b:= eval(subs(csgn=1,simplify(b) assuming(x>0)));
                extp1:= indets(A[3][j][1],{RootOf,name}) minus {t};
                # All the names or RootOf different from RootOf(s,x) in the expression 
                # of A[3][j][1].
                extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All  
                extp2:=indets(ext,{RootOf,name});
                extp3:= extp1 minus extp2; # New names or Rootof which appear. 
                extp5:= extp1 union ext;
                E:=[op(E),[s[2],1/x,[convert(evala(b),polynom)],acc,extp3,extp5]];
                nonpole:= nonpole*(1)^acc1; # we take for infinity [1/x,pol]=[1,1]. 
             else
                b:=`if`(degree(s[1],x) > 1,subs(t=x-s[2],d),subs(t=s[1],d));
                a1:=convert(series(numer(b),x=s[2],100),polynom);
                a2:=expand(convert(series(denom(b),x=s[2],100),polynom));
                b:= factor(a1/a2);
                b:= convert(series(b,x=s[2],100),polynom);
                b:= eval(subs(csgn=1,simplify(b) assuming(x>0)));
                extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                extp1:= indets(A[3][j][1],{RootOf,name}) minus {t};
                extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                extp2:=indets({op(ext),ReplirrRoot(s[2],{})},{RootOf,name});
                extp3:= extp1 minus extp2; # New names or Rootof which appear.
                extp4:= extp1 minus extp0; # All the names or RootOf 
                                           # different from RootOf(s,x) in the                                                     
                                           # expression of A[3][j][1].
                extp6:= ext minus extp0;
                extp7:= extp4 union extp6;
                v1:={};
                for v in extp7 do
                   if extp0 intersect indets(v,{RootOf,name}) <> {} then
                      v1:= v1 union {v}
                   fi;
                od:
                extp5:= extp7 minus v1;
                E:=[op(E),[s[2],s[1],[convert(evala(b),polynom)],acc,extp3,extp5]];
                nonpole:= nonpole*(s[1])^acc1;
             fi;
          od;   
          dA:= degree(B,x)+dA;
          H1:= table();
          H1[1]:=E;
          H1[2]:=dA;
          H1[3]:=neq;
          H1[4]:=B;
          H1[5]:=nonpole;
          H1:= convert(H1,'list');
          return H1
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate v for the Bessel parameter must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = ZeroesmultordpolBesSqRoot(f,x);
# n = nops(B1);
# RS = newregsingBesSqRoot(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# v = candidate for the Bessel parameter.
CondinuexpBesSqRoot:= proc(E1,E2,m,v,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(-m*v-E2[1]),integer) and type(evala(m*v-E2[2]),integer) and
                 type(evala(2*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]+m*v),integer) and
                 type(evala(E1[2][1]-m*v),integer)) then
                 return true
              elif (type(evala(m*v-E2[1]),integer) and type(evala(-m*v-E2[2]),integer) and
                 type(evala(2*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][1]-E1[3][2]),integer) and type(evala(E1[1][1]-m*v),integer) and
                 type(evala(E1[2][1]+m*v),integer)) then
                 return true
              elif (type(evala(m*v-E2[1]),integer) and type(evala(-m*v-E2[2]),integer) and
                 type(evala(-2*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*v),integer) and
                 type(evala(E1[2][1]+m*v),integer)) then
                 return true
              elif (type(evala(-m*v-E2[1]),integer) and type(evala(m*v-E2[2]),integer) and
                 type(evala(-2*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][1]-E1[3][2]),integer) and type(evala(E1[1][1]+m*v),integer) and
                 type(evala(E1[2][1]-m*v),integer)) then
                 return true
              elif (type(evala(m*v-E2[1]),integer) and type(evala(m*v-E2[2]),integer) and
                 type(evala(2*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]+m*v),integer) and
                 type(evala(E1[2][1]-m*v),integer)) then
                 return true
              elif (type(evala(m*v-E2[1]),integer) and type(evala(m*v-E2[2]),integer) and
                 type(evala(2*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][1]-m*v),integer) and
                 type(evala(E1[2][2]+m*v),integer)) then
                 return true
              elif (type(evala(m*v-E2[1]),integer) and type(evala(-m*v-E2[2]),integer) and
                 type(evala(2*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][2]+m*v),integer) and
                 type(evala(E1[2][2]-m*v),integer)) then
                 return true
              elif (type(evala(-m*v-E2[1]),integer) and type(evala(m*v-E2[2]),integer) and
                 type(evala(2*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][2]-m*v),integer) and
                 type(evala(E1[2][2]+m*v),integer)) then
                 return true
              elif (type(evala(-m*v-E2[1]),integer) and type(evala(-m*v-E2[2]),integer) and
                 type(evala(-2*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*v),integer) and
                 type(evala(E1[2][1]+m*v),integer)) then
                 return true
              elif (type(evala(-m*v-E2[1]),integer) and type(evala(-m*v-E2[2]),integer) and
                 type(evala(-2*m*v-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][1]+m*v),integer) and
                 type(evala(E1[2][2]-m*v),integer)) then
                 return true
              elif (type(evala(-m*v-E2[1]),integer) and type(evala(m*v-E2[2]),integer) and
                 type(evala(-2*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][2]),integer) and
                 type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][2]-m*v),integer) and
                 type(evala(E1[2][2]+m*v),integer)) then
                 return true
              elif (type(evala(m*v-E2[1]),integer) and type(evala(-m*v-E2[2]),integer) and
                 type(evala(-2*m*v-E2[3]),integer)) and (type(evala(E1[1][2]-E1[3][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][2]+m*v),integer) and
                 type(evala(E1[2][2]-m*v),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
ZeroesmultordpolBesSqRoot:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsingBesSqRoot(L,t,H,ext)[2];
# B = ZeroesmultordpolBesSqRoot(f,x);
# n = nops(B)
newregsingBesSqRoot:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],CondiBesSqRootreg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],CondiBesSqRootreg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g.
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Bespossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:=table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     if member(-h,{op(C)})=false and member(h,{op(C)})=false then
                           C[k]:= h;
                     fi;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

simplef := proc(flist)
           #option trace;
           local f,res;
           res := {};
           for f in flist do
              f := evala(Normal(f, expanded));
              f := evala(Sqrfree(f,x));
              f := f[1]*mul(i[1]^i[2],i=f[2]);
              res := res union {f};
           od;
           res;
end:

############################################################################################################
############################################################################################################
#################################################   EASY CASE   ############################################
############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],x,ext).
easyBesSqRootf := proc(L,R,info,x,t,ext)
                  local Dx,a,f,fn,p,i,j,linearr,temp,d,d1,nonpole,und,B1,n,SR,g,
                        g1,A1,A2,G,H1,H,c1,c2,c3,C;
                  Dx,x:=op(_Envdiffopdomain); 
                  ####### info[4] gives us the denominator of f.  
                  ####### degree(A,x)=info[2].
                  H:=table();
                  f := add(a[i]*x^i, i = 0 .. info[2])/info[4];
                  und := indets(f) minus {x} ;
                  C:=map(s->s[1],R[2][1]);       ####### In easy case, Sreg<>[]
                  if has({op(C)}, infinity) then 
                     f:=subs(a[info[2]]=0,f);
                     und:=subs(a[info[2]]=NULL,und);
                  fi;
                  fn := numer(f); 
                  for p in R[2][1] do 
                     if p[1]=infinity then
                        p[1]:= 1/x;
                     fi;
                     fn := fn/(p[1]); 
                  od;
                  linearr:=convert(collect(eval(subs(csgn=1,simplify(evala(Rem(numer(fn),denom(fn),x))) assuming(x>0))),x),polynom);
                  linearr := coeffs(linearr, x); 
                  temp := SolveTools:-Linear({linearr},und); 
                  f := subs(op(temp), f);
                  H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
                  j := 0; # use j indicate that if infinity is an irregular singularity
                  for p in info[1] do 
                     if p[1] = infinity then 
                        j := ceil(p[4]/2);
                        break;
                     fi;
                  od;
                  G := Bespossibility(H1);
                  for d in G do
                     c1:= convert(series(numer(f-d),x=0,100),polynom);
                     c2:= convert(series(denom(f-d),x=0,100),polynom);
                     c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                     nonpole := Normalizer(c3*info[5]);
                     A1:=convert(series(numer(nonpole),x=0,100),polynom);
                     A2:=convert(series(denom(nonpole),x=0,100),polynom);
                     temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                     temp := collect(temp, x, Normalizer);
                     linearr := coeffs(temp, x);
                     if j<>0 then # the equations at infinity
                        temp := evala(Quo(A1, A2,x));
                        temp:=convert(series(numer(temp),x=0,100),polynom);
                        temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                        temp := collect(temp, x, Normalizer);
                        d1:=degree(temp, x); 
                        temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1); 
                        linearr := linearr, temp;
                     fi;
                     temp := SolveTools:-Linear({linearr},und);
                     if temp=NULL then
                        next
                     fi;
                     g := subs(temp, f);
                     g := factor(op(simplef({g})));
                     if member(-g,convert(H,'set'))=false then
                        B1:= ZeroesmultordpolBesSqRoot(g,x);
                        if B1=NULL then
                           next
                        fi;
                        n:= nops(B1);
                        SR:= newregsingBesSqRoot(L,t,x,B1,R[2],n,ext);
                        H[d]:= [g,SR[2],SR[1]];
                     fi;
                  od;
                  H:= convert(H,'set');
                  if H = {} then
                     return NULL
                  else
                     return H
                 fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],x,ext);
# f = element of easyBesSqRootf(L,R,info,x,t,ext).
findBesSqRootnueasyIrr:= proc(L,f,x,t,ext)
                local Dx,i,vi,V,n,j,s;
                Dx,x:=op(_Envdiffopdomain);
                n:= nops(f[3][1]);
                for i from 1 to n do
                   vi := [seq(f[3][3][i][1]+j,j=0..floor(f[2][i]))]/(f[2][i]);
                   vi:=[seq(`if`(MatchInt([0,1/2,op(vi[1..j-1])],vi[j]), NULL, vi[j]),j=1..nops(vi))];
                   if vi=[] then
                      return NULL
                   fi;
                   vi:= map(s-> `if`(CondinuexpBesSqRoot(f[3][4][i],f[3][3][i],f[2][i],s,ext)=false,NULL,s), vi);
                   if vi=[] then
                      return NULL
                   fi;
                   vi:=map(s-> ReplRoot2(s,{}),vi);
                   if i = 1 then
                      V:=vi
                   else
                      V:=[seq(`if`(MatchInt1(V,vi[i])<>false, MatchInt1(V,vi[i]), NULL),i=1..nops(vi))];
                   fi;
                   if V=[] then
                      return NULL
                   fi
                od;
                if V=[] then
                   return NULL
                else
                   return [V,f[1]]
                fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],x,ext);
# dA = degree(A,x) = info[2]
# RSreg = R[3][1].
helpBesSqRootnueasyRat:= proc(Sreg,dA,x)
                local Dx,i,dv,dnu,d1,d2,s;
                Dx,x:=op(_Envdiffopdomain);
                dnu := {};
                if Sreg=[] then
                   for i from 3 to floor(dA/2) do
                      if irem(dA,i)=0 then
                         dnu:= dnu union {i}
                      fi;
                   od;
                   dnu := dnu union {dA};
                else
                   dv := denom(Sreg[3][1][1][1]);
                   for i from 1 to nops(Sreg[1]) do
                      dv:=lcm(dv,denom(Sreg[3][i][1][1]));
                   od;
                   if dv >= dA then
                      for i from 3 to dA*dv do
                         if irem(i,dv)=0 then
                            dnu := dnu union {i};
                         fi;
                      od;
                   else
                      for i from 1 to iquo(dA,dv) do
                         d1:=0;
                         d2:=0;
                         for s from 1 to nops(Sreg[1]) do
                            d1:=d1+i*dv/denom(Sreg[3][s][1][1])*(`if`(Sreg[1][s][1]=infinity,1,
                                                                degree(Sreg[1][s][1],x)));
                            d2 := gcd(d2, i*dv/denom(Sreg[3][s][1][1]));
                         od;
                         if d1<=dA and irem(dA,d2)=0 then
                            dnu := dnu union {i*dv};
                         fi;
                      od;
                   fi;
                fi;
                if dnu={} then
                   return NULL
                else
                   return dnu
                fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],x,ext);
# f = element of easyBesSqRootf(L,R,info,x,t,ext);
# d = candidate for the denominator of nu.
findBesSqRootnueasyRat:= proc(L,f,d,x,t,ext,n)
                    local Dx,v,i,k,V,j,s;
                    Dx,x:=op(_Envdiffopdomain);
                    V:= table();
                    for k from 1 to floor(d/2) do
                       if gcd(k,d)=1 then
                          v := k/d;
                          j:=0;
                          for i from 1 to n do
                             if CondinuexpBesSqRoot(f[3][4][i],f[3][3][i],f[2][i],v,ext)=false and
                                CondinuexpBesSqRoot(f[3][4][i],f[3][3][i],f[2][i],-v,ext)=false then
                                break
                             fi;
                             j:=j+1;
                          od;
                          if j=n then
                             V[k]:= v
                          fi;
                       fi;
                    od;
                    V:= convert(V,'list');
                    if V <> [] then
                       return [seq(`if`(MatchInt([0,1/2,op(V[1..s-1])],V[s]), NULL, V[s]),s=1..nops(V))];
                    else
                       return NULL
                    fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],x,ext).
easyBesSqRoot:= proc(L,R,info,x,t,ext)
                local Dx,H,n1,B,f,ext1,ext2,dnu,n,V,V1,i,s;
                Dx,x:=op(_Envdiffopdomain);
                B:= easyBesSqRootf(L,R,info,x,t,ext);
                if B = NULL then
                   return NULL
                fi;
                H:= table();
                if R[3][3] <> [] then
                   for f in B do
                      H[f]:= findnuLogBesSqRoot(L,f[1],f[3],x,t,ext);
                   od;
                else
                   ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                   ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                   n1:=nops(R[2][1]);
                   if n1 > 1 then
                      for i from 2 to n1 do
                         ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                         ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                         ext1:= ext1 intersect ext2;
                      od;
                   fi;
                   if ext1<>{} then
                      for f in B do
                         H[f]:= findBesSqRootnueasyIrr(L,f,x,t,ext)
                      od;
                   else
                      dnu:= helpBesSqRootnueasyRat(R[3][1],info[2],x);
                      for f in B do
                         n:=nops(f[3][1]);
                         V:= table();
                         for i in dnu do
                            V1:=findBesSqRootnueasyRat(L,f,i,x,t,ext,n);
                            if V1<>NULL then
                               V[i]:= op(findBesSqRootnueasyRat(L,f,i,x,t,ext,n));
                            fi;
                         od;
                         V:= convert(V,'set');
                         if V<>{} then
                            H[f]:=[[op(V)],f[1]]
                         fi;
                      od;
                   fi;
                fi;
                H:= convert(H,'set');
                if H = {} then
                   return NULL
                else
                   return H
                fi;
end:

############################################################################################################
############################################################################################################
############################################   LOGARITHMIC CASE   ##########################################
############################################################################################################
############################################################################################################

# g = f^2 = (C.A1.A2^b)/B. We want to find candidates for A1. Sreg=regsingtrue(L,t,ext)[1] is the set of the
# zeros of A1.
# input: H and d, where H=subs(Sreg[4]=NULL,Sreg) and d is the degree of A1. We do a combinatorial distribution of d to # the elements of H[1].
# return: candidates for A1 = \prod_{p\in Sreg[1]}(p^{k_p}) where
# p = an element of Sreg[1] (minimal polynomial)
# d = \sum_{p\in Sreg[1]}(k_p*deg(p)) = degree of A1.
#
# E = Singular(L,extpp);
# F = NotAppSing(L,E,ext)
# Sirr = irrsingBesSqRoot(L,t,G,ext) or IrrRegAppsingBesSqRoot(L,t,E,ext)[1];
# C = regsingtrueBesSqRoot(L,t,Sirr[-1],ext);
# H = IrrRegAppsing0F2(L,t,G,ext)[2];
searchKnLog :=proc(H,d)
              local res,temp,s,deg,t,k;
              res:=NULL;
              if nops(H[1])=1 then
                 deg:=`if`(H[1][1][1]=infinity,1,degree(H[1][1][1],x));
                 if d mod deg = 0 then
                    if H[1][1][1]= infinity then
                       res:=res,1;
                    else
                       res:= res, H[1][1][1]^(d/degree(H[1][1][1],x));
                    fi
                 fi;
              else
                 deg := `if`(H[1][1][1]=infinity,1,degree(H[1][1][1],x));
                 for k from 1 to iquo(d,deg) do
                    temp := searchKnLog([seq([seq(H[i][j],j=2..nops(H[i]))],i=1..nops(H))],d-k*deg);
                    if temp<>NULL then
                       if H[1][1][1] <> infinity then
                          for t in temp do
                             res:= res,H[1][1][1]^k*t;
                          od;
                       else
                          for t in temp do
                             res:=res,t;
                          od;
                       fi;
                    fi;
                 od;
              fi;
              return [res];
end:

############################################################################################################
############################################################################################################

# We compute the Bessel square parameter nu related to a candidate of f^2.
#
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# Sreg = R[2];
# g = candidate for the change of variable parameter (f).
findnuLogBesSqRoot := proc(L,g,Sreg,x,t,ext)
             local Dx,i,v,n,m,a;
             m:= nops(Sreg[1]);
             n := max(degree(numer(g),x),degree(denom(g),x));
             ####### Here all the exponent differences at the zeroes of g are integers
             a:=0;
             for i from 1 to m do
                if Sreg[1][i][1]=infinity then
                   a:= a+abs(evala(ReplirrRoot(Sreg[3][i][1],{})))
                else
                   a:= a+abs(evala(ReplirrRoot(Sreg[3][i][1],{}))*degree(Sreg[1][i][1],x))
                fi;
             od;
             v:= floor(a/n);
             return {[[v],g]}
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],x,ext).
findBesSqRootln:= proc(L,R,info,x,t,T,ext)
                 local Dx,f,p,res,knl,kn,g,i,j,linearr,temp,d,d1,nonpole,B,dA,k,e,G,A1,A2,c1,c2,c3,h,H1,H;
                 Dx,x:=op(_Envdiffopdomain);
                 res:= NULL;
                 dA:= info[2];
                 B:= info[4];
                 knl := searchKnLog(R[2],dA);
                 f := {};
                 H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
                 j := 0; # use j indicate that if infinity is irregularsingularity
                 k:=0;
                 for p in info[1] do
                    k:=k+1;
                    if p[1] = infinity then
                       j := ceil(info[1][k][4]/2);
                       break;
                    fi;
                 od;
                 G := Bespossibility(H1);
                 for kn in knl do
                    for d in G do
                       c1:= convert(series(numer(c*kn/B-d),x=0,100),polynom);
                       c2:= convert(series(denom(c*kn/B-d),x=0,100),polynom);
                       c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                       nonpole := Normalizer(c3*info[5]);
                       A1:=convert(series(numer(nonpole),x=0,100),polynom);
                       A2:=convert(series(denom(nonpole),x=0,100),polynom);
                       temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                       temp := collect(temp, x, Normalizer);
                       linearr := coeffs(temp, x);
                       if j<>0 then # the equations at infinity
                          temp := evala(Quo(A1, A2,x));
                          temp:=convert(series(numer(temp),x=0,100),polynom);
                          temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                          temp := collect(temp, x, Normalizer);
                          d1:=degree(temp, x);
                          temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                          linearr := linearr, temp;
                       fi;
                       temp := SolveTools:-Linear({linearr},{c});
                       if temp=NULL then
                          next
                       fi;
                       temp := op(temp);
                       h:= subs(temp, c*kn/B);
                       if member(-h,f)=false then
                          f := f union {h};
                       fi;
                    od;
                 od;
                 if f<>{} then
                    f := simplef(f);
                 fi;
                 for g in f do
                    res:= res union findnuLogBesSqRoot(L,factor(g),R[2],x,t,ext);
                 od;
                 return res
end:

############################################################################################################
############################################################################################################
##############################################   IRRATIONAL CASE   ##########################################
############################################################################################################
############################################################################################################

leadcoeffRootnameinter:=proc(A,x,ext1)
                    local Dx,l,l1,ira,ira1,ira2,sum1,sum2;
                    Dx,x:=op(_Envdiffopdomain);
                    l:=A;
                    l1:=ext1;
                    while l1<>{} and (l1 intersect ext1 <> {}) do
                       ira1:=ChooseRootOflistinter(l,l1)[1];
                       if ira1<>{} then
                          ira:=ira1[1];
                       else
                          ira2:= l1 intersect ext1:
                          ira:=ira2[1];
                       fi;
                       sum1:=convert(collect(l,ira,evala),polynom);
                       sum2:= `if`(ldegree(sum1,ira)<0,convert(subs(ira=1/ira,sum1),polynom),sum1);
                       l:= lcoeff(sum2,ira);
                       l1:= indets(l,{RootOf,name});
                    od;
                    return l
end:

############################################################################################################
############################################################################################################

#
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,G,ext);
# info = SirrBesSqRootinfo1(L,R[1],R[2],x,ext)
findBesSqRootIrr := proc(L,R,info,x,t,T,ext)
             local Dx,f,vf,ira,dA,B,ext1,ext2,A,m,i,r,sum,a,j,h,H1,s,p,G,d,A1,A2,d1,c1,c2,c3,c,linearr,temp,
                   nonpole,n,b1,v,V,g,ira1,ira2,H,vi,M,l1;
             Dx,x:=op(_Envdiffopdomain);
             f := {};
             M:=[];
             dA:= info[2];
             B:= info[4];
             n:=nops(R[2][1]);
             ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
             ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
             if n > 1 then
                for i from 2 to n do
                   ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                   ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                   ext1:= ext1 intersect ext2;
                od;
             fi;
             ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
             ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
             ####### irrational case.
             r[1]:=1;
             for i from 1 to n do
                H[i]:= ReplirrRoot(R[2][3][i][1],{});
                l1:= leadcoeffRootnameinter(H[i],x,ext1);
                if type(l1,rational)=false or l1=0 then
                   return NULL
                fi;
                a[i]:= l1*(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));
                if i>1 then
                   r[i] := Normalizer(abs(a[i]/a[1])); #   r[i]=(m_i*degree(p_i,x))/(m_1*degree(p_1,x))
                   if not type(r[i],rational) then
                      return NULL
                   fi;
                fi;
             od;
             sum:=add(r[i],i=1..n);
             b1:=dA/sum; # b1 = m_1*degree(p_1,x)
             if type(b1, integer) then
                A:=c;
                for i from 1 to n do
                   m[i]:= (r[i]*b1)/(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));
                   A:=A*(`if`(R[2][1][i][1]=infinity,1,(R[2][1][i][1])^(m[i])));
                   M:=[op(M),m[i]];
                od;
             else
                return NULL;
             fi;
             H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
             j := 0; # use j indicate that if infinity is irregular singularity
             for p in info[1] do
                if p[1] = infinity then
                   j := ceil(p[4]/2);
                   break;
                fi;
             od;
             G := Bespossibility(H1);
             for d in G do
                c1:= convert(series(numer(A/B-d),x=0,100),polynom);
                c2:= convert(series(denom(A/B-d),x=0,100),polynom);
                c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                nonpole := Normalizer(c3*info[5]);
                A1:=convert(series(numer(nonpole),x=0,100),polynom);
                A2:=convert(series(denom(nonpole),x=0,100),polynom);
                temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                temp := collect(temp, x, Normalizer);
                linearr := coeffs(temp, x);
                if j<>0 then # the equations at infinity
                   temp := evala(Quo(A1, A2,x));
                   temp:=convert(series(numer(temp),x=0,100),polynom);
                   temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                   temp := collect(temp, x, Normalizer);
                   d1:=degree(temp, x);
                   temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                   linearr := linearr, temp;
                fi;
                temp := SolveTools:-Linear({linearr},{c});
                if temp=NULL then
                   next
                fi;
                temp := op(temp);
                h:= subs(temp, A/B);
                if member(-h,f)=false then
                   f := f union {h};
                fi;
             od;
             if f <> {} then
                f := simplef(f);
                for i from 1 to n do
                   vi := [seq(H[i]+j,j=0..floor(m[i]))]/(m[i]);   # Delta = m_j*nu + k with k an integer.
                   vi:=[seq(`if`(MatchInt([0,1/2,op(vi[1..j-1])],vi[j]), NULL, vi[j]),j=1..nops(vi))];
                   if vi=[] then
                      return NULL
                   fi;
                   vi:= map(s-> `if`(CondinuexpBesSqRoot(R[2][4][i],R[2][3][i],m[i],s,ext)=false,NULL,s), vi);
                   if vi=[] then
                      return NULL
                   fi;
                   vi:=map(s-> ReplRoot2(s,{}),vi);
                   if i = 1 then
                      V:=vi
                   else
                      V:=[seq(`if`(MatchInt1(V,vi[i])<>false, MatchInt1(V,vi[i]), NULL),i=1..nops(vi))];
                   fi;
                   if V=[] then
                      return NULL
                   fi
                od;
                vf:=table();
                for g in f do
                   vf[g]:= [V,factor(g)];
                od;
                vf:= convert(vf,'set');
                return vf
             else
                return NULL
             fi;
end:

############################################################################################################
############################################################################################################
##########################################   RATIONAL CASE   ###############################################
############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute all the candidates for A1 associated with
# the degree of A2. The output is a list with element in the form [s,k] where s=A1 and k=degree(A2).
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# G = R[3][1];
# dr = degree(A1)
# k = degree(A2)
# dA = degree(cA1.A2^d) = dr+k*d = degree of the numerator of f^2.
# s = A1.
searchBesSqRootA1k:=proc(G,dA,d,x)
            local Dx,s,dr,k,res,temp;
            Dx,x:=op(_Envdiffopdomain);
            res :=NULL;
            if G = [] then
              if irem(dA,d)<>0 then
                 return {};
              else
                 return {[1,dA/d]};
              fi;
           fi;
           for k from 0 to iquo(dA,d) do
              dr :=dA-k*d;
              temp :=searchBesSqRootA1(G,dr,d,x);
              for s in temp do
                 res := res, [s,k]
              od;
           od;
           {res};
end:

############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute here all the candidates for A1 having the same
# degree.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# G = R[3][1]
# de = degree(A1)
# k = degree(A2)
# dA = degree(cA1.A2^d) = de+k*d = degree of the numerator of f^2.
searchBesSqRootA1 :=proc(G,de,d,x)
           local res,d1,i,temp,k,t,E,n,j,s,m;
           res :=NULL;
           if nops(G[1])=0  then
              if de=0 then
                 return [1];
              else
                 return NULL;
              fi;
           fi;
           s :=G[1][1][1];
           d1:=max(denom(G[3][1][1]));
           i :=abs(d/d1);
           if type(i,integer)=false then
              return NULL
           fi;
           if (de-i*(`if`(s=infinity,1,degree(s,x))))<0 then
              return NULL;
           fi;
           E:= [seq([seq(G[n][j],j=2..nops(G[n]))],n=1..nops(G))];
           m:=add(`if`(j[1]=infinity,1,degree(j[1],x)),j=E[1]);
           for k from 1 to iquo(de-m,i*(`if`(s=infinity,1,degree(s,x)))) do
              if i*k>=d then
                 break
              fi;
              temp := searchBesSqRootA1(E,de-k*i*(`if`(s=infinity,1,degree(s,x))),d,x);
              if temp <> NULL then
                 for t in temp do
                    res := res,(`if`(s=infinity,1,s))^(i*k)*t;
                 od;
              fi;
           od;
           [res];
end:

############################################################################################################
############################################################################################################

poly_to_powerseries := proc(P, x, p, T, acc)
                       local d, l, i, S,a;
                       a:=convert(P,polynom);
                       if a=0 then
                          # error "only treat non-zero powseries"
                          return NULL
                       fi;
                       if p = infinity then
                          d, l:= degree(a,x), lcoeff(a,x);
                          [l, -d,add(evala(coeff(a,x,d-i)/l)*T^i, i=0..acc-1), acc]
                       else
                          S := collect(eval(a, x = T + p), T, evala);
                          d, l := ldegree(S,T), tcoeff(S,T);
                          # Note: the evala in the line S := .. is needed to ensure that d is correct.
                          [l, d,add(evala(coeff(S,T,d+i)/l)*T^i, i=0..acc-1), acc]
                       fi;
end:

############################################################################################################
############################################################################################################

pow_reciprocal := proc(L1,tp)
                  [Normalizer(1/L1[1]),-L1[2],convert(series(1/L1[3],tp=0,L1[4]),polynom),L1[4]]
end:

############################################################################################################
############################################################################################################

pow_time := proc(L1,L2,tp)
            local acc,f,ind;
            acc := min(L1[4],L2[4]);
            f := expand(L1[3]*L2[3]);
            ind := [op(indets(f))];
            if nops(ind) = 2 then
               f := subs(ind[1] = tp, f);
               f := subs(ind[2] = tp, f)
            fi;
            [evala(L1[1]*L2[1]),L1[2]+L2[2],convert(series(f,tp=0,acc),polynom),acc]
end:

############################################################################################################
############################################################################################################

dthroot_powseries := proc(L, tp, d, ext)
                     local X,v,res,i,s,l;
                     l:=L[1];
                     try
                        v:=evala(Factors(X^d -l,ext));
                     catch :
                        v:=evala(Factors(numer(normal(X^d - l)),{op(ext)}))
                     end try;
                     v:=select(has,v[2],X);
                     res := NULL;
                     for i in v do
                        if degree(i[1],X) = 1 then
                           res := res, -coeff(i[1],X,0)/coeff(i[1],X,1)
                        fi;
                     od;
                     v := series(L[3]^(1/d), tp=0, L[4]);
                     v := collect(convert(v,polynom),tp,evala);
                     res :={seq([i, L[2]/d, v, L[4]],  i = [res])};
                     res;
end:

############################################################################################################
############################################################################################################

simplef := proc(flist)
           local f,res;
           res := {};
           for f in flist do
              f := evala(Normal(f, expanded));
              f := evala(Sqrfree(f,x));
              f := f[1]*mul(i[1]^i[2],i=f[2]);
              res := res union {f};
           od;
           res;
end:

############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute here all the candidates for f^2.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
#
## ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# dA = degree(A);
# d = denom(nu);
# knl = searchBesSqRootA1k(R[3][1],dA,d,x);
# info = SirrBesSqRootinfo1(L,R[1],R[2],x,ext);
# B = info[4];
# kn element of knl.
helpBesSqRootfRat1 := proc(info,B,d,kn,x,t,T,ext)
                      local Dx,a,l,acc,ep, Bs, temp, ser,lp, uk, und,i, j, lr,f, extp,ukn,uki,p;
                      Dx,x:=op(_Envdiffopdomain);
                      for i from 1 to nops(info[1]) do
                         acc := info[1][i][4];
                         p:= info[1][i][1];
                         ep := poly_to_powerseries(kn[1],x,p,T,acc);
                         if ep = NULL then
                            return NULL
                         fi;
                         ep := pow_reciprocal(ep,T);
                         Bs := poly_to_powerseries(B,x,p,T,acc);
                         if Bs = NULL then
                            return NULL
                         fi;
                         temp := pow_time(ep, Bs, T);
                         ep := poly_to_powerseries(info[1][i][3][1],x,p,T,acc);
                         if ep = NULL then
                            return NULL
                         fi;
                         ser[i] := pow_time(temp, ep,T);
                         lp := ser[i][1];
                         und := indets(lp,RootOf) minus ext;
                         if und={} then
                            if not assigned(l) then
                               l:=lp;
                            else
                               if length(lp)<length(l) then
                                  l := lp;
                               fi;
                            fi;
                         fi;
                      od;
                      if not assigned(l) then
                         return {};
                      fi;
                      uk := add(a[i]*x^i,i=0..kn[2]);
                      und := indets(uk) minus {x};
                      for i from 1 to nops(info[1]) do
                         acc := info[1][i][4];
                         p:= info[1][i][1];
                         temp := ser[i];
                         if temp[2] <> 0 and p <> infinity then
                            error "we made an error"
                         fi;
                         temp[1] := temp[1]/l;
                         extp := indets(p,RootOf) union ext;
                         ep := dthroot_powseries(temp,T,d,extp);
                         if ep={} then
                            next
                         fi;
                         ukn := NULL;
                         for uki in {uk} do
                            temp := poly_to_powerseries(uki,x,p,T,acc);
                            if temp = NULL then
                               next
                            fi;
                            for j in ep do
                               lr := temp[3]-j[3];
                               lr := coeffs(lr,T);
                               lr := {lr,temp[1]-j[1]};
                               lr := map(numer,lr);
                               if type(p,RootOf) and not member(p,ext) then
                                  lr := map(evala@Expand, lr);
                                  lr := map(coeffs, lr,p)
                               fi;
                               lr := map(s-> `if`(indets(s) intersect und={},NULL,s),lr);
                               lr := SolveTools:-Linear(lr,und);
                               if lr <> NULL then
                                  f:= subs(op(lr),uki);
                                  ukn := ukn,f;
                               fi;
                            od;
                         od;
                         uk := ukn;
                         if uk = NULL then
                            break
                         fi;
                      od;
                      if uk <> NULL then
                         return {seq(evala(l*i^d*kn[1]/B),i=[uk])}
                      else
                         return NULL
                      fi;
end:

############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute here all the candidates for f^2.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# dA = degree(A)
# d = denom(nu)
# knl = searchBesSqRootA1k(R[3][1],dA,d,x)
# info = SirrBesSqRootinfo1(L,R[1],R[2],x,t,ext)
helpBesSqRootfRat2:= proc(L,info,d,knl,x,t,T,ext)
                local Dx,res,kn,uk,B,h;
                Dx,x:=op(_Envdiffopdomain);
                res := {};
                if knl <> {} then
                   B:= info[4];
                   for kn in knl do
                      uk:= helpBesSqRootfRat1(info,B,d,kn,x,t,T,ext);
                      if uk <> NULL then
                        for h in uk do
                            if member(-h,res)=false then
                               res := res union {h};
                            fi;
                         od;
                      fi;
                   od;
                fi;
                res;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# dA = degree(A)
# d = denom(nu)
# knl = searchBesSqRootA1k(R[3][1],dA,d)
# info = SirrBesSqRootinfo1(L,R[1],R[2],x,ext)
# B = info[4]
# g = candidate for f^2
# B1 = ZeroesmultordpolBesSqRoot(g,x);
# n1 = nops(B1);
# H = newregsingBesSqRoot(L,t,x,B1,R[2],n1,ext);
findnuBesSqRootRat:= proc(g,d,H,x,t,ext)
                  local Dx,V,j,v,k,i,n,s;
                  Dx,x:=op(_Envdiffopdomain);
                  V:= table();
                  n:= nops(H[1][1]);
                  for k from 1 to floor(d/2) do
                     if gcd(k,d)=1 then
                        v := k/d;
                        j:=0;
                        for i from 1 to n do
                           if CondinuexpBesSqRoot(H[1][4][i],H[1][3][i],H[2][i],v,ext)=false and
                              CondinuexpBesSqRoot(H[1][4][i],H[1][3][i],H[2][i],-v,ext)=false then
                              break
                           fi;
                           j:=j+1;
                        od;
                        if j=n then
                           V[k]:= v
                        fi;
                     fi;
                  od;
                  V:= convert(V,'list');
                  if V <> [] then
                     V:=[seq(`if`(MatchInt([0,1/2,op(V[1..s-1])],V[s]), NULL, V[s]),s=1..nops(V))];
                     return {[V,factor(g)]}
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingBesSqRoot(L,t,E,ext);
# info = SirrBesSqRootinfo1(L,R[1],R[2],x,ext).
findBesSqRootRat:= proc(L,R,info,x,t,T,ext)
           local  Dx,s,B1,n,RS,l,vf,dnu,dv,d1,d2,res,i,knl,j,k,f,ff,B,dA,Sreg;
           Dx,x:=op(_Envdiffopdomain);
           vf:=table();
           Sreg:=R[3][1];
           dA:= info[2];
           if Sreg=[] then
              dnu := {};
              for i from 3 to floor(dA/2) do
                 if irem(dA,i)=0 then
                    dnu:= dnu union {i}
                 fi;
              od;
              dnu := dnu union {dA};
           else
              dnu := {};
              dv := denom(Sreg[3][1][1][1]);
              for i from 1 to nops(Sreg[1]) do
                 dv:=lcm(dv,denom(Sreg[3][i][1][1]));
              od;
              if dv >= dA then
                 for i from 3 to dA*dv do
                    if irem(i,dv)=0 then
                       dnu := dnu union {i};
                    fi;
                 od;
              else
                 for i from 1 to iquo(dA,dv) do
                    d1:=0;
                    d2:=0;
                    for s from 1 to nops(Sreg[1]) do
                       d1:=d1+i*dv/denom(Sreg[3][s][1][1])*(`if`(Sreg[1][s][1]=infinity,1,
                                                                degree(Sreg[1][s][1],x)));
                       d2 := gcd(d2, i*dv/denom(Sreg[3][s][1][1]));
                    od;
                    if d1<=dA and irem(dA,d2)=0 then
                       dnu := dnu union {i*dv};
                    fi;
                 od;
              fi;
           fi;
           B:= info[4];
           for i in dnu do
              knl := searchBesSqRootA1k(Sreg,dA,i,x);
              if knl={} then
                 next
              fi;
              ff := helpBesSqRootfRat2(L,info,i,knl,x,t,T,ext);
              ff := simplef(ff);
              if assigned(_Env_original_field) then
                 f := {};
                 for k in ff do
                    if indets(k, RootOf) minus _Env_original_field ={} then
                       f:=f union {factor(k)};
                    fi;
                 od;
              else
                 f:= map(j->factor(j),ff);
              fi;
              res:={};
              for j in f do
                 B1:= ZeroesmultordpolBesSqRoot(j,x);
                 if B1=NULL then
                    next
                 fi;
                 n:= nops(B1);
                 RS:= newregsingBesSqRoot(L,t,x,B1,R[2],n,ext);
                 res := res union findnuBesSqRootRat(j,i,RS,x,t,ext);
              od;
              if res<>{} then
                 vf[i]:= op(res);
              fi;
           od;
           vf:= convert(vf,'set');
           if vf={} then
              return NULL
           else
              return vf
           fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature 1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1, l2, p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

#
# E = Singular(L,ext);
# A = NotAppSing(L,E,ext);
# r = irrsingBesSqRoot(L,t,A,ext);
# Sreg = regsingtrueBesSqRoot(L,t,r[-1],ext);
# RSreg = SregseptrueBesSqRoot(L,Sreg,ext);
# info = Sirr1F2info1(L,r,Sreg,x,ext).
helpBesSqRootSolutions1:= proc(L,R,info,x,t,T,ext)
           local Dx,n1,ext1,ext2,i,H,H1,s;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H1:= findBesSqRootRat(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easyBesSqRoot(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           elif R[3][3]<>[] then
              H1:= findBesSqRootln(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easyBesSqRoot(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           elif R[3][2]<>[] then
              H1:= findBesSqRootRat(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easyBesSqRoot(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           else
              ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
              ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
              n1:=nops(R[2][1]);
              if n1 > 1 then
                 for i from 2 to n1 do
                    ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                    ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                    ext1:= ext1 intersect ext2;
                 od;
              fi;
              if ext1 <> {} then
                 H1:= findBesSqRootIrr(L,R,info,x,t,T,ext);
                 if H1=NULL then
                    if info[3]>info[2] then
                       H:= easyBesSqRoot(L,R,info,x,t,ext);
                    else
                       return NULL
                    fi;
                 else
                    H:= H1
                 fi;
              else
                 H1:= findBesSqRootRat(L,R,info,x,t,T,ext);
                 if H1=NULL then
                    if info[3]>info[2] then
                       H:= easyBesSqRoot(L,R,info,x,t,ext);
                    else
                       return NULL
                    fi;
                 else
                    H:= H1
                 fi;
              fi;
           fi:
           if H=NULL then
              return NULL
           else
              H:= {op(map(s-> op([s,subs(s[-1]=-s[-1],s)]),H))};
              return [op(H)]
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsingBesSqRoot(L,t,G,ext);
# a = element of helpBesSqRootSolutions1(L,r,RSreg,info,ext,x);
helpBesSqRootSolutions2:= proc(L,a,A,x,t,ext)
              local Dx,LBB,LBB1,eq,Y,J,n1,M,L22,e1,R,i,T;
              Dx,x:=op(_Envdiffopdomain);
              n1:= nops(a[1]);
              eq:=HolonomicDE(BesselI(e1, x)^2,Y(x));
              LBB1:=de2diffop(eq, Y(x));
              for i from 1 to n1 do
                 LBB:=subs(e1=a[1][i],LBB1);
                 M:= ChangeOfVariables(LBB,sqrt(a[2]));
                 R:=EquivExpgaugeBesSqRoot(M,L,A,x,t,T,ext);
                 if R<>NULL then
                    return [{[a[1][i],R]},a[2]]
                 fi;
              od;
              return NULL
end:

############################################################################################################
############################################################################################################

helpBesSqRootSolutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

FindBesSqRootSolutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsingBesSqRoot(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= SirrBesSqRootinfo1(L,R[1],R[2],x,t,ext);
           if F=NULL then
              return NULL
           fi;
           C:= table();
           H:= helpBesSqRootSolutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= helpBesSqRootSolutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

BesSqRootSolutions:= proc(L)
           local Dx,x,s,k,F,E,H1,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:= table();
           H1:= FindBesSqRootSolutions(L);
           if H1=NULL then
              return "No Bessel square roots type solutions"
           fi;
           F:= map(s-> [op(s[1]),s[2]],H1);
           for C in F do
              E:= subs(C[1][2]=map(k->helpBesSqRootSolutions3(k),C[1][2]),C[1]);
              E:= seq([E[1],op(E[2][i]),C[2]],i=1..nops(E[2]));
              H[C]:= E
           od;
           H:= convert(H,'set');
           return H
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################ 2F2 TYPE SOLUTIONS ############################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing2F2(L2,t,G,ext);       ####### L2 can be our L operator.
# C1 = Exptest2F2equiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator.
# i = index of an element in C2[5][1].     ####### C2[5][1] = set of non-apparent singularities of L2
towopegenexpdiff2F2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;
                  # options trace;
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equiva2F22(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equiva2F21(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsing2F2(L2,t,G,ext);       ####### L2 can be our L operator.
Exptest2F2equiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= Condi2F2irr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing2F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest2F2equiv2(L1,C2,x,t,ext)       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
Exptest2F2equiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing2F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest2F2equiv2(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equiva2F21:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equiva2F22:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing2F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest2F2equiv1(L1,C2,x,t,ext)     ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff2F2(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge2F21:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,G,ext).
EquivExpgauge2F21:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest2F2equiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff2F2(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge2F21(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing2F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest2F2equiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff2F2(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge2F2:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=Exptest2F2equiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(Exptest2F2equiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,G,ext).
EquivExpgauge2F2:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest2F2equiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff2F2(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge2F2(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 2F2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi2F2irr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l;
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 2 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              if l <> 1 then
                 return NULL
              fi;
              C:= map(s-> convert(series(subs(t=1/t,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              if (evala(A[1]-A[2])=0 or evala(A[1]+A[2])=0) then
                 return [E[2],E[4],A,B]
              elif (evala(A[1]-A[3])=0 or evala(A[1]+A[3])=0) then
                    return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]],[A[1],A[3],A[2]],[B[1],B[3],B[2]]]
              elif (evala(A[2]-A[3])=0 or evala(A[2]+A[3])=0) then
                    return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]],[A[2],A[3],A[1]],[B[2],B[3],B[1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm "irrsing2F2"
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext).
irrsing2F2:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or
                    member(t,indets(A[i[1]][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d:= Condi2F2irr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 2F2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol.
# E = genexpdiff(L,i,t,ext).
Condi2F2reg:=proc(E,t,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if type(evala(2*(E[2][2]-E[2][1]-E[2][3])),integer) then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif type(evala(2*(E[2][1]-E[2][2]-E[2][3])),integer) then
                 return [[E[2][2],E[2][1],E[2][3]],[E[4][2],E[4][1],E[4][3]]]
              elif type(evala(2*(E[2][3]-E[2][1]-E[2][2])),integer) then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# R = Singular(L,extppp);
# G = NotAppSing(L,R,ext);
# H = irrsing2F2(L,t,G,ext)[-1] or IrrRegAppsing2F2(L,t,R,ext)[1][-1].
regsingtrue2F2:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                 # must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= Condi2F2reg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1];
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in two parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer)
#       - removable singularities with integer exponent differences.
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# E = Singular(L,extppp);
# G = NotAppSing(L,E,ext);
# Sirr = irrsing2F2(L,t,G,ext) or IrrRegAppsing2F2(L,t,E,ext)[1];
# H = regsingtrue2F2(L,t,Sirr[-1],ext) or IrrRegAppsing2F2(L,t,E,ext)[2].
Sregseptrue2F2:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# G = NotAppSing(L,H,ext);
# Sirr = irrsing2F2(L,t,G,ext) or IrrRegAppsing2F2(L,t,H,ext)[1];
# Sreg = regsingtrue2F2(L,t,Sirr[-1],ext) or IrrRegAppsing2F2(L,t,H,ext)[2];
# RSreg = Sregseptrue2F2(L,Sreg,ext) or IrrRegAppsing2F2(L,t,H,ext)[3].
UnionSregseptrue2F2:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################

# Delta = exponent differences of the singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# Sirr = irrsing2F2(L,t,F,ext) or IrrRegAppsing2F2(L,t,G,ext)[1];
# Sreg = regsingtrue2F2(L,t,Sirr[-1],ext) or IrrRegAppsing2F2(L,t,G,ext)[2].
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp).
IrrRegAppsing2F2:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,
                    H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or
                    member(t,indets(A[i][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d1:= Condi2F2irr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                    # must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # generalized exponents have to be positive integers.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= Condi2F2reg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1];
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31];
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31];
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= Condi2F2reg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1];
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31];
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31];
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsing1F1sq(L,t,F,ext)[1][i],x),
#######                      with i=position of the minimal polynom of p) is 1.
############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g.
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsing2F2(L,t,ext).
Hyperpossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:= table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     C[k]:= h;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g.
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Bespossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:=table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     if member(-h,{op(C)})=false and member(h,{op(C)})=false then
                           C[k]:= h;
                     fi;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# Sirr = irrsing2F2(L,t,F,ext) or IrrRegAppsing2F2(L,t,H,ext)[1].
Hyp2F2Subst:=proc(L,x,t,Sirr,ext)
               local Dx,m,A,ff,B,extp0,extp1,extp2,extp3,extp4,extp5,v,v1,i,j,s;
               Dx,x:=op(_Envdiffopdomain);
               A := table();
               i:=0;
               for s in Sirr[1] do
                  i:=i+1;
                  m:= ldegree(ReplirrRoot(Sirr[3][i][1],{}),t);
                  if m=FAIL then
                     return NULL
                  fi;
                  ff := add( (coeff(convert(eval(subs(csgn=1,simplify(ReplirrRoot(Sirr[3][i][1],{})) assuming(t>0))),                                                   polynom),t,j)/j)*(`if`(s[1]=infinity,1/x,x-s[2]))^j,j=m .. -1);
                  ff:=evala(Reduce(ff));
                  if s[1]=infinity then
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All the names or RootOf
                                           # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp5:= ext union extp1
                  else
                     extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                     extp2:= extp1 minus extp0; # All the names or RootOf
                                                 # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp3:= ext minus extp0;
                     extp4:= extp2 union extp3;
                     v1:={};
                     for v in extp4 do
                        if extp0 intersect indets(v,{RootOf,name}) <> {} then
                           v1:= v1 union {v}
                        fi;
                     od:
                     extp5:= extp4 minus v1;
                  fi;
                  A[s]:= evala(Trace(convert(evala(ff),polynom),extp5,extp5),x)
               od;
               A:= convert(A,'list');
               B := Hyperpossibility(A);
               B:= map(j-> `if`(member(x,indets(j,{name})),j,NULL),B);
               if B=[] then
                  return NULL
               fi;
               return map(s-> Normalizer(s),B)
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
Zeroesmultordpol2F2:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsing2F2(L,t,H,ext)[2];
# B = Zeroesmultordpol2F2(f,x);
# n = nops(B).
newregsing2F2:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],Condi2F2reg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],Condi2F2reg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################

# Input: a function f(x) and a minimal polynomial i. The minimal polynomial with Root infinity is infinity.
# Output: f(x)+c such that f(p)+c=0 with p=RootOf(i,x). If i=infinity then f(x) is returned i.e c=0.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# Sirr = R[1];
# f = element of Hyp2F2Subst(L,x,t,Sirr,ext);
# i = element of R[2][1].
changeconstant2F2 := proc(f,x,i,ext)
                 local Dx,c,A,ff,C;
                 Dx,x:=op(_Envdiffopdomain);
                 if i[1]=infinity then
                    A:= f;
                 else
                    ff:=numer(Normalizer(f+C));
                    c:=SolveTools:-Linear({coeffs(convert(collect(eval(subs(csgn=1,simplify(evala(Rem(ff,i[1],x)))
                                                               assuming(x>0))),x),polynom),x)},{C});
                    if c=NULL or nops(c) <> 1 then
                       return NULL;
                    else
                       A:=Normalizer(f+subs(c[1],C));
                    fi;
                 fi;
                 return factor(A);
end:

############################################################################################################
############################################################################################################

# Give the good candidates for f by looking at those candidates which vanish at elements of regsingtrue (all
# elements of regsingtrue should give the same constant c). Note that regsingtrue <> {} in this case.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# Sirr = R[1];
# F = Hyp2F2Subst(L,x,t,Sirr,ext).
Candichangvar2F2:=proc(F,R,ext) ### For all the case except the integer case.
               local Dx,x,i,f,A,j,s,B1,n1,SR,n;
               Dx,x:=op(_Envdiffopdomain);
               A:= table();
               n:=nops(R[2][1]);
               for s in F do
                  j:=0;
                  f:= changeconstant2F2(s,x,R[2][1][1],ext);
                  if f=NULL then
                     next
                  fi;
                  j:=j+1;
                  if n>1 then
                     for i from 2 to n do
                        if R[2][1][i][1]=infinity then
                           if not degree(numer(f),x)<degree(denom(f),x) then
                              break
                           fi;
                           j:=j+1;
                        else
                           if not evala(Rem(numer(f),R[2][1][i][1],x))=0 then
                              break
                           fi;
                           j:=j+1;
                        fi;
                     od;
                  fi;
                  if j <> n then
                     next
                  fi;
                  B1:= Zeroesmultordpol2F2(f,x);
                  if B1=NULL then
                     next
                  fi;
                  n1:= nops(B1);
                  SR:= newregsing2F2(L,t,x,B1,R[2],n1,ext);
                  A[s] := [f,SR[2],SR[1]];
               od;
               A:= convert(A,'set');
               if A={} then
                  return NULL
               else
                  return A
               fi;
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate [b1,b2] for the lower parameter of 2F2 must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = Zeroesmultordpol2F2(f,x);
# n = nops(B1);
# RS = newregsing2F2(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# (b1,b2) = candidate for the lower parameters.
Condibiexp2F2:= proc(E1,E2,m,b1,b2,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(m*(1-b1)-E2[1]),integer) and type(evala(m*(1-b2)-E2[2]),integer) and
                 type(evala(m*(b1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b1)),integer) and
                 type(evala(E1[2][1]-m*(1-b2)),integer)) then
                 return true
              elif (type(evala(m*(1-b2)-E2[1]),integer) and type(evala(m*(1-b1)-E2[2]),integer) and
                 type(evala(-m*(b1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b2)),integer) and
                 type(evala(E1[2][1]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(1-b1)-E2[1]),integer) and type(evala(m*(b1-b2)-E2[2]),integer) and
                 type(evala(m*(1-b2)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b1)),integer) and
                 type(evala(E1[2][1]-m*(1-b2)),integer)) then
                 return true
              elif (type(evala(m*(b1-b2)-E2[1]),integer) and type(evala(-m*(1-b1)-E2[2]),integer) and
                 type(evala(-m*(1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-m*(1-b2)),integer) and
                 type(evala(E1[1][2]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(1-b2)-E2[1]),integer) and type(evala(-m*(b1-b2)-E2[2]),integer) and
                 type(evala(m*(1-b1)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b2)),integer) and
                 type(evala(E1[2][1]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(b1-b2)-E2[1]),integer) and type(evala(-m*(1-b2)-E2[2]),integer) and
                 type(evala(-m*(1-b1)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-m*(1-b1)),integer) and
                 type(evala(E1[1][2]-m*(1-b2)),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

####### Let E be a list of two element [a,b]. We return a list [c,d] where c and d are the constant part of a and b,
####### respectively.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# i = index of an element in R[1][1];
# m = R[1][5][i];
# E = an element in R[1][6][i]. E is a list of two elements a and b (E=[a,b]);
helpCondiaibiexp2F2:= proc(E,t,m,ext)
              local Dx,x,a,i,c1,C;
              Dx,x:=op(_Envdiffopdomain);
              C:= map(s-> convert(series(subs(t=1/t^m,s),t=0,100),polynom),E);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              C:= map(s-> evala(coeff(s,t,0)),C);
              return C
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# k = index of an element in R[1][1];
# m1 = R[1][5][k];
# E2 = R[1][8][k];
# E1 = map(s-> helpCondiaibiexp2F2(s,t,m1,ext),R[1][6][k]);
# m = R[1][4][k].
Condiaibiexp2F2:= proc(E2,E1,m,b1,b2,a1,a2,ext)
           local Dx,x;
           Dx,x:=op(_Envdiffopdomain);
          if (type(evala(m*(a2-a1)-E2[3]),integer) and type(evala(m*(b1+b2-2*a1-a2)-E2[1]),integer) and
              type(evala(m*(b1+b2-a1-2*a2)-E2[2]),integer)) and type(evala(E1[1][1]-E1[2][1]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][2]-E1[3][1]),integer) then
              if indets(a1+a2,{RootOf,name}) subset indets(evala(E1[1][1]-m*(b1+b2)),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) and
                 indets(a2,{RootOf,name}) subset indets(evala(E1[2][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(m*(a2-a1)-E2[3]),integer) and type(evala(m*(b1+b2-2*a1-a2)-E2[1]),integer) and
              type(evala(m*(b1+b2-a1-2*a2)+E2[2]),integer)) and type(evala(E1[1][1]-E1[2][2]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a1+a2,{RootOf,name}) subset indets(evala(E1[1][1]-m*(b1+b2)),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) and
                 indets(a2,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(m*(a1-a2)-E2[3]),integer) and type(evala(m*(b1+b2-a1-2*a2)-E2[1]),integer) and
              type(evala(m*(b1+b2-2*a1-a2)-E2[2]),integer)) and type(evala(E1[1][1]-E1[2][1]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][2]-E1[3][1]),integer) then
              if indets(a1+a2,{RootOf,name}) subset indets(evala(E1[1][1]-m*(b1+b2)),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[2][2]),{RootOf,name}) and
                 indets(a2,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(m*(a1-a2)-E2[3]),integer) and type(evala(m*(b1+b2-a1-2*a2)-E2[1]),integer) and
              type(evala(m*(b1+b2-2*a1-a2)+E2[2]),integer)) and type(evala(E1[1][1]-E1[2][2]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a1+a2,{RootOf,name}) subset indets(evala(E1[1][1]-m*(b1+b2)),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) and
                 indets(a2,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(m*(a2-a1)-E2[3]),integer) and type(evala(m*(b1+b2-2*a1-a2)+E2[1]),integer) and
              type(evala(m*(b1+b2-a1-2*a2)+E2[2]),integer)) and type(evala(E1[1][2]-E1[2][2]),integer) and
              type(evala(E1[1][1]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a1+a2,{RootOf,name}) subset indets(evala(E1[1][2]-m*(b1+b2)),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[1][1]),{RootOf,name}) and
                 indets(a2,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(m*(a1-a2)-E2[3]),integer) and type(evala(m*(b1+b2-a1-2*a2)+E2[1]),integer) and
              type(evala(m*(b1+b2-2*a1-a2)+E2[2]),integer)) and type(evala(E1[1][2]-E1[2][2]),integer) and
              type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-E1[3][2]),integer) then
              if indets(a1+a2,{RootOf,name}) subset indets(evala(E1[1][2]-m*(b1+b2)),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) and
                 indets(a2,{RootOf,name}) subset indets(evala(E1[1][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           else
              return false
           fi;
end:

############################################################################################################
############################################################################################################

# For a given candidates [[[b1,b2],[b1,b2],...]],f], we find associated candidates for [{{a1,a2},...,},[b1,b2],f]
# related an element of [[b1,b2],[b1,b2],...] and an irregular singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# F = R[1];
# C = Candidates for [[b1,b2],[b1,b2],[b1,b2]];
# i = index of an element in F[1].
helpcandai2F2:=proc(L,F,C,i,t,ext,ext1)
            local Dx,x,E1,E2,E,H,s,A,B,j,m,m1,A1,A2,A3,B1,B2,k,a1,a2;
            Dx,x:=op(_Envdiffopdomain);
            m:= F[4][i];
            m1:= F[5][i];
            E2:= F[8][i];
            E1:= map(s-> helpCondiaibiexp2F2(s,t,m1,ext),F[6][i]);
            ####### candidates for a2-a1 modulo an integer
            B:= [seq(E2[3]+j,j=-m+1..m-1),seq(-E2[3]+j,j=-m+1..m-1)]/m;
            B:=[op({op(B)})];
            H:= table();
            for j in C do
               ####### candidates for a1+a2 modulo an integer
               A:= [seq((2*(j[1]+j[2])-(E2[1]+E2[2]+s)/m)/3,s=-3*m+1..3*m-1),
                    seq((2*(j[1]+j[2])-(E2[1]-E2[2]+s)/m)/3,s=-3*m+1..3*m-1),
                    seq((2*(j[1]+j[2])-(-E2[1]+E2[2]+s)/m)/3,s=-3*m+1..3*m-1),
                    seq((2*(j[1]+j[2])-(-E2[1]-E2[2]+s)/m)/3,s=-3*m+1..3*m-1)];
               A:=[op({op(A)})];
               ####### We compute the candidates for "a1" and "a2" modulo an integer
               A1:= table();
               A2:= table();
               A3:= table();
               for k in B do
                  for s in A do
                     a1:= (s-k)/2;
                     a2:= (s+k)/2;
                     ####### We make the rational part of "a1" and "a2" belong to ]0,1[ union {-1,1}
                     B1:=Ratiopartmodint(a1,ext1);
                     B2:=Ratiopartmodint(a2,ext1);
                     A1[k,s]:= `if`(B1[1]=0,op([[-1,-1,B1[3],B1[4]],[1,1,B1[3],B1[4]]]),B1);
                     A2[k,s]:= `if`(B2[1]=0,op([[-1,-1,B2[3],B2[4]],[1,1,B2[3],B2[4]]]),B2);
                  od;
               od;
               A1:= convert(A1,'set');
               A2:= convert(A2,'set');
               for k in A1 do
                  for s in A2 do
                     if Condiaibiexp2F2(E2,E1,m,j[1],j[2],k[1],s[1],ext)=true then
                        A3[k,s]:= [k[1],s[1]],[s[1],k[1]]
                     fi;
                  od;
               od;
               A3:= convert(A3,'set');
               H[j]:= [A3,j];
            od;
            H:= convert(H,'list');
            return H
end:

############################################################################################################
############################################################################################################

# We compute the candiates for [{{a1,a2},{a1,a2}...},[b1,b2]] related to a candidate of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,H,ext);
# F = R[1];
# C = Candidates for [[b1,b2],[b1,b2],[b1,b2]].
findcandai2F2:=proc(L,F,C,x,t,ext,ext1)
                  local j,B,k,E,H;
                  H := table();
                  E:= helpcandai2F2(L,F,C,1,t,ext,ext1);
                  if E=NULL then
                     return NULL
                  fi;
                  if nops(F[1]) > 1 then
                     for j from 2 to nops(F[1]) do
                        B:= helpcandai2F2(L,F,C,j,t,ext,ext1);
                        if B=NULL then
                           return NULL
                        fi;
                        for k from 1 to nops(B) do
                           E[k][1]:= B[k][1] intersect E[k][1];
                        od;
                     od;
                  fi;
                  for j from 1 to nops(E) do
                     if E[j][1]<>{} then
                        E[j][1]:={seq(`if`(Matchlist([op(E[j][1][1..k-1])],E[j][1][k]), NULL, E[j][1][k]),
                                  k=1..nops(E[j][1]))};
                        H[j]:= E[j];
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
####################################   Logarithmic Case   ##################################################
############################################################################################################
############################################################################################################

####### b11, b12 and b13 are sets of candidates for 1-b1, 1-b2 and b1-b2 respectively. We compute here the candidates
####### for [b1,b2] with the rational part of b1 and b2 in ]0,1], related to a zero of the change of variable parameter.
#######
####### m = multiplicity order of the element as a zero of f.
####### e = [Delta1,Delta2,Delta3]
####### a = [[e,e],[e,e],[e,e]].
casesbiLog2F2:= proc(b11,b21,b31,e,a,m,ext,ext1)
          local k,s,b1,b2,b3,B,B1,B2,B3,d1,d2,i,j;
          b1:=map(s->1-s,b11);
          b2:=map(s->1-s,b21);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          b2:= [seq(`if`(MatchInt2([op(b2[1..k-1])],b2[k]), NULL, b2[k]),k=1..nops(b2))];
          b3:= [seq(`if`(MatchInt2([op(b31[1..k-1])],b31[k]), NULL, b31[k]),k=1..nops(b31))];
          ####### We compute the candidates for b1, b2 and b1-b2 such that their rational parts are in [0,1]
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          B2:={op(map(s-> Ratiopartmodint(s,ext1),b2))};
          B3:={op(map(s-> Ratiopartmodint(s,ext1),b3))};
          B3:=map(s-> [s[1],s[2],s[3]],B3);
          B1:=map(s->`if`(s[1]=0,[1,1,s[3],s[4]],s),B1);
          B2:=map(s->`if`(s[1]=0,[1,1,s[3],s[4]],s),B2);
          B:= table();
          for i in B1 do
             for j in B2 do
                ####### the condition for the logarithmic case
                if type(i[1],integer) or type(j[1],integer) or type(i[1]-j[1],integer) or (type(i[1]*2,integer)
                   and type(j[1]*2,integer) and i[1]*j[1]< 0) then
                   d1:= Ratiopartmodint(i[1]-j[1],ext1);
                   d2:= Ratiopartmodint(j[1]-i[1],ext1);
                   d1:= [d1[1],d1[2],d1[3]];
                   d2:= [d2[1],d2[2],d2[3]];
                   if {d1,d2} intersect B3 <> {} then
                      ####### the condition that a candidate for (b1,b2) must satisfy
                      if Condibiexp2F2(a,e,m,i[1],j[1],ext)=true or Condibiexp2F2(a,e,m,j[1],i[1],ext)=true then
                         B[i,j]:= {i[1],j[1]}
                      fi;
                   fi;
                fi;
             od;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameters [b1,b2] related to a zero of the change of variable
# parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,G,ext);
# F = Hyp2F2Subst(L,x,t,R[1],ext);
# B = Candichangvar2F2(F,R,ext) ;
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiffLog2F2:= proc(j,G,m,ext,ext1)
          local d,i,ext2,b11,b21,b31,B,H1,H2,H3,C1,C,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          B:= table();
          C:=[[G[3][j][1],G[3][j][2],G[3][j][3]],[G[3][j][1],G[3][j][3],G[3][j][2]],[G[3][j][2],G[3][j][3],G[3][j][1]]];
          for i in C do
             ####### b11 is the list of candidates for +-(1-b1) modulo an integer
             b11:=[seq(i[1]+s,s=-m+1..m-1),seq(-i[1]+s,s=-m+1..m-1)]/m;
             ####### b21 is the list of candidates for +-(1-b2) modulo an integer
             b21:= [seq(i[2]+s,s=-m+1..m-1),seq(-i[2]+s,s=-m+1..m-1)]/m;
             ####### b31 is the list of candidates for +-(b1-b2) modulo an integer
             b31:= [seq(i[3]+s,s=-m+1..m-1),seq(-i[3]+s,s=-m+1..m-1)]/m;
             C1:=casesbiLog2F2(b11,b21,b31,G[3][j],G[4][j],m,ext,ext1);
             if C1 <> NULL then
                B[i]:=op(C1)
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters b1 and b2 related to a candidate of f: [[[b1,b2],[b1,b2]...],f,M],
# where M is the list of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,G,ext);
# F = Hyp2F2Subst(L,x,t,R[1],ext);
# B1 = Candichangvar2F2(F,R,ext);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter.
# f[2] list of multiplicity orders of zeroes of f[1].
# f[3][1] list of zeroes of the change of variable parameter.
findbi2F2ln:=proc(L,f,x,t,ext,ext1,n1)
                  local H,B,j,s,i;
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiffLog2F2(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> `if`(nops(s)=1,[s[1],s[1]],[s[1],s[2]]),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a1,a2,b1,b2), also and f:[{{a1,a2},...},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,E,ext);
# F = Hyp2F2Subst(L,x,t,R[1],ext).
find2F2ln:=proc(L,R,F,ext,x,t)
                  local B,ab,f,C,H,E,n1,n,ext1,ext2,ext3,ext4,i,n2,s;
                  B:= Candichangvar2F2(F,R,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                  ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                  n1:=nops(R[2][1]);
                  if n1 > 1 then
                     for i from 2 to n1 do
                        ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                        ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                        ext1:= ext1 intersect ext2;
                     od;
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi2F2ln(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        ab:= findcandai2F2(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##################################   RATIONAL AND IRRATIONAL CASE   ########################################
############################################################################################################
############################################################################################################

####### b11, b12 and b13 are sets of candidates for 1-b1, 1-b2 and b1-b2 respectively. We compute here the candidates
####### for [b1,b2] with the rational part of b1 and b2 in ]0,1].
#######
####### m = multiplicity order of the element as a zero of f.
####### e = [Delta1,Delta2,Delta3]
####### a = [[e,e],[e,e],[e,e]]].
casesbi2F2:= proc(b11,b21,b31,e,a,m,ext,ext1)
          local k,s,b1,b2,b3,B,B1,B2,B3,d1,d2,i,j;
          b1:=map(s->1-s,b11);
          b2:=map(s->1-s,b21);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          b2:= [seq(`if`(MatchInt2([op(b2[1..k-1])],b2[k]), NULL, b2[k]),k=1..nops(b2))];
          b3:= [seq(`if`(MatchInt2([op(b31[1..k-1])],b31[k]), NULL, b31[k]),k=1..nops(b31))];
          ####### We compute the candidates for b1, b2 and b1-b2 such that their rational parts are in [0,1]
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          B2:={op(map(s-> Ratiopartmodint(s,ext1),b2))};
          B3:={op(map(s-> Ratiopartmodint(s,ext1),b3))};
          B3:=map(s-> [s[1],s[2],s[3]],B3);
          B1:=map(s->`if`(s[1]=0,NULL,s),B1);
          B2:=map(s->`if`(s[1]=0,NULL,s),B2);
          if B1={} or B2={} then
             return NULL
          fi;
          B:= table();
          for i in B1 do
             for j in B2 do
                ####### the condition for the non-logarithmic case
                if type(i[1]-j[1],integer)=false and (type(i[1]*2,integer)=false or type(j[1]*2,integer)=false or
                   i[1]*j[1]>= 0) then
                   d1:= Ratiopartmodint(i[1]-j[1],ext1);
                   d2:= Ratiopartmodint(j[1]-i[1],ext1);
                   d1:= [d1[1],d1[2],d1[3]];
                   d2:= [d2[1],d2[2],d2[3]];
                   if {d1,d2} intersect B3 <> {} then
                      ####### the condition that a candidate for (b1,b2) must satisfy
                      if Condibiexp2F2(a,e,m,i[1],j[1],ext)=true or Condibiexp2F2(a,e,m,j[1],i[1],ext)=true then
                         B[i,j]:= {i[1],j[1]}
                      fi;
                   fi;
                fi;
             od;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameters [b1,b2] related to a zero of the change of variable
# parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,G,ext);
# F = Hyp2F2Subst(L,x,R[1],ext);
# B = Candichangvar2F2(F,R,ext) ;
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiff2F2:= proc(j,G,m,ext,ext1)
          local d,i,ext2,b11,b21,b31,B,H1,H2,H3,C1,C,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          B:= table();
          C:=[[G[3][j][1],G[3][j][2],G[3][j][3]],[G[3][j][1],G[3][j][3],G[3][j][2]],[G[3][j][2],G[3][j][3],G[3][j][1]]];
          for i in C do
             ####### b11 is the list of candidates for +-(1-b1) modulo an integer
             b11:=[seq(i[1]+s,s=-m+1..m-1),seq(-i[1]+s,s=-m+1..m-1)]/m;
             ####### b21 is the list of candidates for +-(1-b2) modulo an integer
             b21:= [seq(i[2]+s,s=-m+1..m-1),seq(-i[2]+s,s=-m+1..m-1)]/m;
             ####### b31 is the list of candidates for +-(b1-b2) modulo an integer
             b31:= [seq(i[3]+s,s=-m+1..m-1),seq(-i[3]+s,s=-m+1..m-1)]/m;
             C1:=casesbi2F2(b11,b21,b31,G[3][j],G[4][j],m,ext,ext1);
             if C1 <> NULL then
                B[i]:=op(C1)
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters b1 and b2 related to a candidate of f: [[[b1,b2],[b1,b2]...],f,M],
# where M is the list of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,G,ext);
# F = Hyp2F2Subst(L,x,R[1],ext);
# B1 = Candichangvar2F2(F,R,ext);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter.
# f[2] list of multiplicity orders of zeroes of f[1].
# f[3][1] list of zeroes of the change of variable parameter.
findbi2F2:=proc(L,f,x,t,ext,ext1,n1)
                  local H,B,j,s,i;
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiff2F2(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> `if`(nops(s)=1,[s[1],s[1]],[s[1],s[2]]),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a1,a2,b1,b2), also and f:[{{a1,a2},...},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,E,ext);
# F = Hyp2F2Subst(L,x,R[1],ext).
find2F2RatIrr:=proc(L,R,F,x,t,ext)
                  local Dx,B,ab,f,C,H,E,ext1,ext2,ext3,ext4,i,n1,n2,n,s,G1;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= Candichangvar2F2(F,R,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                  ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                  n1:=nops(R[2][1]);
                  if n1 > 1 then
                     for i from 2 to n1 do
                        ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                        ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                        ext1:= ext1 intersect ext2;
                     od;
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi2F2(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        ab:= findcandai2F2(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
######################################    INTEGER CASE    ##################################################
############################################################################################################
############################################################################################################

# Input: A monic polynomial f\in K[x] and a positive integer p\in\N
# Output: g\in K[x], such that if there exists a solution to y^p=f, then y=g is such a solution.
ispower := proc(f,x,p)
           local d,n,A,Ap,a,i,b;
           if p=1 or degree(f,x)=0 then
                return f
           fi;
           d:=degree(f,x);
           n:=d/p;
           if not type(n,integer) then
                return NULL
           fi;
           A:=add(a[i]*x^i,i=0..n-1)+x^n;
           Ap:=collect(A^p,x);
           for i from 1 to n do
                # Solve a linear equation in one unknown:
                b[n-i]:=solve(subs({seq(a[n-j]=b[n-j],j=1..i-1)},coeff(Ap,x,d-i))-coeff(f,x,d-i),a[n-i] );
           od;
           Ap:=subs({seq(a[n-j]=b[n-j],j=1..n)},Ap);
           subs({seq(a[n-j]=b[n-j],j=1..i-1)},A)
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,E,ext);
# F = Hyp2F2Subst(L,x,t,R[1],ext).
find2F2Int:= proc(L,R,F,x,t,ext)
          local Dx,ff,f,g1,n,i,g,lc,a,k,V1,ext3,ext4,v,C,c,cc,fctrs,bf,s,inff,B1,n1,RS,v1,j,n2,ab,res,E;
           Dx,x:=op(_Envdiffopdomain);
           ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
           ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
           n2:=nops(R[1][1]);
           if n2 > 1 then
              for i from 2 to n2 do
                 ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                 ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                 ext3:= ext3 intersect ext4;
              od;
           fi;
           bf:=table();
           for f in F do
               ff:=Normalizer(f+C);
               inff:=map(s->s[1],R[1][1]);
               if member(infinity,{op(inff)}) then
                   n:=degree(numer(ff));
               else
                   n:=degree(denom(ff));
               fi;
               # denom(2*nu) divises n. So if n=1 then denom(2*nu)=1 and that means nu belong to 1/2+Z or
               # nu belong to Z which is not good. nu also have to be different from 0 (see the proof of
               # one of corollary in the thesis related to this part).
               if n<=1 then
                  next
               fi;
               g:=numer(ff);
               lc:=lcoeff(g,x);
               res:=table();
               for i from 2 to n do
                  # i must be a divisor of n
                  V1:= table();
                  if irem(n,i)=0 then
                     # f must be in the form g^i.
                     a:=ispower(collect(g/lc,x,Normalizer),x,i);
                     if a=NULL then
                        next
                     fi;
                     s:=numer(Normalizer(evala(Content(g-lc*a^i,x))));
                     s:=evala(Factors(s,ext));
                     cc:={seq(`if`(degree(k[1],C)=1,-coeff(k[1],C,0)/coeff(k[1],C,1),NULL),k=s[2])};
                     if member(infinity,{op(inff)})=false then
                        fctrs:=sqrfree(subs(C=0,g),x);
                        # the multiplicity of each factor of g is a multiple of i
                        if add(modp(k[2],i),k=fctrs[2])=0 then
                           cc:=cc union {0};
                        fi;
                     fi;
                     if cc <> {} then
                        for c in cc do
                           g:= factor(subs(C=c,ff));
                           B1:= Zeroesmultordpol2F2(g,x);
                           if B1=NULL then
                              next
                           fi;
                           n1:= nops(B1);
                           RS:= newregsing2F2(L,t,x,B1,R[2],n1,ext);
                           g1:= [g,RS[2],RS[1]];
                           n1:= nops(g1[3][1]);
                           E:= findbi2F2(L,g1,x,t,ext,{},n1);
                           if E<>NULL then
                              ab:= findcandai2F2(L,R[1],E[1],x,t,ext,ext3);
                              if ab <> NULL then
                                 V1[c]:= [op(ab),E[2]]
                              fi;
                           fi;
                        od;
                     fi;
                  fi;
                  V1:= convert(V1,'list');
                  if V1<>[] then
                     res[i]:= op(V1);
                  fi;
               od;
               res:= convert(res,'list');
               if res<>[] then
                  bf[f]:= op(res);
               fi;
            od;
            bf:= convert(bf,'list');
            if bf=[] then
               return NULL
            else
               return bf
            fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature 1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1, l2, p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing2F2(L,t,E,ext);
# F = Hyp2F2Subst(L,x,t,R[1],ext).
help2F2Solutions1:= proc(L,R,F,x,t,T,ext)
           local Dx,B,s,H;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H:= find2F2Int(L,R,F,x,t,ext);
           elif R[3][3]<>[] then
              H:= find2F2ln(L,R,F,ext,x,t);
           else
              H:= find2F2RatIrr(L,R,F,x,t,ext);
           fi;
           if H=NULL then
              return NULL
           else
              H:= {op(map(s-> op([s,subs(s[-1]=-s[-1],s)]),H))};
              return [op(H)]
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsing2F2(L,t,G,ext);
# a = element of help2F2Solutions1(L,r,RSreg,info,ext,x).
help2F2Solutions2:= proc(L,a,A,x,t,ext)
           local Dx,J,n2,C1,M,L22,k1,j,R,i,e1,e2,e3,e4,L12,T;
           Dx,x:=op(_Envdiffopdomain);
           n2:= nops(a);
           C1:= sumdiffeq(hyperterm([e1,e2],[e3,e4],x,k1),k1,J(x));
           L12:= de2diffop(C1,J(x));
           for i from 1 to n2-1 do
              for j in a[i][1] do
                 L22:=subs({e1=j[1],e2=j[2],e3=a[i][2][1],e4=a[i][2][2]},L12);
                 M:= ChangeOfVariables(L22,a[n2]);
                 R:=EquivExpgauge2F2(M,L,A,x,t,T,ext);
                 if R<>NULL then
                    return [{[[j],a[i][2],R]},a[n2]]
                 fi;
              od;
           od;
           return NULL
end:

############################################################################################################
############################################################################################################

help2F2Solutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

help2F2Solutions4:= proc(a,b)
            local Dx,x,n1,n2;
            Dx,x:=op(_Envdiffopdomain);
            if a[3]<>b[3] then
               return [false,b]
            elif Matchlistmod2(a[1],b[1])=false or Matchlistmod2(a[2],b[2])=false then
               return [false,b]
            else
               n1:= length(a[4]);
               n2:= length(b[4]);
               if n2<n1 then
                  return [true,b]
               else
                  return [true,a]
               fi;
            fi;
end:

############################################################################################################
############################################################################################################

help2F2Solutions5:= proc(G,b)
            local Dx,x,n1,n2,n,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n1:= nops(G);
            n2:= 0;
            for i from 1 to n1 do
               E:= help2F2Solutions4(G[i],b);
               if E[1]=false then
                  n2:= n2+1
               else
                  n:= i;
                  break
               fi;
            od;
            if n2=n1 then
              return [false,b]
            else
               return [true,[E[2],n]]
            fi;
end:

############################################################################################################
############################################################################################################

help2F2Solutions6:= proc(H)
            local Dx,x,n,G,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n:= nops(H);
            G:=table();
            G[1]:= H[1];
            if n>1 then
               for i from 2 to n do
                  E:= help2F2Solutions5(convert(G,'list'),H[i]);
                  if E[1]=true then
                     G[E[2][2]]:= E[2][1]
                  else
                     G[i]:= E[2]
                  fi;
               od;
            fi;
            G:= convert(G,'list');
            return G
end:

############################################################################################################
############################################################################################################

#
Find2F2Solutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsing2F2(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= Hyp2F2Subst(L,x,t,R[1],ext);
           if F=NULL then
              return NULL
           fi;
           C:= table();
           H:= help2F2Solutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= help2F2Solutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

Hyp2F2Solutions:= proc(L)
           local Dx,x,s,k,F,E,H1,H2,G,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:= table();
           H1:= Find2F2Solutions(L);
           if H1=NULL then
              return "No 2F2 type solutions"
           fi;
           F:= table();
           G:= map(s->s[2],H1);
           for i in G do
              H2:= [[op(map(s->`if`(s[2]=i,op(s[1]),NULL),H1))],i];
              F[i]:= H2
           od;
           F:= convert(F,'set');
           for C in F do
              E:= map(s->subs(s[3]=map(k->help2F2Solutions3(k),s[3]),s),C[1]);
              E:= map(s->seq([s[1],s[2],op(s[3][i])],i=1..nops(s[3])),E);
              E:= help2F2Solutions6(E);
              H[C]:= [E,C[2]]
           od;
           H:= convert(H,'set');
           return H
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
######################################## 1F1 SQUARE TYPE SOLUTIONS ########################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F1sq(L2,t,G,ext);       ####### L2 can be our L operator.
# C1 = Exptest1F1sqequiv2(L1,C2,x,t,ext)       ####### L1 can be our M operator.
# i = index of an element in C2[5][1]    ####### C2[5][1] = set of non-apparent singularities of L2.
towopegenexpdiff1F1sq:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equiva1F1sq2(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equiva1F1sq1(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsing1F1sq(L2,t,G,ext);       ####### L2 can be our L operator.
Exptest1F1sqequiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= Condi1F1sqirr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F1sqequiv2(L1,C2,x,t,ext)       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
Exptest1F1sqequiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F1sqequiv2(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equiva1F1sq1:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               if extp0 subset ext then
                  return extp3 union ext
               fi;
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equiva1F1sq2:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F1sqequiv1(L1,C2,x,t,ext)       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff1F1sq(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge1F1sq1:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,G,ext).
EquivExpgauge1F1sq1:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest1F1sqequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff1F1sq(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge1F1sq1(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M
# L2 = L
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F1sqequiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff1F1sq(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge1F1sq:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=Exptest1F1sqequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(Exptest1F1sqequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,G,ext).
EquivExpgauge1F1sq:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest1F1sqequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff1F1sq(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge1F1sq(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 1F1^2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi1F1sqirr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l;
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 3 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              if l <> 1 then
                 return NULL
              fi;
              C:= map(s-> convert(series(subs(t=1/t,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              if (evala(A[1]-A[2])=0 or evala(A[1]+A[2])=0) and (evala(A[3]-2*A[1])=0 or evala(A[3]+2*A[1])=0) then
                 return [E[2],E[4],A,B]
              elif (evala(A[1]-A[3])=0 or evala(A[1]+A[3])=0) and (evala(A[2]-2*A[1])=0 or evala(A[2]+2*A[1])=0) then
                    return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]],[A[1],A[3],A[2]],[B[1],B[3],B[2]]]
              elif (evala(A[2]-A[3])=0 or evala(A[2]+A[3])=0) and (evala(A[1]-2*A[2])=0 or evala(A[1]+2*A[2])=0) then
                    return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]],[A[2],A[3],A[1]],[B[2],B[3],B[1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm "irrsing1F1sq"
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext).
irrsing1F1sq:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or
                    member(t,indets(A[i[1]][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d:= Condi1F1sqirr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 1F1^2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi1F1sqreg:=proc(E,t,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if evala(E[2][1]+E[2][2])=0 and evala(E[2][3]+2*E[2][1])=0 and evala(E[2][3]-2*E[2][2])=0 and
                 evala(E[2][1]-E[2][2]+E[2][3])=0 then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif evala(E[2][1]-E[2][3])=0 and evala(E[2][2]-2*E[2][1])=0 and evala(E[2][2]-2*E[2][3])=0 and
                 evala(E[2][1]+E[2][3]-E[2][2])=0 then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif evala(E[2][2]+E[2][3])=0 and evala(E[2][1]-2*E[2][2])=0 and evala(E[2][1]+2*E[2][3])=0 and
                 evala(E[2][2]-E[2][3]-E[2][1])=0 then
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              elif type(evala(E[2][1]+E[2][2]),integer)  and type(evala(E[2][3]+2*E[2][1]),integer) and
                 type(evala(E[2][3]-2*E[2][2]),integer) and type(evala(E[2][1]-E[2][2]+E[2][3]),integer) then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif type(evala(E[2][1]-E[2][3]),integer) and type(evala(E[2][2]-2*E[2][1]),integer) and
                 type(evala(E[2][2]-2*E[2][3]),integer) and type(evala(E[2][1]+E[2][3]-E[2][2]),integer) then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif type(evala(E[2][2]+E[2][3]),integer) and type(evala(E[2][1]-2*E[2][2]),integer) and
                 type(evala(E[2][1]+2*E[2][3]),integer) and type(evala(E[2][2]-E[2][3]-E[2][1]),integer) then
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H1 = NotAppSing(L,G,ext);
# H = irrsing1F1sq(L,t,H1,ext)[-1].
regsingtrue1F1sq:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                 # must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= Condi1F1sqreg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1];
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in two parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer)
#       - removable singularities with integer exponent differences.
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsing1F1sq(L,t,F,ext)[-1];
# H = regsingtrue1F1sq(L,t,R,ext).
Sregseptrue1F1sq:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsing1F1sq(L,t,F,ext)[-1];
# Sreg = regsingtrue1F1sq(L,t,R,ext);
# RSreg = Sregseptrue1F1sq(L,Sreg,ext).
UnionSregseptrue1F1sq:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################

# Delta = exponent differences of the singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# G = Singular(L,ext);
# F = NotAppSing(L,G,ext);
# Sreg = regsingtrue1F1sq(L,t,F,ext);
# Sirr = irrsing1F1sq(L,t,F,ext).
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp).
IrrRegAppsing1F1sq:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,
                    H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or
                    member(t,indets(A[i][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d1:= Condi1F1sqirr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                    # must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # generalized exponents have to be positive integers.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= Condi1F1sqreg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1];
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31];
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31];
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= Condi1F1sqreg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1];
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31];
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31];
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsing1F1sq(L,t,F,ext)[1][i],x),
#######                      with i=position of the minimal polynom of p) is 1.
############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g.
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n);
# Here n = nops(A[1]) where A = irrsing2F2(L,t,ext).
Hyperpossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:= table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     C[k]:= h;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g;
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n);
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Bespossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:=table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     if member(-h,{op(C)})=false and member(h,{op(C)})=false then
                           C[k]:= h;
                     fi;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# G = Singular(L,ext);
# F = NotAppSing(L,G,ext);
# Sirr = irrsing1F1sq(L,t,F,ext).
Hyp1F1sqSubst:=proc(L,x,t,Sirr,ext)
               local Dx,m,A,ff,B,extp0,extp1,extp2,extp3,extp4,extp5,v,v1,i,j,s;
               Dx,x:=op(_Envdiffopdomain);
               A := table();
               i:=0;
               for s in Sirr[1] do
                  i:=i+1;
                  m:= ldegree(ReplirrRoot(Sirr[3][i][1],{}),t);
                  if m=FAIL then
                     return NULL
                  fi;
                  ff := add( (coeff(convert(eval(subs(csgn=1,simplify(ReplirrRoot(Sirr[3][i][1],{})) assuming(t>0))),                                                   polynom),t,j)/j)*(`if`(s[1]=infinity,1/x,x-s[2]))^j,j=m .. -1);
                  ff:=evala(Reduce(ff));
                  if s[1]=infinity then
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All the names or RootOf
                                           # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp5:= ext union extp1
                  else
                     extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                     extp2:= extp1 minus extp0; # All the names or RootOf
                                                 # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp3:= ext minus extp0;
                     extp4:= extp2 union extp3;
                     v1:={};
                     for v in extp4 do
                        if extp0 intersect indets(v,{RootOf,name}) <> {} then
                           v1:= v1 union {v}
                        fi;
                     od:
                     extp5:= extp4 minus v1;
                  fi;
                  A[s]:= evala(Trace(convert(evala(ff),polynom),extp5,extp5),x)
               od;
               A:= convert(A,'list');
               B := Hyperpossibility(A);
               B:= map(j-> `if`(member(x,indets(j,{name})),j,NULL),B);
               if B=[] then
                  return NULL
               fi;
               return map(s-> Normalizer(s),B)
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
Zeroesmultordpol1F1sq:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsing1F1sq(L,t,H,ext)[2];
# B = Zeroesmultordpol1F1sq(f,x);
# n = nops(B).
newregsing1F1sq:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],Condi1F1sqreg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],Condi1F1sqreg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################

# Input: a function f(x) and a minimal polynomial i. The minimal polynomial with Root infinity is infinity.
# Output: f(x)+c such that f(p)+c=0 with p=RootOf(i,x). If i=infinity then f(x) is returned i.e c=0;
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# Sirr = R[1];
# f = element of Hyp1F1sqSubst(L,x,Sirr,ext);
# i = element of R[2][1].
changeconstant1F1sq := proc(f,x,i,ext)
                 local Dx,c,A,ff,C;
                 Dx,x:=op(_Envdiffopdomain);
                 if i[1]=infinity then
                    A:= f;
                 else
                    ff:=numer(Normalizer(f+C));
                    c:=SolveTools:-Linear({coeffs(convert(collect(eval(subs(csgn=1,simplify(evala(Rem(ff,i[1],x)))
                                                               assuming(x>0))),x),polynom),x)},{C});
                    if c=NULL or nops(c) <> 1 then
                       return NULL;
                    else
                       A:=Normalizer(f+subs(c[1],C));
                    fi;
                 fi;
                 return factor(A);
end:

############################################################################################################
############################################################################################################

# Give the good candidates for f by looking at those candidates which vanish at elements of regsingtrue (all
# elements of regsingtrue should give the same constant c). Note that regsingtrue <> {} in this case.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# Sirr = R[1];
# F = Hyp1F1sqSubst(L,x,Sirr,ext).
Candichangvar1F1sq:=proc(F,R,ext) ### For all the case except the integer case.
               local Dx,x,i,f,A,j,s,B1,n1,SR,n;
               Dx,x:=op(_Envdiffopdomain);
               A:= table();
               n:=nops(R[2][1]);
               for s in F do
                  j:=0;
                  f:= changeconstant1F1sq(s,x,R[2][1][1],ext);
                  if f=NULL then
                     next
                  fi;
                  j:=j+1;
                  if n>1 then
                     for i from 2 to n do
                        if R[2][1][i][1]=infinity then
                           if not degree(numer(f),x)<degree(denom(f),x) then
                              break
                           fi;
                           j:=j+1;
                        else
                           if not evala(Rem(numer(f),R[2][1][i][1],x))=0 then
                              break
                           fi;
                           j:=j+1;
                        fi;
                     od;
                  fi;
                  if j <> n then
                     next
                  fi;
                  B1:= Zeroesmultordpol1F1sq(f,x);
                  if B1=NULL then
                     next
                  fi;
                  n1:= nops(B1);
                  SR:= newregsing1F1sq(L,t,x,B1,R[2],n1,ext);
                  A[s] := [f,SR[2],SR[1]];
               od;
               A:= convert(A,'set');
               if A={} then
                  return NULL
               else
                  return A
               fi;
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate b for the lower parameter in 1F1^2 must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = Zeroesmultordpol1F1sq(f,x);
# n = nops(B1);
# RS = newregsing1F1sq(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# b = candidate for the lower parameter.
Condibiexp1F1sq:= proc(E1,E2,m,b,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(m*(1-b)-E2[1]),integer) and type(evala(-m*(1-b)-E2[2]),integer) and
                 type(evala(2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b)),integer) and
                 type(evala(E1[2][2]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(-m*(1-b)-E2[1]),integer) and type(evala(m*(1-b)-E2[2]),integer) and
                 type(evala(2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[1][2]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b)),integer) and
                 type(evala(E1[1][2]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(m*(1-b)-E2[1]),integer) and type(evala(m*(1-b)-E2[2]),integer) and
                 type(evala(2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b)),integer) and
                 type(evala(E1[2][1]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(m*(1-b)-E2[1]),integer) and type(evala(m*(1-b)-E2[2]),integer) and
                 type(evala(2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[1][2]-E1[2][1]),integer) and type(evala(E1[1][1]-2*m*(1-b)),integer) and
                 type(evala(E1[1][2]-m*(1-b)),integer)) then
                 return true
              elif (type(evala(-m*(1-b)-E2[1]),integer) and type(evala(m*(1-b)-E2[2]),integer) and
                 type(evala(-2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[1][1]-m*(1-b)),integer) and
                 type(evala(E1[1][2]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(m*(1-b)-E2[1]),integer) and type(evala(-m*(1-b)-E2[2]),integer) and
                 type(evala(-2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][1]-m*(1-b)),integer) and
                 type(evala(E1[2][2]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(-m*(1-b)-E2[1]),integer) and type(evala(-m*(1-b)-E2[2]),integer) and
                 type(evala(-2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][2]),integer) and
                 type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[1][1]-m*(1-b)),integer) and
                 type(evala(E1[1][2]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(-m*(1-b)-E2[1]),integer) and type(evala(-m*(1-b)-E2[2]),integer) and
                 type(evala(-2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][2]-m*(1-b)),integer) and
                 type(evala(E1[2][2]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(-m*(1-b)-E2[1]),integer) and type(evala(m*(1-b)-E2[2]),integer) and
                 type(evala(2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b)),integer) and
                 type(evala(E1[2][1]-2*m*(1-b)),integer)) then
                 return true
              elif (type(evala(m*(1-b)-E2[1]),integer) and type(evala(-m*(1-b)-E2[2]),integer) and
                 type(evala(2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-2*m*(1-b)),integer) and
                 type(evala(E1[1][2]-m*(1-b)),integer)) then
                 return true
              elif (type(evala(m*(1-b)-E2[1]),integer) and type(evala(-m*(1-b)-E2[2]),integer) and
                 type(evala(-2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-2*m*(1-b)),integer) and
                 type(evala(E1[1][2]-m*(1-b)),integer)) then
                 return true
              elif (type(evala(-m*(1-b)-E2[1]),integer) and type(evala(m*(1-b)-E2[2]),integer) and
                 type(evala(-2*m*(1-b)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][2]),integer) and type(evala(E1[1][2]-m*(1-b)),integer) and
                 type(evala(E1[2][1]-2*m*(1-b)),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

####### Let E be a list of two element [a,b]. We return a list [c,d] where c and d are the constant part of a and b,
####### respectively.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# i = index of an element in R[1][1];
# m = R[1][5][i];
# E = an element in R[1][6][i]. E is a list of two elements a and b (E=[a,b]).
helpCondiaibiexp1F1sq:= proc(E,t,m,ext)
              local Dx,x,a,i,c1,C;
              Dx,x:=op(_Envdiffopdomain);
              C:= map(s-> convert(series(subs(t=1/t^m,s),t=0,100),polynom),E);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              C:= map(s-> evala(coeff(s,t,0)),C);
              return C
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# k = index of an element in R[1][1];
# m1 = R[1][5][k];
# E2 = R[1][8][k];
# E1 = map(s-> helpCondiaibiexp1F1sq(s,t,m1,ext),R[1][6][k]);
# m = R[1][4][k].
Condiaibiexp1F1sq:= proc(E2,E1,m,b,a,ext)
           local Dx,x;
           Dx,x:=op(_Envdiffopdomain);
           if (type(evala(2*m*(b-2*a)-E2[3]),integer) and type(evala(m*(b-2*a)-E2[1]),integer) and
              type(evala(m*(b-2*a)+E2[2]),integer)) and type(evala(E1[1][1]-E1[2][1]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][2]-E1[3][1]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[2][2]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a)-E2[3]),integer) and type(evala(m*(b-2*a)-E2[1]),integer) and
              type(evala(m*(b-2*a)-E2[2]),integer)) and type(evala(E1[1][1]-E1[2][2]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[2][1]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a)+E2[3]),integer) and type(evala(m*(b-2*a)+E2[1]),integer) and
              type(evala(m*(b-2*a)-E2[2]),integer)) and type(evala(E1[1][1]-E1[2][1]),integer) and
              type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][2]-E1[3][2]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[1][2]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[2][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a)+E2[3]),integer) and type(evala(m*(b-2*a)+E2[1]),integer) and
              type(evala(m*(b-2*a)+E2[2]),integer)) and type(evala(E1[1][1]-E1[2][2]),integer) and
              type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-E1[3][2]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[1][2]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a)-E2[3]),integer) and type(evala(m*(b-2*a)+E2[1]),integer) and
              type(evala(m*(b-2*a)-E2[2]),integer)) and type(evala(E1[1][2]-E1[2][2]),integer) and
              type(evala(E1[1][1]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[2][1]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[1][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a)+E2[3]),integer) and type(evala(m*(b-2*a)-E2[1]),integer) and
              type(evala(m*(b-2*a)+E2[2]),integer)) and type(evala(E1[1][2]-E1[2][2]),integer) and
              type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-E1[3][2]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[1][1]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           else
              return false
           fi;
end:

############################################################################################################
############################################################################################################

# For a given candidates [[b,b,...],f], we find associated candidates for [{a,a,...,},[b],f]
# related an element of [b,b,...] and an irregular singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# F = R[1];
# C = Candidates for [[b1,b2],[b1,b2],[b1,b2]];
# i = index of an element in F[1].
helpcandai1F1sq:=proc(L,F,C,i,t,ext,ext1)
            local Dx,x,p,H,s,A,A1,A2,k,m,m1,E,E1,E2,j;
            Dx,x:=op(_Envdiffopdomain);
            m:= F[4][i];
            m1:= F[5][i];
            E2:= F[8][i];
            E1:= map(s-> helpCondiaibiexp1F1sq(s,t,m1,ext),F[6][i]);
            H:= table();
            for j in C do
               ####### candidates for "a" modulo an integer
               A:= [seq((j-(E2[1]+s)/m)/2,s=-2*m+1..2*m-1),
                    seq((j-(-E2[1]+s)/m)/2,s=-2*m+1..2*m-1)];
               A:=[op({op(A)})];
               ####### We make the rational part of "a" belong to ]0,1[ union {-1,1}
               A1:={op(map(s->Ratiopartmodint(s,ext1),A))};
               A1:=map(s->`if`(s[1]=0,op([[-1,-1,s[3],s[4]],[1,1,s[3],s[4]]]),s),A1);
               A2:= table();
               for k in A1 do
                  if Condiaibiexp1F1sq(E2,E1,m,j,k[1],ext)=true then
                     A2[k]:= k[1]
                  fi;
               od;
               A2:= convert(A2,'set');
               H[j]:= [A2,[j]];
            od;
            H:= convert(H,'list');
            return H
end:

############################################################################################################
############################################################################################################

# We compute the candiates for [{a,a...},b] related to a candidate of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# F = R[1];
# C = Candidates for [b,b,...].
findcandai1F1sq:=proc(L,F,C,x,t,ext,ext1)
                  local j,B,k,E,H;
                  H := table();
                  E:= helpcandai1F1sq(L,F,C,1,t,ext,ext1);
                  if E=NULL then
                     return NULL
                  fi;
                  if nops(F[1]) > 1 then
                     for j from 2 to nops(F[1]) do
                        B:= helpcandai1F1sq(L,F,C,j,t,ext,ext1);
                        if B=NULL then
                           return NULL
                        fi;
                        for k from 1 to nops(B) do
                           E[k][1]:= B[k][1] intersect E[k][1];
                        od;
                     od;
                  fi;
                  for j from 1 to nops(E) do
                     if E[j][1]<>{} then
                        H[j]:= E[j];
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
####################################   Logarithmic Case   ##################################################
############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a,b), also and f:[{a,a,...},b,f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extpp);
# R = IrrRegAppsing1F1sq(L,t,E,ext);
# Sreg = IrrRegAppsing1F1sq(L,t,E,ext)[2];
# F = Hyp1F1sqSubst(L,x,t,R[1],ext).
find1F1sqln:=proc(L,R,F,x,t,ext)
                  local ab,f,C,H,ext3,ext4,n2,B,i,s;
                  B:= Candichangvar1F1sq(F,R,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     C:=[[2],f[1],f[2]];
                     ab:= findcandai1F1sq(L,R[1],C[1],x,t,ext,ext3);
                     if ab <> NULL then
                        H[f]:= [op(ab),C[2]]
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##################################   RATIONAL AND IRRATIONAL CASE   ########################################
############################################################################################################
############################################################################################################

####### b11 is the sets of candidates for 1-b. We compute here the candidates
####### for b with its the rational part in ]0,1], related to a zero of the change of variable parameter.
#######
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,H,ext);
# f = candidate for the change of variable parameter;
# F = Zeroesmultordpol1F1sq(f,x);
# n = nops(F);
# RS = newregsing1F1sq(L,t,x,F,R[2],n,ext);
# j = index of an element in RS[1][1];
# m = multiplicity order of the j^th zero of f that we know;
# e = RS[1][3][j];
# a = RS[1][4][j].
casesbi1F1sq:= proc(b11,e,a,m,ext,ext1)
          local k,s,b1,B,B1,i;
          b1:=map(s->1-s,b11);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          ####### We compute the candidates for b such that their rational part are in [0,1[
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          ####### the condition for the non-logarithmic case
          B1:= map(s->`if`(s[1]=0,NULL,s),B1);
          if B1={} then
             return NULL
          fi;
          B:= table();
          for i in B1 do
             ####### the condition that a candidate for b must satisfy
             if Condibiexp1F1sq(a,e,m,i[1],ext)=true then
                B[i]:= {i[1]}
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameter [b] related to a zero of the change of variable parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,G,ext);
# F = Hyp1F1sqSubst(L,x,R[1],ext);
# B = Candichangvar1F1sq(F,R,ext) ;
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiff1F1sq:= proc(j,G,m,ext,ext1)
          local b11,C1,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          ####### b11 is the list of candidates for +-(1-b) modulo an integer
          b11:=[seq(G[3][j][1]+s,s=-m+1..m-1),seq(-G[3][j][1]+s,s=-m+1..m-1)]/m;
          C1:=casesbi1F1sq(b11,G[3][j],G[4][j],m,ext,ext1);
          if C1 = NULL then
             return NULL
          else
             return C1
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameter b related to a candidate of f: [[b,b,...],f,M], where M is the list
# of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,G,ext);
# F = Hyp2F2Subst(L,x,R[1],ext);
# B1 = Candichangvar1F1sq(F,R,ext);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter;
# f[2] list of multiplicity orders of zeroes of f[1];
# f[3][1] list of zeroes of the change of variable parameter.
findbi1F1sq:=proc(L,f,x,t,ext,ext1,n1)
                  local H,B,j,s,i;
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiff1F1sq(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> op(s),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a,b), also and f:[{a,a,...},b,f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,E,ext);
# F = Hyp1F1sqSubst(L,x,R[1],ext).
find1F1sqRatIrr:=proc(L,R,F,x,t,ext)
                  local Dx,B,ab,f,C,H,E,ext1,ext2,ext3,ext4,i,n1,n2,n,s,G1;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= Candichangvar1F1sq(F,R,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                  ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                  n1:=nops(R[2][1]);
                  if n1 > 1 then
                     for i from 2 to n1 do
                        ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                        ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                        ext1:= ext1 intersect ext2;
                     od;
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi1F1sq(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        ab:= findcandai1F1sq(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
######################################    INTEGER CASE    ##################################################
############################################################################################################
############################################################################################################

# Input: A monic polynomial f\in K[x] and a positive integer p\in\N;
# Output: g\in K[x], such that if there exists a solution to y^p=f, then y=g is such a solution.
ispower := proc(f,x,p)
           local d,n,A,Ap,a,i,b;
           if p=1 or degree(f,x)=0 then
                return f
           fi;
           d:=degree(f,x);
           n:=d/p;
           if not type(n,integer) then
                return NULL
           fi;
           A:=add(a[i]*x^i,i=0..n-1)+x^n;
           Ap:=collect(A^p,x);
           for i from 1 to n do
                # Solve a linear equation in one unknown:
                b[n-i]:=solve(subs({seq(a[n-j]=b[n-j],j=1..i-1)},coeff(Ap,x,d-i))-coeff(f,x,d-i),a[n-i] );
           od;
           Ap:=subs({seq(a[n-j]=b[n-j],j=1..n)},Ap);
           subs({seq(a[n-j]=b[n-j],j=1..i-1)},A)
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,E,ext);
# F = Hyp1F1sqSubst(L,x,t,R[1],ext).
find1F1sqInt:= proc(L,R,F,x,t,ext)
          local Dx,ff,f,g1,n,i,g,lc,a,k,V1,ext3,ext4,v,C,c,cc,fctrs,bf,s,inff,B1,n1,RS,v1,j,n2,ab,res,E;
           Dx,x:=op(_Envdiffopdomain);
           ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
           ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
           n2:=nops(R[1][1]);
           if n2 > 1 then
              for i from 2 to n2 do
                 ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                 ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                 ext3:= ext3 intersect ext4;
              od;
           fi;
           bf:=table();
           for f in F do
               ff:=Normalizer(f+C);
               inff:=map(s->s[1],R[1][1]);
               if member(infinity,{op(inff)}) then
                   n:=degree(numer(ff));
               else
                   n:=degree(denom(ff));
               fi;
               # denom(2*nu) divises n. So if n=1 then denom(2*nu)=1 and that means nu belong to 1/2+Z or
               # nu belong to Z which is not good. nu also have to be different from 0 (see the proof of
               # one of corollary in the thesis related to this part).
               if n<=1 then
                  next
               fi;
               g:=numer(ff);
               lc:=lcoeff(g,x);
               res:=table();
               for i from 2 to n do
                  # i must be a divisor of n
                  V1:= table();
                  if irem(n,i)=0 then
                     # f must be in the form g^i.
                     a:=ispower(collect(g/lc,x,Normalizer),x,i);
                     if a=NULL then
                        next
                     fi;
                     s:=numer(Normalizer(evala(Content(g-lc*a^i,x))));
                     s:=evala(Factors(s,ext));
                     cc:={seq(`if`(degree(k[1],C)=1,-coeff(k[1],C,0)/coeff(k[1],C,1),NULL),k=s[2])};
                     if member(infinity,{op(inff)})=false then
                        fctrs:=sqrfree(subs(C=0,g),x);
                        # the multiplicity of each factor of g is a multiple of i
                        if add(modp(k[2],i),k=fctrs[2])=0 then
                           cc:=cc union {0};
                        fi;
                     fi;
                     if cc <> {} then
                        for c in cc do
                           g:= factor(subs(C=c,ff));
                           B1:= Zeroesmultordpol1F1sq(g,x);
                           if B1=NULL then
                              next
                           fi;
                           n1:= nops(B1);
                           RS:= newregsing1F1sq(L,t,x,B1,R[2],n1,ext);
                           g1:= [g,RS[2],RS[1]];
                           n1:= nops(g1[3][1]);
                           E:= findbi1F1sq(L,g1,x,t,ext,{},n1);
                           if E<>NULL then
                              ab:= findcandai1F1sq(L,R[1],E[1],x,t,ext,ext3);
                              if ab <> NULL then
                                 V1[c]:= [op(ab),E[2]]
                              fi;
                           fi;
                        od;
                     fi;
                  fi;
                  V1:= convert(V1,'list');
                  if V1<>[] then
                     res[i]:= op(V1);
                  fi;
               od;
               res:= convert(res,'list');
               if res<>[] then
                  bf[f]:= op(res);
               fi;
            od;
            bf:= convert(bf,'list');
            if bf=[] then
               return NULL
            else
               return bf
            fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature 1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1, l2, p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F1sq(L,t,E,ext);
# F = Hyp1F1sqSubst(L,x,t,R[1],ext).
help1F1sqSolutions1:= proc(L,R,F,x,t,T,ext)
           local Dx,B,s,H;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H:= find1F1sqInt(L,R,F,x,t,ext);
           elif R[3][3]<>[] then
              H:= find1F1sqln(L,R,F,x,t,ext);
           else
              H:= find1F1sqRatIrr(L,R,F,x,t,ext);
           fi;
           if H=NULL then
              return NULL
           else
              H:= {op(map(s-> op([s,subs(s[-1]=-s[-1],s)]),H))};
              return [op(H)]
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsing1F1sq(L,t,G,ext);
# a = element of help1F1sqSolutions1(L,r,RSreg,info,ext,x).
help1F1sqSolutions2:= proc(L,a,A,x,t,ext)
           local Dx,n1,M,L11,L12,j,R,i,e1,e2,T;
           Dx,x:=op(_Envdiffopdomain);
           n1:= nops(a);
           L11:= MinOp(hypergeom([e1],[e2],x)^2);
           for i from 1 to n1-1 do
              for j in a[i][1] do
                 L12:= subs({e1=j,e2=op(a[i][2])},L11);
                 M:= ChangeOfVariables(L12,a[n1]);
                 R:=EquivExpgauge1F1sq(M,L,A,x,t,T,ext);
                 if R<>NULL then
                    return [{[[j],op(a[i][2]),R]},a[n1]]
                 fi;
              od;
           od;
           return NULL
end:

############################################################################################################
############################################################################################################

help1F1sqSolutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

help1F1sqSolutions4:= proc(a,b)
            local Dx,x,n1,n2;
            Dx,x:=op(_Envdiffopdomain);
            if a[3]<>b[3] then
               return [false,b]
            elif Matchlistmod1(a[1],b[1])=false or Matchlistmod1(a[2],b[2])=false then
               return [false,b]
            else
               n1:= length(a[4]);
               n2:= length(b[4]);
               if n2<n1 then
                  return [true,b]
               else
                  return [true,a]
               fi;
            fi;
end:

############################################################################################################
############################################################################################################

help1F1sqSolutions5:= proc(G,b)
            local Dx,x,n1,n2,n,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n1:= nops(G);
            n2:= 0;
            for i from 1 to n1 do
               E:= help1F1sqSolutions4(G[i],b);
               if E[1]=false then
                  n2:= n2+1
               else
                  n:= i;
                  break
               fi;
            od;
            if n2=n1 then
              return [false,b]
            else
               return [true,[E[2],n]]
            fi;
end:

############################################################################################################
############################################################################################################

help1F1sqSolutions6:= proc(H)
            local Dx,x,n,G,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n:= nops(H);
            G:=table();
            G[1]:= H[1];
            if n>1 then
               for i from 2 to n do
                  E:= help1F1sqSolutions5(convert(G,'list'),H[i]);
                  if E[1]=true then
                     G[E[2][2]]:= E[2][1]
                  else
                     G[i]:= E[2]
                  fi;
               od;
            fi;
            G:= convert(G,'list');
            return G
end:

############################################################################################################
############################################################################################################

#
Find1F1sqSolutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsing1F1sq(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= Hyp1F1sqSubst(L,x,t,R[1],ext);
           if F=NULL then
              return NULL
           fi;
           C:= table();
           H:= help1F1sqSolutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= help1F1sqSolutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

Hyp1F1sqSolutions:= proc(L)
           local Dx,x,s,k,F,E,H1,H2,G,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:= table();
           H1:= Find1F1sqSolutions(L);
           if H1=NULL then
              return "No 1F1 square type solutions"
           fi;
           F:= table();
           G:= map(s->s[2],H1);
           for i in G do
              H2:= [[op(map(s->`if`(s[2]=i,op(s[1]),NULL),H1))],i] ;
              F[i]:= H2
           od;
           F:= convert(F,'set');
           for C in F do
              E:= map(s->subs(s[3]=map(k->help1F1sqSolutions3(k),s[3]),s),C[1]);
              E:= map(s->seq([s[1],s[2],op(s[3][i])],i=1..nops(s[3])),E);
              E:= help1F1sqSolutions6(E);
              H[C]:= [E,C[2]]
           od;
           H:= convert(H,'set');
           return H
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
################################### DIFFERENTIAL OF 1F1 SQUARE TYPE SOLUTIONS ##############################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingdiff1F1sq(L2,t,G,ext);       ####### L2 can be our L operator.
# C1 = Exptestdiff1F1sqequiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator.
# i = index of an element in C2[5][1];     ####### C2[5][1] = set of non-apparent singularities of L2.
towopegenexpdiffdiff1F1sq:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equivadiff1F1sq2(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equivadiff1F1sq1(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsingdiff1F1sq(L2,t,G,ext);       ####### L2 can be our L operator.
Exptestdiff1F1sqequiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= Condidiff1F1sqirr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingdiff1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptestdiff1F1sqequiv2(L1,C2,x,t,ext);       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
Exptestdiff1F1sqequiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingdiff1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptestdiff1F1sqequiv2(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equivadiff1F1sq1:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               if extp0 subset ext then
                  return extp3 union ext
               fi;
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equivadiff1F1sq2:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingdiff1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptestdiff1F1sqequiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiffdiff1F1sq(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgaugediff1F1sq1:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,G,ext).
EquivExpgaugediff1F1sq1:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptestdiff1F1sqequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiffdiff1F1sq(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgaugediff1F1sq1(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsingdiff1F1sq(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptestdiff1F1sqequiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiffdiff1F1sq(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgaugediff1F1sq:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=Exptestdiff1F1sqequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(Exptestdiff1F1sqequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,G,ext).
EquivExpgaugediff1F1sq:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptestdiff1F1sqequiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiffdiff1F1sq(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgaugediff1F1sq(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the (1F1^2)' case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol.
# E = genexpdiff(L,i,t,ext)
Condidiff1F1sqirr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l;
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 3 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              if l <> 1 then
                 return NULL
              fi;
              C:= map(s-> convert(series(subs(t=1/t,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              if (evala(A[1]-A[2])=0 or evala(A[1]+A[2])=0) and (evala(A[3]-2*A[1])=0 or evala(A[3]+2*A[1])=0) then
                 return [E[2],E[4],A,B]
              elif (evala(A[1]-A[3])=0 or evala(A[1]+A[3])=0) and (evala(A[2]-2*A[1])=0 or evala(A[2]+2*A[1])=0) then
                    return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]],[A[1],A[3],A[2]],[B[1],B[3],B[2]]]
              elif (evala(A[2]-A[3])=0 or evala(A[2]+A[3])=0) and (evala(A[1]-2*A[2])=0 or evala(A[1]+2*A[2])=0) then
                    return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]],[A[2],A[3],A[1]],[B[2],B[3],B[1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm is "irrsingdiff1F1sq".
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext).
irrsingdiff1F1sq:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or
                    member(t,indets(A[i[1]][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d:= Condidiff1F1sqirr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the (1F1^2)' case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condidiff1F1sqreg:=proc(E,t,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if evala(E[2][1]+E[2][2])=0 and evala(E[2][3]+2*E[2][1])=0 and evala(E[2][3]-2*E[2][2])=0 and
                 evala(E[2][1]-E[2][2]+E[2][3])=0 then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif evala(E[2][1]-E[2][3])=0 and evala(E[2][2]-2*E[2][1])=0 and evala(E[2][2]-2*E[2][3])=0 and
                 evala(E[2][1]+E[2][3]-E[2][2])=0 then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif evala(E[2][2]+E[2][3])=0 and evala(E[2][1]-2*E[2][2])=0 and evala(E[2][1]+2*E[2][3])=0 and
                 evala(E[2][2]-E[2][3]-E[2][1])=0 then
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              elif type(evala(E[2][1]+E[2][2]),integer)  and type(evala(E[2][3]+2*E[2][1]),integer) and
                 type(evala(E[2][3]-2*E[2][2]),integer) and type(evala(E[2][1]-E[2][2]+E[2][3]),integer) then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif type(evala(E[2][1]-E[2][3]),integer) and type(evala(E[2][2]-2*E[2][1]),integer) and
                 type(evala(E[2][2]-2*E[2][3]),integer) and type(evala(E[2][1]+E[2][3]-E[2][2]),integer) then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              elif type(evala(E[2][2]+E[2][3]),integer) and type(evala(E[2][1]-2*E[2][2]),integer) and
                 type(evala(E[2][1]+2*E[2][3]),integer) and type(evala(E[2][2]-E[2][3]-E[2][1]),integer) then
                 return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H1 = NotAppSing(L,G,ext);
# H = irrsingdiff1F1sq(L,t,H1,ext)[-1].
regsingtruediff1F1sq:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                 # must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= Condidiff1F1sqreg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1];
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in three parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer);
#       - removable singularities with integer exponent differences;
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingdiff1F1sq(L,t,F,ext)[-1];
# H = regsingtruediff1F1sq(L,t,R,ext).
Sregseptruediff1F1sq:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# R = irrsingdiff1F1sq(L,t,F,ext)[-1];
# Sreg = regsingtruediff1F1sq(L,t,R,ext);
# RSreg = Sregseptruediff1F1sq(L,Sreg,ext).
UnionSregseptruediff1F1sq:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# G = Singular(L,ext);
# F = NotAppSing(L,G,ext);
# Sreg = regsingtruediff1F1sq(L,t,F,ext);
# Sirr = irrsingdiff1F1sq(L,t,F,ext).
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp).
IrrRegAppsingdiff1F1sq:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,
                    H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or
                    member(t,indets(A[i][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d1:= Condidiff1F1sqirr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                    # must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # generalized exponents have to be positive integers.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= Condidiff1F1sqreg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1];
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31];
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31];
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= Condidiff1F1sqreg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1];
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31];
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31];
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsing1F1sq(L,t,F,ext)[1][i],x),
#######                      with i=position of the minimal polynom of p) is 1.
############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g;
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsing2F2(L,t,ext).
Hyperpossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:= table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     C[k]:= h;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g;
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Bespossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:=table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     if member(-h,{op(C)})=false and member(h,{op(C)})=false then
                           C[k]:= h;
                     fi;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# G = Singular(L,ext);
# F = NotAppSing(L,G,ext);
# Sirr = irrsingdiff1F1sq(L,t,F,ext);
Hypdiff1F1sqSubst:=proc(L,x,t,Sirr,ext)
               local Dx,m,A,ff,B,extp0,extp1,extp2,extp3,extp4,extp5,v,v1,i,j,s;
               Dx,x:=op(_Envdiffopdomain);
               A := table();
               i:=0;
               for s in Sirr[1] do
                  i:=i+1;
                  m:= ldegree(ReplirrRoot(Sirr[3][i][1],{}),t);
                  if m=FAIL then
                     return NULL
                  fi;
                  ff := add( (coeff(convert(eval(subs(csgn=1,simplify(ReplirrRoot(Sirr[3][i][1],{})) assuming(t>0))),                                                   polynom),t,j)/j)*(`if`(s[1]=infinity,1/x,x-s[2]))^j,j=m .. -1);
                  ff:=evala(Reduce(ff));
                  if s[1]=infinity then
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All the names or RootOf
                                           # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp5:= ext union extp1
                  else
                     extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                     extp1:= indets(evala(Sirr[3][i][1]),{RootOf,name}) minus {t};
                     extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                     extp2:= extp1 minus extp0; # All the names or RootOf
                                                 # different from RootOf(s,x) in the expression of Sirr[3][i][1].
                     extp3:= ext minus extp0;
                     extp4:= extp2 union extp3;
                     v1:={};
                     for v in extp4 do
                        if extp0 intersect indets(v,{RootOf,name}) <> {} then
                           v1:= v1 union {v}
                        fi;
                     od:
                     extp5:= extp4 minus v1;
                  fi;
                  A[s]:= evala(Trace(convert(evala(ff),polynom),extp5,extp5),x)
               od;
               A:= convert(A,'list');
               B := Hyperpossibility(A);
               B:= map(j-> `if`(member(x,indets(j,{name})),j,NULL),B);
               if B=[] then
                  return NULL
               fi;
               return map(s-> Normalizer(s),B)
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
Zeroesmultordpoldiff1F1sq:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsingdiff1F1sq(L,t,H,ext)[2];
# B = Zeroesmultordpoldiff1F1sq(f,x);
# n = nops(B).
newregsingdiff1F1sq:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],Condidiff1F1sqreg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],Condidiff1F1sqreg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################

# Input: a function f(x) and a minimal polynomial i. The minimal polynomial with Root infinity is infinity;
# Output: f(x)+c such that f(p)+c=0 with p=RootOf(i,x). If i=infinity then f(x) is returned i.e c=0.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# Sirr = R[1];
# f = element of Hypdiff1F1sqSubst(L,x,Sirr,ext);
# i = element of R[2][1].
changeconstantdiff1F1sq:= proc(f,x,i,ext)
                 local Dx,c,A,ff,C;
                 Dx,x:=op(_Envdiffopdomain);
                 if i[1]=infinity then
                    A:= f;
                 else
                    ff:=numer(Normalizer(f+C));
                    c:=SolveTools:-Linear({coeffs(convert(collect(eval(subs(csgn=1,simplify(evala(Rem(ff,i[1],x)))
                                                               assuming(x>0))),x),polynom),x)},{C});
                    if c=NULL or nops(c) <> 1 then
                       return NULL;
                    else
                       A:=Normalizer(f+subs(c[1],C));
                    fi;
                 fi;
                 return factor(A);
end:

############################################################################################################
############################################################################################################

# Give the good candidates for f by looking at those candidates which vanish at elements of regsingtrue (all
# elements of regsingtrue should give the same constant c). Note that regsingtrue <> {} in this case.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# Sirr = R[1];
# F = Hypdiff1F1sqSubst(L,x,Sirr,ext).
Candichangvardiff1F1sq:=proc(F,R,ext) ### For all the case except the integer case.
               local Dx,x,i,f,A,j,s,B1,n1,SR,n;
               Dx,x:=op(_Envdiffopdomain);
               A:= table();
               n:=nops(R[2][1]);
               for s in F do
                  j:=0;
                  f:= changeconstantdiff1F1sq(s,x,R[2][1][1],ext);
                  if f=NULL then
                     next
                  fi;
                  j:=j+1;
                  if n>1 then
                     for i from 2 to n do
                        if R[2][1][i][1]=infinity then
                           if not degree(numer(f),x)<degree(denom(f),x) then
                              break
                           fi;
                           j:=j+1;
                        else
                           if not evala(Rem(numer(f),R[2][1][i][1],x))=0 then
                              break
                           fi;
                           j:=j+1;
                        fi;
                     od;
                  fi;
                  if j <> n then
                     next
                  fi;
                  B1:= Zeroesmultordpoldiff1F1sq(f,x);
                  if B1=NULL then
                     next
                  fi;
                  n1:= nops(B1);
                  SR:= newregsingdiff1F1sq(L,t,x,B1,R[2],n1,ext);
                  A[s] := [f,SR[2],SR[1]];
               od;
               A:= convert(A,'set');
               if A={} then
                  return NULL
               else
                  return A
               fi;
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate b for the lower parameter in (1F1^2)' must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = Zeroesmultordpoldiff1F1sq(f,x);
# n = nops(B1);
# RS = newregsingdiff1F1sq(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# b = candidate for the lower parameter.
Condibiexpdiff1F1sq:= proc(E1,E2,m,b,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(-m*b-E2[1]),integer) and type(evala(-m*b-E2[2]),integer) and
                 type(evala(-2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]+m*b),integer) and
                 type(evala(E1[2][1]+2*m*b),integer)) then
                 return true
              elif (type(evala(-m*b-E2[1]),integer) and type(evala(-m*b-E2[2]),integer) and
                 type(evala(-2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[1][2]-E1[2][1]),integer) and type(evala(E1[1][1]+2*m*b),integer) and
                 type(evala(E1[1][2]+m*b),integer)) then
                 return true
              elif (type(evala(-m*b-E2[1]),integer) and type(evala(m*b-E2[2]),integer) and
                 type(evala(-2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][1]+m*b),integer) and
                 type(evala(E1[2][2]+2*m*b),integer)) then
                 return true
              elif (type(evala(m*b-E2[1]),integer) and type(evala(-m*b-E2[2]),integer) and
                 type(evala(-2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[1][2]-E1[3][1]),integer) and type(evala(E1[1][1]+m*b),integer) and
                 type(evala(E1[1][2]+2*m*b),integer)) then
                 return true
              elif (type(evala(m*b-E2[1]),integer) and type(evala(-m*b-E2[2]),integer) and
                 type(evala(-2*m*b-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]+m*b),integer) and
                 type(evala(E1[2][1]+2*m*b),integer)) then
                 return true
              elif (type(evala(-m*b-E2[1]),integer) and type(evala(m*b-E2[2]),integer) and
                 type(evala(-2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][1]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]+2*m*b),integer) and
                 type(evala(E1[1][2]+m*b),integer)) then
                 return true
              elif (type(evala(-m*b-E2[1]),integer) and type(evala(m*b-E2[2]),integer) and
                 type(evala(2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]+2*m*b),integer) and
                 type(evala(E1[1][2]+m*b),integer)) then
                 return true
              elif (type(evala(m*b-E2[1]),integer) and type(evala(-m*b-E2[2]),integer) and
                 type(evala(2*m*b-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][2]),integer) and type(evala(E1[1][2]+m*b),integer) and
                 type(evala(E1[2][1]+2*m*b),integer)) then
                 return true
              elif (type(evala(m*b-E2[1]),integer) and type(evala(-m*b-E2[2]),integer) and
                 type(evala(2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[1][1]+m*b),integer) and
                 type(evala(E1[1][2]+2*m*b),integer)) then
                 return true
              elif (type(evala(-m*b-E2[1]),integer) and type(evala(m*b-E2[2]),integer) and
                 type(evala(2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][1]+m*b),integer) and
                 type(evala(E1[2][2]+2*m*b),integer)) then
                 return true
              elif (type(evala(m*b-E2[1]),integer) and type(evala(m*b-E2[2]),integer) and
                 type(evala(2*m*b-E2[3]),integer)) and (type(evala(E1[1][1]-E1[2][2]),integer) and
                 type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[1][1]+m*b),integer) and
                 type(evala(E1[1][2]+2*m*b),integer)) then
                 return true
              elif (type(evala(m*b-E2[1]),integer) and type(evala(m*b-E2[2]),integer) and
                 type(evala(2*m*b-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][1]),integer) and
                 type(evala(E1[2][2]-E1[3][2]),integer) and type(evala(E1[1][2]+m*b),integer) and
                 type(evala(E1[2][2]+2*m*b),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

####### Let E be a list of two element [a,b]. We return a list [c,d] where c and d are the constant part of a and b,
####### respectively.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# i = index of an element in R[1][1];
# m = R[1][5][i];
# E = an element in R[1][6][i]. E is a list of two elements a and b (E=[a,b]).
helpCondiaibiexpdiff1F1sq:= proc(E,t,m,ext)
              local Dx,x,a,i,c1,C;
              Dx,x:=op(_Envdiffopdomain);
              C:= map(s-> convert(series(subs(t=1/t^m,s),t=0,100),polynom),E);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              C:= map(s-> evala(coeff(s,t,0)),C);
              return C
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# k = index of an element in R[1][1];
# m1 = R[1][5][k];
# E2 = R[1][8][k];
# E1 = map(s-> helpCondiaibiexpdiff1F1sq(s,t,m1,ext),R[1][6][k]);
# m = R[1][4][k].
Condiaibiexpdiff1F1sq:= proc(E2,E1,m,b,a,ext)
           local Dx,x;
           Dx,x:=op(_Envdiffopdomain);
           if (type(evala(2*m*(b-2*a-1)-E2[3]),integer) and type(evala(m*(b-2*a-1)-E2[1]),integer) and
              type(evala(m*(b-2*a-1)-E2[2]),integer)) and type(evala(E1[1][1]-E1[2][2]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[2][1]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a-1)-E2[3]),integer) and type(evala(m*(b-2*a-1)-E2[1]),integer) and
              type(evala(m*(b-2*a-1)+E2[2]),integer)) and type(evala(E1[1][1]-E1[2][1]),integer) and
              type(evala(E1[1][2]-E1[3][2]),integer) and type(evala(E1[2][2]-E1[3][1]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[2][2]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a-1)-E2[3]),integer) and type(evala(m*(b-2*a-1)+E2[1]),integer) and
              type(evala(m*(b-2*a-1)-E2[2]),integer)) and type(evala(E1[1][2]-E1[2][2]),integer) and
              type(evala(E1[1][1]-E1[3][2]),integer) and type(evala(E1[2][1]-E1[3][1]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[2][1]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[1][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a-1)+E2[3]),integer) and type(evala(m*(b-2*a-1)-E2[1]),integer) and
              type(evala(m*(b-2*a-1)+E2[2]),integer)) and type(evala(E1[1][2]-E1[2][2]),integer) and
              type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-E1[3][2]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[1][1]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a-1)+E2[3]),integer) and type(evala(m*(b-2*a-1)+E2[1]),integer) and
              type(evala(m*(b-2*a-1)-E2[2]),integer)) and type(evala(E1[1][1]-E1[2][1]),integer) and
              type(evala(E1[2][2]-E1[3][1]),integer) and type(evala(E1[1][2]-E1[3][2]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[1][2]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[2][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif (type(evala(2*m*(b-2*a-1)+E2[3]),integer) and type(evala(m*(b-2*a-1)+E2[1]),integer) and
              type(evala(m*(b-2*a-1)+E2[2]),integer)) and type(evala(E1[1][1]-E1[2][2]),integer) and
              type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-E1[3][2]),integer) then
              if indets(a,{RootOf,name}) subset indets(evala(E1[1][2]-2*m*b),{RootOf,name}) and
                 indets(a,{RootOf,name}) subset indets(evala(E1[2][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           else
              return false
           fi;
end:

############################################################################################################
############################################################################################################

# For a given candidates [[b,b,...],f], we find associated candidates for [{a,a,...,},[b],f]
# related an element of [b,b,...] and an irregular singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# F = R[1];
# C = Candidates for [[b1,b2],[b1,b2],[b1,b2]];
# i = index of an element in F[1].
helpcandaidiff1F1sq:=proc(L,F,C,i,t,ext,ext1)
            local Dx,x,p,H,s,A,A1,A2,k,m,m1,E,E1,E2,j;
            Dx,x:=op(_Envdiffopdomain);
            m:= F[4][i];
            m1:= F[5][i];
            E2:= F[8][i];
            E1:= map(s-> helpCondiaibiexpdiff1F1sq(s,t,m1,ext),F[6][i]);
            H:= table();
            for j in C do
               ####### candidates for "a" modulo an integer
               A:= [seq((j-1-(E2[2]+s)/m)/2,s=-2*m+1..2*m-1),
                    seq((j-1-(-E2[2]+s)/m)/2,s=-2*m+1..2*m-1)];
               A:=[op({op(A)})];
               ####### We make the rational part of "a" belong to ]0,1[ union {-1,1}
               A1:={op(map(s->Ratiopartmodint(s,ext1),A))};
               A1:=map(s->`if`(s[1]=0,op([[-1,-1,s[3],s[4]],[1,1,s[3],s[4]]]),s),A1);
               A2:= table();
               for k in A1 do
                  if Condiaibiexpdiff1F1sq(E2,E1,m,j,k[1],ext)=true then
                     A2[k]:= k[1]
                  fi;
               od;
               A2:= convert(A2,'set');
               H[j]:= [A2,[j]];
            od;
            H:= convert(H,'list');
            return H
end:

############################################################################################################
############################################################################################################

# We compute the candiates for [{a,a...},b] related to a candidate of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# F = R[1];
# C = Candidates for [b,b,...].
findcandaidiff1F1sq:=proc(L,F,C,x,t,ext,ext1)
                  local j,B,k,E,H;
                  H := table();
                  E:= helpcandaidiff1F1sq(L,F,C,1,t,ext,ext1);
                  if E=NULL then
                     return NULL
                  fi;
                  if nops(F[1]) > 1 then
                     for j from 2 to nops(F[1]) do
                        B:= helpcandaidiff1F1sq(L,F,C,j,t,ext,ext1);
                        if B=NULL then
                           return NULL
                        fi;
                        for k from 1 to nops(B) do
                           E[k][1]:= B[k][1] intersect E[k][1];
                        od;
                     od;
                  fi;
                  for j from 1 to nops(E) do
                     if E[j][1]<>{} then
                        H[j]:= E[j];
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
####################################   Logarithmic Case   ##################################################
############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a,b), also and f:[{a,a,...},b,f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,ext);
# R = IrrRegAppsingdiff1F1sq(L,t,E,ext);
# F = Hypdiff1F1sqSubst(L,x,t,R[1],ext).
finddiff1F1sqln:=proc(L,R,F,x,t,ext)
                  local ab,f,C,H,ext3,ext4,n2,B,i,s;
                  B:= Candichangvardiff1F1sq(F,R,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     C:=[[1],f[1],f[2]];
                     ab:= findcandaidiff1F1sq(L,R[1],C[1],x,t,ext,ext3);
                     if ab <> NULL then
                        H[f]:= [op(ab),C[2]]
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##################################   RATIONAL AND IRRATIONAL CASE   ########################################
############################################################################################################
############################################################################################################

####### b11 is the sets of candidates for -b. We compute here the candidates
####### for b with its the rational part in ]0,1], related to a zero of the change of variable parameter.
#######
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,H,ext);
# f = candidate for the change of variable parameter;
# F = Zeroesmultordpoldiff1F1sq(f,x);
# n = nops(F);
# RS = newregsingdiff1F1sq(L,t,x,F,R[2],n,ext);
# j = index of an element in RS[1][1];
# m = multiplicity order of the j^th zero of f that we know;
# e = RS[1][3][j];
# a = RS[1][4][j].
casesbidiff1F1sq:= proc(b11,e,a,m,ext,ext1)
          local k,s,b1,B,B1,i;
          b1:=map(s->-s,b11);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          ####### We compute the candidates for b such that their rational part are in [0,1[
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          ####### the condition for the non-logarithmic case
          B1:= map(s->`if`(s[1]=0,NULL,s),B1);
          if B1={} then
             return NULL
          fi;
          B:= table();
          for i in B1 do
             ####### the condition that a candidate for b must satisfy
             if Condibiexpdiff1F1sq(a,e,m,i[1],ext)=true then
                B[i]:= {i[1]}
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameter [b] related to a zero of the change of variable parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,G,ext);
# F = Hypdiff1F1sqSubst(L,x,R[1],ext);
# B = Candichangvardiff1F1sq(F,R,ext) ;
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiffdiff1F1sq:= proc(j,G,m,ext,ext1)
          local b11,C1,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          ####### b11 is the list of candidates for +-(1-b) modulo an integer
          b11:=[seq(G[3][j][1]+s,s=-m+1..m-1),seq(-G[3][j][1]+s,s=-m+1..m-1)]/m;
          C1:=casesbidiff1F1sq(b11,G[3][j],G[4][j],m,ext,ext1);
          if C1 = NULL then
             return NULL
          else
             return C1
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameter b related to a candidate of f: [[b,b,...],f,M], where M is the list
# of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,G,ext);
# F = Hypdiff1F1sqSubst(L,x,R[1],ext);
# B1 = Candichangvardiff1F1sq(F,R,ext) ;
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter;
# f[2] list of multiplicity orders of zeroes of f[1];
# f[3][1] list of zeroes of the change of variable parameter.
findbidiff1F1sq:=proc(L,f,x,t,ext,ext1,n1)
                  local H,B,j,s;
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiffdiff1F1sq(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> op(s),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a,b), also and f:[{a,a,...},b,f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,E,ext);
# F = Hypdiff1F1sqSubst(L,x,R[1],ext).
finddiff1F1sqRatIrr:=proc(L,R,F,x,t,ext)
                  local Dx,B,ab,f,C,H,E,ext1,ext2,ext3,ext4,i,n1,n2,n,s,G1;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= Candichangvardiff1F1sq(F,R,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                  ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                  n1:=nops(R[2][1]);
                  if n1 > 1 then
                     for i from 2 to n1 do
                        ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                        ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                        ext1:= ext1 intersect ext2;
                     od;
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbidiff1F1sq(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        ab:= findcandaidiff1F1sq(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
######################################    INTEGER CASE    ##################################################
############################################################################################################
############################################################################################################

# Input: A monic polynomial f\in K[x] and a positive integer p\in\N;
# Output: g\in K[x], such that if there exists a solution to y^p=f, then y=g is such a solution.
ispower := proc(f,x,p)
           local d,n,A,Ap,a,i,b;
           if p=1 or degree(f,x)=0 then
                return f
           fi;
           d:=degree(f,x);
           n:=d/p;
           if not type(n,integer) then
                return NULL
           fi;
           A:=add(a[i]*x^i,i=0..n-1)+x^n;
           Ap:=collect(A^p,x);
           for i from 1 to n do
                # Solve a linear equation in one unknown:
                b[n-i]:=solve(subs({seq(a[n-j]=b[n-j],j=1..i-1)},coeff(Ap,x,d-i))-coeff(f,x,d-i),a[n-i] );
           od;
           Ap:=subs({seq(a[n-j]=b[n-j],j=1..n)},Ap);
           subs({seq(a[n-j]=b[n-j],j=1..i-1)},A)
end:

############################################################################################################
############################################################################################################

#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,E,ext);
# F = Hypdiff1F1sqSubst(L,x,t,R[1],ext).
finddiff1F1sqInt:= proc(L,R,F,x,t,ext)
          local Dx,ff,f,g1,n,i,g,lc,a,k,V1,ext3,ext4,v,C,c,cc,fctrs,bf,s,inff,B1,n1,RS,v1,j,n2,ab,res,E;
           Dx,x:=op(_Envdiffopdomain);
           ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
           ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
           n2:=nops(R[1][1]);
           if n2 > 1 then
              for i from 2 to n2 do
                 ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                 ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                 ext3:= ext3 intersect ext4;
              od;
           fi;
           bf:=table();
           for f in F do
               ff:=Normalizer(f+C);
               inff:=map(s->s[1],R[1][1]);
               if member(infinity,{op(inff)}) then
                   n:=degree(numer(ff));
               else
                   n:=degree(denom(ff));
               fi;
               # denom(2*nu) divises n. So if n=1 then denom(2*nu)=1 and that means nu belong to 1/2+Z or
               # nu belong to Z which is not good. nu also have to be different from 0 (see the proof of
               # one of corollary in the thesis related to this part).
               if n<=1 then
                  next
               fi;
               g:=numer(ff);
               lc:=lcoeff(g,x);
               res:=table();
               for i from 2 to n do
                  # i must be a divisor of n
                  V1:= table();
                  if irem(n,i)=0 then
                     # f must be in the form g^i.
                     a:=ispower(collect(g/lc,x,Normalizer),x,i);
                     if a=NULL then
                        next
                     fi;
                     s:=numer(Normalizer(evala(Content(g-lc*a^i,x))));
                     s:=evala(Factors(s,ext));
                     cc:={seq(`if`(degree(k[1],C)=1,-coeff(k[1],C,0)/coeff(k[1],C,1),NULL),k=s[2])};
                     if member(infinity,{op(inff)})=false then
                        fctrs:=sqrfree(subs(C=0,g),x);
                        # the multiplicity of each factor of g is a multiple of i
                        if add(modp(k[2],i),k=fctrs[2])=0 then
                           cc:=cc union {0};
                        fi;
                     fi;
                     if cc <> {} then
                        for c in cc do
                           g:= factor(subs(C=c,ff));
                           B1:= Zeroesmultordpoldiff1F1sq(g,x);
                           if B1=NULL then
                              next
                           fi;
                           n1:= nops(B1);
                           RS:= newregsingdiff1F1sq(L,t,x,B1,R[2],n1,ext);
                           g1:= [g,RS[2],RS[1]];
                           n1:= nops(g1[3][1]);
                           E:= findbidiff1F1sq(L,g1,x,t,ext,{},n1);
                           if E<>NULL then
                              ab:= findcandaidiff1F1sq(L,R[1],E[1],x,t,ext,ext3);
                              if ab <> NULL then
                                 V1[c]:= [op(ab),E[2]]
                              fi;
                           fi;
                        od;
                     fi;
                  fi;
                  V1:= convert(V1,'list');
                  if V1<>[] then
                     res[i]:= op(V1);
                  fi;
               od;
               res:= convert(res,'list');
               if res<>[] then
                  bf[f]:= op(res);
               fi;
            od;
            bf:= convert(bf,'list');
            if bf=[] then
               return NULL
            else
               return bf
            fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature 1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1, l2, p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsingdiff1F1sq(L,t,E,ext);
# F = Hypdiff1F1sqSubst(L,x,t,R[1],ext).
helpdiff1F1sqSolutions1:= proc(L,R,F,x,t,T,ext)
           local Dx,B,s,H;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H:= finddiff1F1sqInt(L,R,F,x,t,ext);
           elif R[3][3]<>[] then
              H:= finddiff1F1sqln(L,R,F,x,t,ext);
           else
              H:= finddiff1F1sqRatIrr(L,R,F,x,t,ext);
           fi;
           if H=NULL then
              return NULL
           else
              H:= {op(map(s-> op([s,subs(s[-1]=-s[-1],s)]),H))};
              return [op(H)]
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsingdiff1F1sq(L,t,G,ext);
# a = element of helpdiff1F1sqSolutions1(L,r,RSreg,info,ext,x).
helpdiff1F1sqSolutions2:= proc(L,a,A,x,t,ext)
           local Dx,n1,M,L11,L12,j,R,i,e1,e2,T;
           Dx,x:=op(_Envdiffopdomain);
           n1:= nops(a);
           L11:= MinOp(diff(hypergeom([e1],[e2],x),x)^2);
           for i from 1 to n1-1 do
              for j in a[i][1] do
                 L12:= subs({e1=j,e2=op(a[i][2])},L11);
                 M:= ChangeOfVariables(L12,a[n1]);
                 R:=EquivExpgaugediff1F1sq(M,L,A,x,t,T,ext);
                 if R<>NULL then
                    return [{[[j],op(a[i][2]),R]},a[n1]]
                 fi;
              od;
           od;
           return NULL
end:

############################################################################################################
############################################################################################################

helpdiff1F1sqSolutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

helpdiff1F1sqSolutions4:= proc(a,b)
            local Dx,x,n1,n2;
            Dx,x:=op(_Envdiffopdomain);
            if a[3]<>b[3] then
               return [false,b]
            elif Matchlistmod1(a[1],b[1])=false or Matchlistmod1(a[2],b[2])=false then
               return [false,b]
            else
               n1:= length(a[4]);
               n2:= length(b[4]);
               if n2<n1 then
                  return [true,b]
               else
                  return [true,a]
               fi;
            fi;
end:

############################################################################################################
############################################################################################################

helpdiff1F1sqSolutions5:= proc(G,b)
            local Dx,x,n1,n2,n,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n1:= nops(G);
            n2:= 0;
            for i from 1 to n1 do
               E:= helpdiff1F1sqSolutions4(G[i],b);
               if E[1]=false then
                  n2:= n2+1
               else
                  n:= i;
                  break
               fi;
            od;
            if n2=n1 then
              return [false,b]
            else
               return [true,[E[2],n]]
            fi;
end:

############################################################################################################
############################################################################################################

helpdiff1F1sqSolutions6:= proc(H)
            local Dx,x,n,G,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n:= nops(H);
            G:=table();
            G[1]:= H[1];
            if n>1 then
               for i from 2 to n do
                  E:= helpdiff1F1sqSolutions5(convert(G,'list'),H[i]);
                  if E[1]=true then
                     G[E[2][2]]:= E[2][1]
                  else
                     G[i]:= E[2]
                  fi;
               od;
            fi;
            G:= convert(G,'list');
            return G
end:

############################################################################################################
############################################################################################################

#
Finddiff1F1sqSolutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsingdiff1F1sq(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= Hypdiff1F1sqSubst(L,x,t,R[1],ext);
           if F=NULL then
              return NULL
           fi;
           C:= table();
           H:= helpdiff1F1sqSolutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= helpdiff1F1sqSolutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

Hypdiff1F1sqSolutions:= proc(L)
           local Dx,x,s,k,F,E,H1,H2,G,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:= table();
           H1:= Finddiff1F1sqSolutions(L);
           if H1=NULL then
              return "No differential of 1F1 square type solutions"
           fi;
           F:= table();
           G:= map(s->s[2],H1);
           for i in G do
              H2:= [[op(map(s->`if`(s[2]=i,op(s[1]),NULL),H1))],i] ;
              F[i]:= H2
           od;
           F:= convert(F,'set');
           for C in F do
              E:= map(s->subs(s[3]=map(k->helpdiff1F1sqSolutions3(k),s[3]),s),C[1]);
              E:= map(s->seq([s[1],s[2],op(s[3][i])],i=1..nops(s[3])),E);
              E:= helpdiff1F1sqSolutions6(E);
              H[C]:= [E,C[2]]
           od;
           H:= convert(H,'set');
           return H
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
######################################## 1F2 TYPE SOLUTIONS ###########################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F2equiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# i = index of an element in C2[5][1];     ####### C2[5][1] = set of non-apparent singularities of L2.
towopegenexpdiff1F2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;
                  # options trace;
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equiva1F22(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equiva1F21(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsing1F2(L2,t,G,ext);       ####### L2 can be our L operator.
Exptest1F2equiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= Condi1F2irr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F2equiv2(L1,C2,x,t,ext);       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
Exptest1F2equiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F2equiv2(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equiva1F21:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               if extp0 subset ext then
                  return extp3 union ext
               fi;
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equiva1F22:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F2equiv1(L1,C2,x,t,ext);     ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff1F2(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge1F21:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext).
EquivExpgauge1F21:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest1F2equiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff1F2(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge1F21(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing1F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest1F2equiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff1F2(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge1F2:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=Exptest1F2equiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(Exptest1F2equiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext).
EquivExpgauge1F2:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest1F2equiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff1F2(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge1F2(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 1F2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi1F2irr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l,R;
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 3 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              if l > 2 then
                 return NULL
              fi;
              C:= map(s-> convert(series(subs(t=1/t^l,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              if (type(evala((B[1]-B[2])*2),integer) or type(evala((B[1]+B[2])*2),integer)) and
                 type(evala(B[3]*2),integer) and (evala(A[1]+A[2])=0 or evala(A[1]-A[2])=0) and
                 (evala(A[3]-2*A[1])=0 or evala(A[3]+2*A[1])=0) then
                 return [E[2],E[4],A,B]
              elif (type(evala((B[1]-B[3])*2),integer) or type(evala((B[1]-B[3])*2),integer)) and
                   type(evala(B[2]*2),integer) and (evala(A[1]+A[3])=0 or evala(A[1]-A[3])=0) and
                   (evala(A[2]-2*A[1])=0 or evala(A[2]+2*A[1])=0) then
                    return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]],[A[1],A[3],A[2]],[B[1],B[3],B[2]]]
              elif (type(evala((B[2]-B[3])*2),integer) or type(evala((B[2]-B[3])*2),integer)) and
                   type(evala(B[1]*2),integer) and (evala(A[2]+A[3])=0 or evala(A[2]-A[3])=0) and
                   (evala(A[1]-2*A[2])=0 or evala(A[1]+2*A[2])=0) then
                    return [[E[2][2],E[2][3],E[2][1]],[E[4][2],E[4][3],E[4][1]],[A[2],A[3],A[1]],[B[2],B[3],B[1]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm is "irrsing1F2".
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext).
irrsing1F2:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or
                    member(t,indets(A[i[1]][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d:= Condi1F2irr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 1F2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
# The condition is that:
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi1F2reg:=proc(E,t,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if type(evala(2*(E[2][2]-E[2][1]-E[2][3])),integer) then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif type(evala(2*(E[2][1]-E[2][2]-E[2][3])),integer) then
                 return [[E[2][2],E[2][1],E[2][3]],[E[4][2],E[4][1],E[4][3]]]
              elif type(evala(2*(E[2][3]-E[2][1]-E[2][2])),integer) then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# R = Singular(L,extppp);
# F = NotAppSing(L,R,ext);
# H = irrsing1F2(L,t,F,ext)[-1] or IrrRegAppsing1F2(L,t,R,ext)[1][-1].
regsingtrue1F2:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                 # must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= Condi1F2reg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1];
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in three parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer);
#       - removable singularities with integer exponent differences;
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# R = Singular(L,extppp);
# F = NotAppSing(L,R,ext);
# Sirr = irrsing1F2(L,t,F,ext) or IrrRegAppsing1F2(L,t,R,ext)[1];
# H = regsingtrue1F2(L,t,Sirr[-1],ext) or IrrRegAppsing1F2(L,t,R,ext)[2].
Sregseptrue1F2:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# R = Singular(L,extppp);
# F = NotAppSing(L,R,ext);
# Sirr = irrsing1F2(L,t,F,ext) or IrrRegAppsing1F2(L,t,R,ext)[1];
# Sreg = regsingtrue1F2(L,t,Sirr[-1],ext) or IrrRegAppsing1F2(L,t,R,ext)[2];
# RSreg = Sregseptrue1F2(L,Sreg,ext) or IrrRegAppsing1F2(L,t,R,ext)[3].
UnionSregseptrue1F2:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################


# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# Sirr = irrsing1F2(L,t,F,ext) or IrrRegAppsing1F2(L,t,G,ext)[1];
# Sreg = regsingtrue1F2(L,t,Sirr[-1],ext) or IrrRegAppsing1F2(L,t,G,ext)[2].
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp).
IrrRegAppsing1F2:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,
                    H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or
                    member(t,indets(A[i][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d1:= Condi1F2irr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                    # must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # generalized exponents have to be positive integers.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= Condi1F2reg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1];
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31];
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31];
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= Condi1F2reg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1];
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31];
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31];
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsingstruc(L,t,ext)[1][i],x),
#######                      with i=position of the minimal polynom of p) is 1.
############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g;
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Hyperpossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:= table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     C[k]:= h;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# We compute the product of (x-p)^(2*m) where p is an irregular singularity (pole) of f and m is its multiplicative
# order. That will give us the denominator of g=f^2.
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext)
# C = irrsing1F2(L,t,H,ext) or IrrRegAppsing1F2(L,t,G,ext)[1].
PolirrProd1F2:=proc(L,C)
            local Dx,x,b,i,s;
            Dx,x:=op(_Envdiffopdomain);
            b:=1;
            i:=0;
            for s in C[1] do
               i:=i+1;
               if s[1] <> infinity then
                  b:=b*(s[1])^(2*C[4][i])
               fi;
            od;
            return b
end:

############################################################################################################
############################################################################################################

degpol:=proc(a,x,t,ext)
          local Dx,l1,d;
          Dx,x:=op(_Envdiffopdomain);
          l1:=series(subs(t=1/x,a),x=0,100);
          l1:=eval(subs(csgn=1,simplify(l1) assuming(x>0)));
          d:= degree(convert(l1,polynom), x);
          if d=FAIL then
             return false
          else
             return d
          fi;
end:

############################################################################################################
############################################################################################################

# We take as input A = irrsing1F2(L,t,ext) and l the minimal integer m such that f belong to k(x^(1/m)). We return a
# list of list in the form [[pol,s,d,acc],dA,neq,B,nonpole] where [pol,s,d,acc] is related to s with:
# s = an irregular singularity of L (pole of f);
# pol = polynomial over K with zero s;
# d = truncated series of g=f^2 at an irregular singularity s of L;
# acc = smallest integer greater than or equal to m_p/l;
# dA = degree bound for the numerator of f^2   (A = numer(f^2));
# neq = number of linear equations for the coefficients of A = numer(f^2);
# B = denominator of f^2;
# nonpole = denomintor of f but with integer power (the smallest integer greater than or equal to m_p);
# Here n_e=2.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# A = irrsing1F2(L,t,F,ext) or IrrRegAppsing1F2(L,t,H,ext)[1];
# G = regsingtrue1F2(L,t,A[-1],ext) or IrrRegAppsing1F2(L,t,H,ext)[2].
Sirr1F2info1:=proc(L,A,G,x,t,ext)
          local Dx,m,a1,a2,b1,b2,b3,R,e1,e2,e3,acc2,acc1,j,s,B,d,i,e,acc,a,b,l1,k,d1,v,v1,E,neq,dA,nonpole,
                extp1,extp2,extp3,extp4,extp5,extp6,extp0,extp7,H1;
          Dx,x:=op(_Envdiffopdomain);
          # We find the coefficients e1 and e2 of (x-p)^(1/2) in the two solutions (with t the unknown) of b*t^2 = x-p
          # Those coefficients are the two first solutions of the equation x^2-1/b = 0. Here b=1. So
          # the equation is now x^2-1=0. So e1,e2 belong to {-1,1}. That means e1-e2=-+2.
          E:= [];
          nonpole:= 1;
          if G=[] then
             neq:= 0
          else
             neq := add(`if`(s[1]=infinity,1,degree(s[1],x)), s=G[1]);
          fi;
          B:=1;
          dA:=0;
          j:=0;
          for s in A[1] do
             j:=j+1;
             m:= 2*A[4][j];
             if type(m,integer)=false then
                return NULL
             fi;
             if s[1] <> infinity then
                B:= B*(s[1])^m;
             else
                dA := m;
             fi;
             acc:= m; # degree minimal, en valeur absolue, du polynome de
                                      # la partie polaire de f^2 en p.
             acc2:= ceil(A[4][j]); # nombre d'equations venant de p pour les coefficients de A= f*B. Dans
                                   # les normes, la formule est: ceil((1/n_e)*m_j)=ceil(A[4][j]).
                                   # Here n_e = 2.
             acc1:= floor(A[4][j]); # nombre de termes dans la partie polaire de f qu'on n'a pas. Dans
                                 # les normes, la formule est: floor(((n_e-1)/n_e)*m_j)= floor((n_e-1)*A[4][j]).
                                 # Here n_e = 2.
             neq:= neq + acc2*op(`if`(s[1]=infinity,1,degree(s[1],x)));
             d:= convert(series(A[3][j][1],t=0,100),polynom);
             if type(d,`+`) then
                d1:= [op(d)]
             else
                d1:=[d]
             fi;
             # We remove the constant term in d
             e:=[];
             for i in d1 do
                if member(t,indets(i)) then
                   e:=[op(e),i]
                fi;
             od;
             l1:=map(s->subs(t=t^2,s),e);
             l1:=map(s->series(subs(t=1/x,s),x=0,100),l1);
             l1:=map(s->eval(subs(csgn=1,simplify(s) assuming(x>0))),l1);
             d1:=[e,l1];
             d:= add(d1[1][i]/(ldegree(convert(d1[2][i],polynom), x)), i=1..nops(e));
             d:= eval(subs(csgn=1,simplify(d) assuming(x>0)));
             d:=evala(expand(d^2));
             d:=convert(series(d,t=0,100),polynom);
             if type(d,`+`) then
                d1:= [op(d)]
             else
                d1:=[d]
             fi;
             d:=0;
             for i in d1 do
                l1:= degpol(i,x,t,ext);
                if l1<>false and l1<=m then
                   d:=d+i
                fi;
             od;
             if member(t,indets(d,{name}))=false then
                return NULL
             fi;
             if s[1]=infinity then
                b:=subs(t=1/x,d);
                a1:=convert(series(numer(b),x=0,100),polynom);
                a2:=expand(convert(series(denom(b),x=0,100),polynom));
                b:= factor(a1/a2);
                b:=series(b,x=0,100);
                b:= eval(subs(csgn=1,simplify(b) assuming(x>0)));
                extp1:= indets(A[3][j][1],{RootOf,name}) minus {t};
                extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All the names or RootOf
                                        # different from RootOf(s,x) in the expression of A[3][j][1].
                extp2:=indets(ext,{RootOf,name});
                extp3:= extp1 minus extp2; # New names or Rootof which appear.
                extp5:= extp1 union ext;
                E:=[op(E),[s[2],1/x,[convert(evala(b),polynom)],acc,extp3,extp5]];
                nonpole:= nonpole*(1)^acc1; # we take for infinity [1/x,pol]=[1,1].
             else
                b:=`if`(degree(s[1],x) > 1,subs(t=x-s[2],d),subs(t=s[1],d));
                a1:=convert(series(numer(b),x=s[2],100),polynom);
                a2:=expand(convert(series(denom(b),x=s[2],100),polynom));
                b:= factor(a1/a2);
                b:= series(b,x=s[2],100);
                b:= eval(subs(csgn=1,simplify(b) assuming(x>0)));
                extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                extp1:= indets(A[3][j][1],{RootOf,name}) minus {t};
                extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                extp2:=indets({op(ext),ReplirrRoot(s[2],{})},{RootOf,name});
                extp3:= extp1 minus extp2; # New names or Rootof which appear.
                extp4:= extp1 minus extp0; # All the names or RootOf
                                           # different from RootOf(s,x) in the expression of A[3][j][1].
                extp6:= ext minus extp0;
                extp7:= extp4 union extp6;
                v1:={};
                for v in extp7 do
                   if extp0 intersect indets(v,{RootOf,name}) <> {} then
                      v1:= v1 union {v}
                   fi;
                od:
                extp5:= extp7 minus v1;
                E:=[op(E),[s[2],s[1],[convert(evala(b),polynom)],acc,extp3,extp5]];
                nonpole:= nonpole*(s[1])^acc1;
             fi;
          od;
          dA:= degree(B,x)+dA;
          H1:= table();
          H1[1]:=E;
          H1[2]:=dA;
          H1[3]:=neq;
          H1[4]:=B;
          H1[5]:=nonpole;
          H1:= convert(H1,'list');
          return H1
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate [b1,b2] for the lower parameter of 1F2 must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = Zeroesmultordpol1F2(f,x);
# n = nops(B1);
# RS = newregsing1F2(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# (b1,b2) = candidate for the lower parameters.
Condibiexp1F2:= proc(E1,E2,m,b1,b2,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(m*(1-b1)-E2[1]),integer) and type(evala(m*(1-b2)-E2[2]),integer) and
                 type(evala(m*(b1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b1)),integer) and
                 type(evala(E1[2][1]-m*(1-b2)),integer)) then
                 return true
              elif (type(evala(m*(1-b2)-E2[1]),integer) and type(evala(m*(1-b1)-E2[2]),integer) and
                 type(evala(-m*(b1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b2)),integer) and
                 type(evala(E1[2][1]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(1-b1)-E2[1]),integer) and type(evala(m*(b1-b2)-E2[2]),integer) and
                 type(evala(m*(1-b2)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b1)),integer) and
                 type(evala(E1[2][1]-m*(1-b2)),integer)) then
                 return true
              elif (type(evala(m*(b1-b2)-E2[1]),integer) and type(evala(-m*(1-b1)-E2[2]),integer) and
                 type(evala(-m*(1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-m*(1-b2)),integer) and
                 type(evala(E1[1][2]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(1-b2)-E2[1]),integer) and type(evala(-m*(b1-b2)-E2[2]),integer) and
                 type(evala(m*(1-b1)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b2)),integer) and
                 type(evala(E1[2][1]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(b1-b2)-E2[1]),integer) and type(evala(-m*(1-b2)-E2[2]),integer) and
                 type(evala(-m*(1-b1)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-m*(1-b1)),integer) and
                 type(evala(E1[1][2]-m*(1-b2)),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
Zeroesmultordpol1F2:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsing1F2(L,t,H,ext)[2];
# B = Zeroesmultordpol1F2(f,x);
# n = nops(B).
newregsing1F2:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],Condi1F2reg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],Condi1F2reg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################

####### Let E be a list of two element [a,b]. We return a list [c,d] where c and d are the constant part of a and b,
####### respectively.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,H,ext);
# i = index of an element in R[1][1];
# m = R[1][5][i];
# E = an element in R[1][6][i]. E is a list of two elements a and b (E=[a,b]).
helpCondiaibiexp1F2:= proc(E,t,m,ext)
              local Dx,x,a,i,c1,C;
              Dx,x:=op(_Envdiffopdomain);
              C:= map(s-> convert(series(subs(t=1/t^m,s),t=0,100),polynom),E);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              C:= map(s-> evala(coeff(s,t,0)),C);
              return C
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name};
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,H,ext);
# k = index of an element in R[1][1];
# m1 = R[1][5][k];
# E2 = R[1][8][k];
# E1 = map(s-> helpCondiaibiexp1F2(s,t,m1,ext),R[1][6][k]);
# m = R[1][4][k].
Condiaibiexp1F2:= proc(E2,E1,m,b1,b2,a1,ext)
           local Dx,x;
           Dx,x:=op(_Envdiffopdomain);
           if type(evala(2*(m*(b1+b2-3*a1-1/2)/2-E2[1])),integer) and
              type(evala(2*(m*(b1+b2-3*a1-1/2)/2-E2[2])),integer) and
              type(evala(2*E2[3]),integer) and type(evala(2*(E1[1][1]-E1[2][1])),integer) and
              type(evala(2*(E1[1][2]-E1[2][2])),integer) and type(evala(2*(E1[3][1]-E1[3][2])),integer) and
              type(evala(2*(E1[1][1]-E1[3][1])),integer) then
              if indets(a1,{RootOf,name}) subset indets(evala(E1[1][1]-m*(b1+b2)/2),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[1][2]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif type(evala(2*(m*(b1+b2-3*a1-1/2)/2+E2[1])),integer) and
              type(evala(2*(m*(b1+b2-3*a1-1/2)/2-E2[2])),integer) and
              type(evala(2*E2[3]),integer) and type(evala(2*(E1[1][1]-E1[2][2])),integer) and
              type(evala(2*(E1[1][2]-E1[2][1])),integer) and type(evala(2*(E1[3][1]-E1[3][2])),integer) and
              type(evala(2*(E1[1][2]-E1[3][1])),integer) then
              if indets(a1,{RootOf,name}) subset indets(evala(E1[1][2]-m*(b1+b2)/2),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[1][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           elif type(evala(2*(m*(b1+b2-3*a1-1/2)/2+E2[1])),integer) and
              type(evala(2*(m*(b1+b2-3*a1-1/2)/2+E2[2])),integer) and
              type(evala(2*E2[3]),integer) and type(evala(2*(E1[1][1]-E1[2][1])),integer) and
              type(evala(2*(E1[1][2]-E1[2][2])),integer) and type(evala(2*(E1[3][1]-E1[3][2])),integer) and
              type(evala(2*(E1[1][2]-E1[3][1])),integer) then
              if indets(a1,{RootOf,name}) subset indets(evala(E1[1][2]-m*(b1+b2)/2),{RootOf,name}) and
                 indets(a1,{RootOf,name}) subset indets(evala(E1[1][1]),{RootOf,name}) then
                 return true
              else
                 return false
              fi;
           else
              return false
           fi;
end:

############################################################################################################
############################################################################################################

# For a given candidates [[[b1,b2],[b1,b2],...]],f], we find associated candidates for [{a,a...,},[b1,b2],f]
# related an element of [[b1,b2],[b1,b2],...] and an irregular singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,H,ext);
# F = R[1];
# C = Candidates for [[b1,b2],[b1,b2],[b1,b2]];
# i = index of an element in F[1].
helpcandai1F2:=proc(L,F,C,i,t,ext,ext1)
            local Dx,x,E1,E2,E,s,A,A1,B,j,m1,m;
            Dx,x:=op(_Envdiffopdomain);
            m:= 2*F[4][i];
            m1:= F[5][i];
            E2:= F[8][i];
            E1:= map(s-> helpCondiaibiexp1F2(s,t,m1,ext),F[6][i]);
            E:= table();
            for j in C do
               ####### candidates for a1 modulo an integer
               A:= [seq((j[1]+j[2]-1/2-2*(E2[1]/m))/3+s/(3*m),s=-3*m+1..3*m-1),
                     seq((j[1]+j[2]-1/2+2*(E2[1]/m))/3+s/(3*m),s=-3*m+1..3*m-1)];
               A:=[op({op(A)})];
               ####### We make the rational part of "a1" belong to ]0,1[ union {-1,1}
               B:={op(map(s->Ratiopartmodint(s,ext1),A))};
               B:=map(s->`if`(s[1]=0,op([[-1,-1,s[3],s[4]],[1,1,s[3],s[4]]]),s),B);
               A1:= table();
               for s in B do
                  if Condiaibiexp1F2(E2,E1,m,j[1],j[2],s[1],ext)=true then
                     A1[s]:= s[1];
                  fi;
               od;
               A1:= convert(A1,'set');
               E[j]:= [A1,j];
            od;
            E:= convert(E,'list');
            return E
end:

############################################################################################################
############################################################################################################

# We compute the candiates for a1, b1 and b2 related to a candidate of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,H,ext);
# F = R[1];
# C = Candidates for [[b1,b2],[b1,b2],[b1,b2]].
findcandai1F2:=proc(L,F,C,x,t,ext,ext1)
                  local j,B,k,E,H;
                  H := table();
                  E:= helpcandai1F2(L,F,C,1,t,ext,ext1);
                  if E=NULL then
                     return NULL
                  fi;
                  if nops(F[1]) > 1 then
                     for j from 2 to nops(F[1]) do
                        B:= helpcandai1F2(L,F,C,j,t,ext,ext1);
                        if B=NULL then
                           return NULL
                        fi;
                        for k from 1 to nops(B) do
                           E[k][1]:= B[k][1] intersect E[k][1];
                        od;
                     od;
                  fi;
                  for j from 1 to nops(E) do
                     if E[j][1]<>{} then
                        H[j]:= E[j];
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################

funct1F2:=proc(x,f,a,b)
          local Dx,k;
          Dx,x:=op(_Envdiffopdomain);
          return expand(sum(subs(x=f,hyperterm([a],[b[1],b[2]],x,k)),k));
end:

############################################################################################################
############################################################################################################
#################################################   EASY CASE   ############################################
############################################################################################################
############################################################################################################

# We compute the candiates for a1,a2,b1,b2 and f: [{{a1,...}},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
easy1F2f:= proc(L,R,info,x,t,ext)
                  local Dx,a,f,fn,p,i,j,linearr,temp,d,d1,nonpole,und,B1,n,SR,g,
                        g1,A1,A2,G,H1,H,c1,c2,c3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  ####### info[4] gives us the denominator of f.
                  ####### degree(A,x)=info[2].
                  H:= table();
                  f := add(a[i]*x^i, i = 0 .. info[2])/info[4];
                  und := indets(f) minus {x};
                  C:=map(s->s[1],R[2][1]);       ####### In easy case, Sreg<>[]
                  if has({op(C)}, infinity) then
                     f:=subs(a[info[2]]=0,f);
                     und:=subs(a[info[2]]=NULL,und);
                  fi;
                  fn := numer(f);
                  for p in R[2][1] do
                     if p[1]=infinity then
                        p[1]:= 1/x;
                     fi;
                     fn := fn/(p[1]);
                  od;
                  linearr:=convert(collect(eval(subs(csgn=1,simplify(evala(Rem(numer(fn),denom(fn),x))) assuming(x>0))),x),polynom);
                  linearr := coeffs(linearr, x);
                  temp := SolveTools:-Linear({linearr},und);
                  f := subs(op(temp), f);
                  H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
                  j := 0; # use j indicate that if infinity is an irregular singularity
                  for p in info[1] do
                     if p[1] = infinity then
                        j := ceil(p[4]/2);
                        break;
                     fi;
                  od;
                  G := Hyperpossibility(H1);
                  for d in G do
                     c1:= convert(series(numer(f-d),x=0,100),polynom);
                     c2:= convert(series(denom(f-d),x=0,100),polynom);
                     c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                     nonpole := Normalizer(c3*info[5]);
                     # nonpole := Normalizer((f-d)*info[5]);
                     A1:=convert(series(numer(nonpole),x=0,100),polynom);
                     A2:=convert(series(denom(nonpole),x=0,100),polynom);
                     temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                     temp := collect(temp, x, Normalizer);
                     linearr := coeffs(temp, x);
                     if j<>0 then # the equations at infinity
                        temp := evala(Quo(A1, A2,x));
                        temp:=convert(series(numer(temp),x=0,100),polynom);
                        temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                        temp := collect(temp, x, Normalizer);
                        d1:=degree(temp, x);
                        temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                        linearr := linearr, temp;
                     fi;
                     temp := SolveTools:-Linear({linearr},und);
                     if temp=NULL then
                        next
                     fi;
                     temp := op(temp);
                     g := subs(temp, f);
                     g := factor(op(simplef({g})));
                     B1:= Zeroesmultordpol1F2(g,x);
                     if B1=NULL then
                        next
                     fi;
                     n:= nops(B1);
                     SR:= newregsing1F2(L,t,x,B1,R[2],n,ext);
                     H[d]:= [g,SR[2],SR[1]];
                  od;
                  H:= convert(H,'set');
                  if H = {} then
                     return NULL
                  else
                     return H
                 fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for a1,a2,b1,b2 and f: [{{a1,...}},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
easy1F2:= proc(L,R,info,x,t,ext)
          local B,ab,f,C,H,E,n1,n,ext1,ext2,ext3,ext4,i,n2,s;
          B:= easy1F2f(L,R,info,x,t,ext);
          if B = NULL then
             return NULL
          fi;
          ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
          ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
          n1:=nops(R[2][1]);
          if n1 > 1 then
             for i from 2 to n1 do
                ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                ext1:= ext1 intersect ext2;
             od;
          fi;
          ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
          ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
          n2:=nops(R[1][1]);
          if n2 > 1 then
             for i from 2 to n2 do
                ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                ext3:= ext3 intersect ext4;
             od;
          fi;
          H:= table();
          for f in B do
             n:= nops(f[3][1]);
             if R[3][3] <> [] then
                C:= findbi1F2ln(L,f,x,t,ext,ext1,n);
             else
                C:= findbi1F2(L,f,x,t,ext,ext1,n);
             fi;
             if C<>NULL then
                ab:= findcandai1F2(L,R[1],C[1],x,t,ext,ext3);
                if ab <> NULL then
                   H[f]:= [op(ab),f[1]]
                fi;
             fi;
          od;
          H:= convert(H,'list');
          if H = [] then
             return NULL
          else
             return H
          fi;
end:

############################################################################################################
############################################################################################################
############################################   LOGARITHMIC CASE   ##########################################
############################################################################################################
############################################################################################################

# g = f^2 = (C.A1.A2^b)/B. We want to find candidates for A1. Sreg=regsingtrue(L,t,ext)[1] is the set of the
# zeros of A1.
# input: H and d, where H=subs(Sreg[4]=NULL,Sreg) and d is the degree of A1. We do a combinatorial distribution
# of d to the elements of H[1];
# return: candidates for A1 = \prod_{p\in Sreg[1]}(p^{k_p}) where
# p = an element of Sreg[1] (minimal polynomial)
# d = \sum_{p\in Sreg[1]}(k_p*deg(p)) = degree of A1.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp);
# F = NotAppSing(L,R,ext);
# Sirr = irrsing1F2(L,t,F,ext) or IrrRegAppsing1F2(L,t,R,ext)[1];
# H = regsingtrue1F2(L,t,Sirr[-1],ext) or IrrRegAppsing1F2(L,t,R,ext)[2].
searchKnLog1F2 :=proc(H,d)
              local res,temp,s,deg,t,k;
              res:=NULL;
              if nops(H[1])=1 then
                 deg:=`if`(H[1][1][1]=infinity,1,degree(H[1][1][1],x));
                 if d mod deg = 0 then
                    if H[1][1][1]= infinity then
                       res:=res,1;
                    else
                       res:= res, H[1][1][1]^(d/degree(H[1][1][1],x));
                    fi
                 fi;
              else
                 deg := `if`(H[1][1][1]=infinity,1,degree(H[1][1][1],x));
                 for k from 1 to iquo(d,deg) do
                    temp := searchKnLog1F2([seq([seq(H[i][j],j=2..nops(H[i]))],i=1..nops(H))],d-k*deg);
                    if temp<>NULL then
                       if H[1][1][1] <> infinity then
                          for t in temp do
                             res:= res,H[1][1][1]^k*t;
                          od;
                       else
                          for t in temp do
                             res:=res,t;
                          od;
                       fi;
                    fi;
                 od;
              fi;
              return [res];
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# R = IrrRegAppsing1F2(L,t,R,ext);
# R = NotAppSing(L,R,ext);
# Sreg = R[2];
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
find1F2Logf := proc(L,info,Sreg,x,t,ext) ####### Rapide
                 local Dx,mult1,M,p,res,knl,kn,g,i,j,linearr,temp,d,d1,nonpole,B,B1,n,SR,dA,k,e,G,A1,A2,h,h1,H,c1,c2,c3;
                 Dx,x:=op(_Envdiffopdomain);
                 res:= {};
                 dA:= info[2];
                 B:= info[4];
                 knl := searchKnLog1F2(Sreg,dA);
                 H:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
                 j := 0; # use j indicate that if infinity is an irregular singularity
                 k:=0;
                 for p in info[1] do
                    k:=k+1;
                    if p[1] = infinity then
                       j := ceil(info[1][k][4]/2);
                       break;
                    fi;
                 od;
                 G := Hyperpossibility(H);
                 for kn in knl do
                    for d in G do
                       c1:= convert(series(numer(c*kn/B-d),x=0,100),polynom);
                       c2:= convert(series(denom(c*kn/B-d),x=0,100),polynom);
                       c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                       nonpole := Normalizer(c3*info[5]);
                       A1:=convert(series(numer(nonpole),x=0,100),polynom);
                       A2:=convert(series(denom(nonpole),x=0,100),polynom);
                       temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                       temp := collect(temp, x, Normalizer);
                       linearr := coeffs(temp, x);
                       if j<>0 then #the equations at infinity
                          temp := evala(Quo(A1, A2,x));
                          temp:=convert(series(numer(temp),x=0,100),polynom);
                          temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                          temp := collect(temp, x, Normalizer);
                          d1:=degree(temp, x);
                          temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                          linearr := linearr, temp;
                       fi;
                       temp := SolveTools:-Linear({linearr},{c});
                       if temp=NULL then
                          next
                       fi;
                       temp := op(temp);
                       h:= subs(temp, c*kn/B);
                       B1:= Zeroesmultordpol1F2(h,x);
                       if B1=NULL then
                          next
                       fi;
                       n:= nops(B1);
                       SR:= newregsing1F2(L,t,x,B1,Sreg,n,ext);
                       res := res union {[h,SR[2],SR[1]]};
                    od;
                 od;
                 if res = {} then
                    return NULL
                 else
                    return res
                 fi;
end:

############################################################################################################
############################################################################################################

####### b11, b12 and b13 are sets of candidates for 1-b1, 1-b2 and b1-b2 respectively. We compute here the candidates
####### for [b1,b2] with the rational part of b1 and b2 in ]0,1], related to a zero of the change of variable parameter.
#######
####### m = multiplicity order of the element as a zero of f;
####### e = [Delta1,Delta2,Delta3];
####### a = [[e,e],[e,e],[e,e]].
casesbiLog1F2:= proc(b11,b21,b31,e,a,m,ext,ext1)
          local k,s,b1,b2,b3,B,B1,B2,B3,d1,d2,i,j;
          b1:=map(s->1-s,b11);
          b2:=map(s->1-s,b21);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          b2:= [seq(`if`(MatchInt2([op(b2[1..k-1])],b2[k]), NULL, b2[k]),k=1..nops(b2))];
          b3:= [seq(`if`(MatchInt2([op(b31[1..k-1])],b31[k]), NULL, b31[k]),k=1..nops(b31))];
          ####### We compute the candidates for b1, b2 and b1-b2 such that their rational parts are in [0,1]
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          B2:={op(map(s-> Ratiopartmodint(s,ext1),b2))};
          B3:={op(map(s-> Ratiopartmodint(s,ext1),b3))};
          B3:=map(s-> [s[1],s[2],s[3]],B3);
          B1:=map(s->`if`(s[1]=0,[1,1,s[3],s[4]],s),B1);
          B2:=map(s->`if`(s[1]=0,[1,1,s[3],s[4]],s),B2);
          B:= table();
          for i in B1 do
             for j in B2 do
                ####### the condition for the logarithmic case
                if type(i[1],integer) or type(j[1],integer) or type(i[1]-j[1],integer) or (type(i[1]*2,integer)
                   and type(j[1]*2,integer) and i[1]*j[1]< 0) then
                   d1:= Ratiopartmodint(i[1]-j[1],ext1);
                   d2:= Ratiopartmodint(j[1]-i[1],ext1);
                   d1:= [d1[1],d1[2],d1[3]];
                   d2:= [d2[1],d2[2],d2[3]];
                   if {d1,d2} intersect B3 <> {} then
                      ####### the condition that a candidate for (b1,b2) must satisfy
                      if Condibiexp1F2(a,e,m,i[1],j[1],ext)=true or Condibiexp1F2(a,e,m,j[1],i[1],ext)=true then
                         B[i,j]:= {i[1],j[1]}
                      fi;
                   fi;
                fi;
             od;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameters [b1,b2] related to a zero of the change of variable
# parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext);
# B = find1F2Logf(L,info,R[2],ext,x,t);
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiffLog1F2:= proc(j,G,m,ext1)
          local d,i,ext2,b11,b21,b31,B,H1,H2,H3,C1,C,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          B:= table();
          C:=[[G[3][j][1],G[3][j][2],G[3][j][3]],[G[3][j][1],G[3][j][3],G[3][j][2]],[G[3][j][2],G[3][j][3],G[3][j][1]]];
          for i in C do
             ####### b11 is the list of candidates for +-(1-b1) modulo an integer
             b11:=[seq(i[1]+s,s=-m+1..m-1),seq(-i[1]+s,s=-m+1..m-1)]/m;
             ####### b21 is the list of candidates for +-(1-b2) modulo an integer
             b21:= [seq(i[2]+s,s=-m+1..m-1),seq(-i[2]+s,s=-m+1..m-1)]/m;
             ####### b31 is the list of candidates for +-(b1-b2) modulo an integer
             b31:= [seq(i[3]+s,s=-m+1..m-1),seq(-i[3]+s,s=-m+1..m-1)]/m;
             C1:=casesbiLog1F2(b11,b21,b31,G[3][j],G[4][j],m,ext,ext1);
             if C1 <> NULL then
                B[i]:=op(C1)
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters b1 and b2 related to a candidate of f: [[[b1,b2],[b1,b2]...],f,M],
# where M is the list of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext);
# B1 = find1F2Logf(L,info,R[2],ext,x,t);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter;
# f[2] list of multiplicity orders of zeroes of f[1];
# f[3][1] list of zeroes of the change of variable parameter.
findbi1F2ln:=proc(L,f,x,t,ext,ext1,n1)
                  local Dx,H,B,j,s,i;
                  Dx,x:=op(_Envdiffopdomain);
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiffLog1F2(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> `if`(nops(s)=1,[s[1],s[1]],[s[1],s[2]]),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for a1,a2,b1,b2 and f: [{{a1,...}},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
find1F2ln:=proc(L,R,info,x,t,ext)
                  local Dx,B,ab,f,C,H,E,n1,n,ext1,ext2,ext3,ext4,i,n2,s;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= find1F2Logf(L,info,R[2],x,t,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                  ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                  n1:=nops(R[2][1]);
                  if n1 > 1 then
                     for i from 2 to n1 do
                        ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                        ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                        ext1:= ext1 intersect ext2;
                     od;
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi1F2ln(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        ab:= findcandai1F2(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##############################################   IRRATIONAL CASE   ##########################################
############################################################################################################
############################################################################################################

leadcoeffRootnameinter:=proc(A,x,ext1)
                    local Dx,l,l1,ira,ira1,ira2,sum1,sum2;
                    Dx,x:=op(_Envdiffopdomain);
                    l:=A;
                    l1:=ext1;
                    while l1<>{} and (l1 intersect ext1 <> {}) do
                       ira1:=ChooseRootOflistinter(l,l1)[1];
                       if ira1<>{} then
                          ira:=ira1[1];
                       else
                          ira2:= l1 intersect ext1:
                          ira:=ira2[1];
                       fi;
                       sum1:=convert(collect(l,ira,evala),polynom);
                       sum2:= `if`(ldegree(sum1,ira)<0,convert(subs(ira=1/ira,sum1),polynom),sum1);
                       l:= lcoeff(sum2,ira);
                       l1:= indets(l,{RootOf,name});
                    od;
                    return l
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,ext);
# R = IrrRegAppsing1F2(L,t,E,ext);
# info = Sirr1F2info1(L,R[1],R[2],l,x,ext).
find1F2Irrf:= proc(L,R,info,x,t,ext)
             local Dx,M,B1,n1,SR,res,ira,dA,B,ext1,ext2,A,m,i,r,sum,a,j,h,H1,s,p,G,d,A1,A2,d1,c1,c2,c3,c,linearr,
                   temp,nonpole,n,b1,v,V,g,ira1,ira2,H,E,C,l1;
             Dx,x:=op(_Envdiffopdomain);
             res := {};
             M:=[];
             dA:= info[2];
             B:= info[4];
             n:=nops(R[2][1]);
             ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
             ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
             if n > 1 then
                for i from 2 to n do
                   ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                   ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                   ext1:= ext1 intersect ext2;
                od;
             fi;
             ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
             ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
             ####### irrational case.
             r[1]:=1;
             for i from 1 to n do
                ira:=0;
                for j from 1 to 3 do
                   C:=ReplirrRoot(R[2][3][i][j],{});
                   l1:= leadcoeffRootnameinter(C,x,ext1);
                   if type(l1,rational)=false or l1=0 then
                      next
                   else
                      H[i]:=C;
                      ira:=1;
                      break
                   fi;
                od;
                if ira=0 then
                   return NULL
                fi;
                a[i]:= l1*(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));
                if i>1 then
                   r[i] := Normalizer(abs(a[i]/a[1])); #   r[i]=(m_i*degree(p_i,x))/(m_1*degree(p_1,x))
                   if not type(r[i],rational) then
                      return NULL
                   fi;
                fi;
             od;
             sum:=add(r[i],i=1..n);
             b1:=dA/sum; # b1 = m_1*degree(p_1,x)
             if type(b1, integer) then
                A:=c;
                for i from 1 to n do
                   m[i]:= (r[i]*b1)/(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));   # m[i]=m_i
                   A:=A*(`if`(R[2][1][i][1]=infinity,1,(R[2][1][i][1])^(m[i])));
                   M:=[op(M),m[i]];
                od;
             else
                return NULL;
             fi;
             H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
             j := 0; # use j indicate that if infinity is irregular singularity
             for p in info[1] do
                if p[1] = infinity then
                   j := ceil(p[4]/2);
                   break;
                fi;
             od;
             G := Hyperpossibility(H1);
             for d in G do
                c1:= convert(series(numer(A/B-d),x=0,100),polynom);
                c2:= convert(series(denom(A/B-d),x=0,100),polynom);
                c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                nonpole := Normalizer(c3*info[5]);
                # nonpole := Normalizer((A/B-d)*info[5]);
                A1:=convert(series(numer(nonpole),x=0,100),polynom);
                A2:=convert(series(denom(nonpole),x=0,100),polynom);
                temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                temp := collect(temp, x, Normalizer);
                linearr := coeffs(temp, x);
                if j<>0 then # the equations at infinity
                   temp := evala(Quo(A1, A2,x));
                   temp:=convert(series(numer(temp),x=0,100),polynom);
                   temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                   temp := collect(temp, x, Normalizer);
                   d1:=degree(temp, x);
                   temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                   linearr := linearr, temp;
                fi;
                temp := SolveTools:-Linear({linearr},{c});
                if temp=NULL then
                   next
                fi;
                temp := op(temp);
                h:= subs(temp, A/B);
                B1:= Zeroesmultordpol1F2(h,x);
                if B1=NULL then
                   next
                fi;
                n1:= nops(B1);
                SR:= newregsing1F2(L,t,x,B1,R[2],n1,ext);
                res := res union {[h,SR[2],SR[1]]};
             od;
             if res = {} then
                return NULL
             else
                return [res,n,ext1]
             fi;
end:

############################################################################################################
############################################################################################################

####### b11, b12 and b13 are sets of candidates for 1-b1, 1-b2 and b1-b2 respectively. We compute here the candidates
####### for [b1,b2] with the rational part of b1 and b2 in ]0,1], related to a zero of the change of variable parameter.
#######
####### m = multiplicity order of the element as a zero of f;
####### e = [Delta1,Delta2,Delta3];
####### a = [[e,e],[e,e],[e,e]].
casesbi1F2:= proc(b11,b21,b31,e,a,m,ext,ext1)
          local k,s,b1,b2,b3,B,B1,B2,B3,d1,d2,i,j;
          b1:=map(s->1-s,b11);
          b2:=map(s->1-s,b21);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          b2:= [seq(`if`(MatchInt2([op(b2[1..k-1])],b2[k]), NULL, b2[k]),k=1..nops(b2))];
          b3:= [seq(`if`(MatchInt2([op(b31[1..k-1])],b31[k]), NULL, b31[k]),k=1..nops(b31))];
          ####### We compute the candidates for b1, b2 and b1-b2 such that their rational parts are in [0,1]
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          B2:={op(map(s-> Ratiopartmodint(s,ext1),b2))};
          B3:={op(map(s-> Ratiopartmodint(s,ext1),b3))};
          B3:=map(s-> [s[1],s[2],s[3]],B3);
          B1:=map(s->`if`(s[1]=0,NULL,s),B1);
          B2:=map(s->`if`(s[1]=0,NULL,s),B2);
          if B1={} or B2={} then
             return NULL
          fi;
          B:= table();
          for i in B1 do
             for j in B2 do
                ####### the condition for the non-logarithmic case
                if type(i[1]-j[1],integer)=false and (type(i[1]*2,integer)=false or type(j[1]*2,integer)=false or
                   i[1]*j[1]>= 0) then
                   d1:= Ratiopartmodint(i[1]-j[1],ext1);
                   d2:= Ratiopartmodint(j[1]-i[1],ext1);
                   d1:= [d1[1],d1[2],d1[3]];
                   d2:= [d2[1],d2[2],d2[3]];
                   if {d1,d2} intersect B3 <> {} then
                      ####### the condition that a candidate for (b1,b2) must satisfy
                      if Condibiexp1F2(a,e,m,i[1],j[1],ext)=true or Condibiexp1F2(a,e,m,j[1],i[1],ext)=true then
                         B[i,j]:= {i[1],j[1]}
                      fi;
                   fi;
                fi;
             od;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameters [b1,b2] related to a zero of the change of variable
# parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext);
# B = find1F2Irrf(L,info,R[3][1],ext,x,t);
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiff1F2:= proc(j,G,m,ext,ext1)
          local d,i,ext2,b11,b21,b31,B,H1,H2,H3,C1,C,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          B:= table();
          C:=[[G[3][j][1],G[3][j][2],G[3][j][3]],[G[3][j][1],G[3][j][3],G[3][j][2]],[G[3][j][2],G[3][j][3],G[3][j][1]]];
          for i in C do
             ####### b11 is the list of candidates for +-(1-b1) modulo an integer
             b11:=[seq(i[1]+s,s=-m+1..m-1),seq(-i[1]+s,s=-m+1..m-1)]/m;
             ####### b21 is the list of candidates for +-(1-b2) modulo an integer
             b21:= [seq(i[2]+s,s=-m+1..m-1),seq(-i[2]+s,s=-m+1..m-1)]/m;
             ####### b31 is the list of candidates for +-(b1-b2) modulo an integer
             b31:= [seq(i[3]+s,s=-m+1..m-1),seq(-i[3]+s,s=-m+1..m-1)]/m;
             C1:=casesbi1F2(b11,b21,b31,G[3][j],G[4][j],m,ext,ext1);
             if C1 <> NULL then
                B[i]:=op(C1)
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters b1 and b2 related to a candidate of f: [[[b1,b2],[b1,b2]...],f,M],
# where M is the list of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext);
# B1 = find1F2Irrf(L,info,R[3][1],ext,x,t);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter;
# f[2] list of multiplicity orders of zeroes of f[1];
# f[3][1] list of zeroes of the change of variable parameter.
findbi1F2:=proc(L,f,x,t,ext,ext1,n1)
                  local Dx,H,B,j,s,i;
                  Dx,x:=op(_Envdiffopdomain);
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiff1F2(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> `if`(nops(s)=1,[s[1],s[1]],[s[1],s[2]]),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the upper and lower parameters (a,b1,b2), also and f: [{a,a,...},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = NotAppSing(L,G,ext);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
find1F2Irr:=proc(L,R,info,x,t,ext)
                  local Dx,B,ab,f,C,H,E,ext2,ext3,ext4,i,n2,n,s,G1;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= find1F2Irrf(L,R,info,x,t,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B[1] do
                     n:= nops(f[3][1]);
                     C:= findbi1F2(L,f,x,t,ext,B[3],n);
                     if C<>NULL then
                        ab:= findcandai1F2(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##########################################   RATIONAL CASE   ###############################################
############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute all the candidates for A1 associated with
# the degree of A2. The output is a list with element in the form [s,k] where s=A1 and k=degree(A2).
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# G = R[3][1];
# dr = degree(A1);
# k = degree(A2);
# dA = degree(cA1.A2^d) = dr+k*d = degree of the numerator of f^2;
# s = A1.
search1F2A1k:=proc(G,dA,d,x)
           #option trace;
           local Dx,s,dr,k,res,temp;
           Dx,x:=op(_Envdiffopdomain);
           res :=NULL;
           if G=[] then
              if irem(dA,d)<>0 then
                 return {};
              else
                 return {[1,dA/d]};
              fi;
           fi;
           for k from 0 to iquo(dA,d) do
              dr :=dA-k*d;
              temp :=search1F2A1(G,dr,d,x);
              for s in temp do
                 res := res, [s,k]
              od;
           od;
           {res};
end:

############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute here all the candidates for A1 having the
# same degree.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# G = R[3][1];
# de = degree(A1);
# k = degree(A2);
# dA = degree(cA1.A2^d) = de+k*d = degree of the numerator of f^2.
search1F2A1 :=proc(G,de,d,x)
           local Dx,d1,res,i,temp,k,t,E,n,j,s,m;
           Dx,x:=op(_Envdiffopdomain);
           res :=NULL;
           if nops(G[1])=0 then
              if de=0 then
                 return [1];
              else
                 return NULL;
              fi;
           fi;
           s :=G[1][1][1];
           d1:=denom(G[3][1][1][1]);
           i :=abs(d/d1);
           if type(i,integer)=false then
              return NULL
           fi;
           if (de-i*(`if`(s=infinity,1,degree(s,x))))<0 then
              return NULL;
           fi;
           E:= [seq([seq(G[n][j],j=2..nops(G[n]))],n=1..nops(G))];
           m:=add(`if`(j[1]=infinity,1,degree(j[1],x)),j=E[1]);
           for k from 1 to iquo(de-m,i*(`if`(s=infinity,1,degree(s,x)))) do
              if i*k>=d then
                 break
              fi;
              temp := search1F2A1(E,de-k*i*(`if`(s=infinity,1,degree(s,x))),d,x);
              if temp <> NULL then
                 for t in temp do
                    res := res,(`if`(s=infinity,1,s))^(i*k)*t;
                 od;
              fi;
           od;
           [res];
end:

############################################################################################################
############################################################################################################

poly_to_powerseries := proc(P, x, p, T, acc)
                       # option trace;
                       local d, l, i, S,a;
                       a:=convert(P,polynom);
                       if a=0 then
                          # error "only treat non-zero powseries"
                          return NULL
                       fi;
                       if p = infinity then
                          d, l:= degree(a,x), lcoeff(a,x);
                          [l, -d,add(evala(coeff(a,x,d-i)/l)*T^i, i=0..acc-1), acc]
                       else
                          S := collect(eval(a, x = T + p), T, evala);
                          d, l := ldegree(S,T), tcoeff(S,T);
                          # Note: the evala in the line S := .. is needed to ensure that d is correct.
                          [l, d,add(evala(coeff(S,T,d+i)/l)*T^i, i=0..acc-1), acc]
                       fi;
end:

############################################################################################################
############################################################################################################

pow_reciprocal := proc(L1,tp)
                  [Normalizer(1/L1[1]),-L1[2],convert(series(1/L1[3],tp=0,L1[4]),polynom),L1[4]]
end:

############################################################################################################
############################################################################################################

pow_time := proc(L1,L2,tp)
            local acc,f,ind;
            acc := min(L1[4],L2[4]);
            f := expand(L1[3]*L2[3]);
            ind := [op(indets(f))];
            if nops(ind) = 2 then
               f := subs(ind[1] = tp, f);
               f := subs(ind[2] = tp, f)
            fi;
            [evala(L1[1]*L2[1]),L1[2]+L2[2],convert(series(f,tp=0,acc),polynom),acc]
end:

############################################################################################################
############################################################################################################

dthroot_powseries := proc(L, tp, d, ext)
                     # option trace;
                     local X,v,res,i,s,l;
                     l:=L[1];
                     try
                        v:=evala(Factors(X^d -l,ext));
                     catch :
                        v:=evala(Factors(numer(normal(X^d - l)),{op(ext)}))
                     end try;
                     v:=select(has,v[2],X);
                     res := NULL;
                     for i in v do
                        if degree(i[1],X) = 1 then
                           res := res, -coeff(i[1],X,0)/coeff(i[1],X,1)
                        fi;
                     od;
                     v := series(L[3]^(1/d), tp=0, L[4]);
                     v := collect(convert(v,polynom),tp,evala);
                     res :={seq([i, L[2]/d, v, L[4]],  i = [res])};
                     res;
end:

############################################################################################################
############################################################################################################

simplef := proc(flist)
           #option trace;
           local f,res;
           res := {};
           for f in flist do
              f := evala(Normal(f, expanded));
              f := evala(Sqrfree(f,x));
              f := f[1]*mul(i[1]^i[2],i=f[2]);
              res := res union {f};
           od;
           res;
end:

############################################################################################################
############################################################################################################

# Let f=(cA1.A2^d)/B where B is the denominator of f. We compute here a candidates for f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# dA = degree(A);
# d = denom(b1) or denom(b2) or denom(b1-b2);
# knl = search1F2A1k(R[3][1],dA,d);
# info = Sirr1F2info1(L,R[1],R[2],x,ext);
# B = info[4];
# kn element of knl.
helpcompute1F2f := proc(info,B,d,kn,x,t,T,ext)
                      local Dx,a,l,acc,ep,B1, Bs, temp, ser,lp, uk, und,i, j, lr,f, extp,ukn,uki,p;
                      Dx,x:=op(_Envdiffopdomain);
                      for i from 1 to nops(info[1]) do
                         acc := info[1][i][4];
                         p:= info[1][i][1];
                         ep := poly_to_powerseries(kn[1],x,p,T,acc);
                         ep := pow_reciprocal(ep,T);
                         Bs := poly_to_powerseries(B,x,p,T,acc);
                         temp := pow_time(ep, Bs, T);
                         ep := poly_to_powerseries(info[1][i][3][1],x,p,T,acc);
                         ser[i] := pow_time(temp, ep,T);
                         lp := ser[i][1];
                         und := indets(lp,RootOf) minus ext;
                         if und={} then
                            if not assigned(l) then
                               l:=lp;
                            else
                               if length(lp)<length(l) then
                                  l := lp;
                               fi;
                            fi;
                         fi;
                      od;
                      if not assigned(l) then
                         return {};
                      fi;
                      uk := add(a[i]*x^i,i=0..kn[2]);
                      und := indets(uk) minus {x};
                      for i from 1 to nops(info[1]) do
                         acc := info[1][i][4];
                         p:= info[1][i][1];
                         temp := ser[i];
                         if temp[2] <> 0 and p <> infinity then
                            error "we made an error"
                         fi;
                         temp[1] := temp[1]/l;
                         extp := indets(p,RootOf) union ext;
                         ep := dthroot_powseries(temp,T,d,extp);
                         ukn := NULL;
                         for uki in {uk} do
                            temp := poly_to_powerseries(uki,x,p,T,acc);
                            if temp = NULL then
                               next
                            fi;
                            for j in ep do
                               lr := temp[3]-j[3];
                               lr := coeffs(lr,T);
                               lr := {lr,temp[1]-j[1]};
                               lr := map(numer,lr);
                               if type(p,RootOf) and not member(p,ext) then
                                  lr := map(evala@Expand, lr);
                                  lr := map(coeffs, lr,p)
                               fi;
                               lr := map(s-> `if`(indets(s) intersect und={},NULL,s),lr);
                               lr := SolveTools:-Linear(lr,und);
                               if lr <> NULL then
                                  f:= subs(op(lr),uki);
                                  ukn := ukn,f;
                               fi;
                            od;
                         od;
                         uk := ukn;
                         if uk = NULL then
                            break
                         fi;
                      od;
                      if uk <> NULL then
                         return {seq(evala(l*i^d*kn[1]/B),i=[uk])}
                      else
                         return NULL
                      fi;
end:

############################################################################################################
############################################################################################################

# Let f=(cA1.A2^d)/B where B is the denominator of f. We compute here all the candidates for f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# dA = degree(A);
# d = denom(b1) or denom(b2) or denom(b1-b2);
# knl = search1F2A1k(R[3][1],dA,d,x);
# info = Sirr1F2info1(L,R[1],R[2],x,ext).
help1F2f := proc(L,info,d,knl,x,t,T,ext)
                local Dx,res,kn,uk,B;
                Dx,x:=op(_Envdiffopdomain);
                res := {};
                if knl <> {} then
                   B:= info[4];
                   for kn in knl do
                      uk:= helpcompute1F2f(info,B,d,kn,x,t,T,ext);
                      if uk <> NULL then
                         res:= res union uk;
                      fi;
                   od;
                fi;
                res;
end:

############################################################################################################
############################################################################################################

Goodexpbi1:=proc(A,x,t,ext)
          local Dx,n,B,i;
          Dx,x:=op(_Envdiffopdomain);
          B:=table();
          n:=nops(A);
          for i from 1 to n do
             if nops(A[i][1])=3 then
                if abs(denom(A[i][1][1])*denom(A[i][1][2]))=abs(denom(A[i][1][3])) or
                   abs(denom(A[i][1][1]))=abs(denom(A[i][1][2])) then
                   B[i]:= A[i]
                elif abs(denom(A[i][1][1])*denom(A[i][1][3]))=abs(denom(A[i][1][2])) or
                     abs(denom(A[i][1][1]))=abs(denom(A[i][1][3])) then
                   B[i]:= [[A[i][1][1],A[i][1][3],A[i][1][2]],[]]
                else # abs(denom(A[i][1][1])*denom(A[i][1][2]))=abs(denom(A[i][1][3])) or
                     # abs(denom(A[i][1][1]))=abs(denom(A[i][1][3]))
                   B[i]:= [[A[i][1][2],A[i][1][3],A[i][1][1]],[]]
                fi;
             else
                B[i]:= A[i]
             fi;
          od;
          return convert(B,'list');
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext).
Goodexpbi2:=proc(R,x,t,ext)
          local Dx,l,n,A,B,s,i,j,e,C;
          Dx,x:=op(_Envdiffopdomain);
          n:=nops(R[3][1][1]);
          if n=1 then
             return {R[3][1]}
          else
             A:= Goodexpbi1(R[3][1][3],x,t,ext);
             B:= map(s-> `if`(nops(s[1])=3 and s[1][1]<>s[1][2],{s,[[s[1][2],s[1][1],s[1][3]],[]]},{s}),A);
             C:={[seq(e[i],i=1..n)]};
             for i from 1 to n do
                C:=map(s-> seq(subs(e[i]=B[i][j],s),j=1..nops(B[i])),C)
             od:
             return {seq(subs({seq(R[3][1][3][i]=l[i],i=1..n)},R[3][1]),l in C)}
         fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,ext).
find1F2Ratf:= proc(L,R,info,x,t,T,ext)
           local  Dx,Newp,s, l, dnu,dv, d1,d2, res, i, knl, j, k,f,ff,B,B1,n,SR,dA,Sreg,h,g,H;
           Dx,x:=op(_Envdiffopdomain);
           res:={};
           Newp:=Goodexpbi2(R,x,t,ext);
           for Sreg in Newp do
              dA:= info[2];
              if Sreg=[] then
                 dnu := {};
                 for i from 2 to floor(dA/2) do
                    if irem(dA,i)=0 then
                       dnu:= dnu union {i}
                    fi;
                 od;
                 dnu := dnu union {dA};
              else
                 dnu := {};
                 dv := denom(Sreg[3][1][1][1]);
                 for i from 1 to nops(Sreg[1]) do
                    dv:=lcm(dv,denom(Sreg[3][i][1][1]));
                 od;
                 if dv >= dA then
                    for i from 2 to dA*dv do
                       if irem(i,dv)=0 then
                          dnu := dnu union {i};
                       fi;
                    od;
                 else
                    for i from 1 to iquo(dA,dv) do
                       d1:=0;
                       d2:=0;
                       for s from 1 to nops(Sreg[1]) do
                          d1:=d1+i*dv/denom(Sreg[3][s][1][1])*(`if`(Sreg[1][s][1]=infinity,1,
                                                               degree(Sreg[1][s][1],x)));
                          d2 := gcd(d2, i*dv/denom(Sreg[3][s][1][1]));
                       od;
                       if d1<=dA and irem(dA,d2)=0 then
                          dnu := dnu union {i*dv};
                       fi;
                    od;
                 fi;
              fi;
              B:= info[4];
              for i in dnu do
                 knl := search1F2A1k(Sreg,dA,i,x);
                 if knl={} then
                    next
                 fi;
                 ff := help1F2f(L,info,i,knl,x,t,T,ext);
                 ff := simplef(ff);
                 if assigned(_Env_original_field) then
                    f := {};
                    for k in ff do
                        if indets(k, RootOf) minus _Env_original_field ={} then
                           f:=f union {k};
                        fi;
                     od;
                  else
                     f:= map(j->factor(j),ff);
                  fi;
                  res := res union f
              od;
           od;
           if res = {} then
              return NULL
           else
              H:= table();
              for g in res do
                 B1:= Zeroesmultordpol1F2(g,x);
                 if B1=NULL then
                    next
                 fi;
                 n:= nops(B1);
                 SR:= newregsing1F2(L,t,x,B1,R[2],n,ext);
                 H[g] := [g,SR[2],SR[1]];
              od;
              H:= convert(H,'set');
              return H
           fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for a1,a2,b1,b2 and f: [{{a1,a2}},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
find1F2Rat:=proc(L,R,info,x,t,T,ext)
                  local Dx,B,ab,f,C,H,E,ext1,n1,ext2,ext3,ext4,i,n2,n,s,G1;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= find1F2Ratf(L,R,info,x,t,T,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  if R[2]=[] then
                     ext1:= {}
                  else
                     ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                     ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                     n1:=nops(R[2][1]);
                     if n1 > 1 then
                        for i from 2 to n1 do
                           ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                           ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                           ext1:= ext1 intersect ext2;
                        od;
                     fi;
                  fi;
                  ext3:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][1],{RootOf,name}));
                  ext3:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext3,{RootOf,name}));
                  n2:=nops(R[1][1]);
                  if n2 > 1 then
                     for i from 2 to n2 do
                        ext4:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[1][3][i],{RootOf,name}));
                        ext4:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext4,{RootOf,name}));
                        ext3:= ext3 intersect ext4;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi1F2(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        ab:= findcandai1F2(L,R[1],C[1],x,t,ext,ext3);
                        if ab <> NULL then
                           H[f]:= [op(ab),C[2]]
                        fi;
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature 1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1, l2, p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

# E = Singular(L,ext);
# A = NotAppSing(L,E,ext);
# r = irrsing1F2(L,t,A,ext);
# Sreg = regsingtrue1F2(L,t,r[-1],ext);
# RSreg = Sregseptrue1F2(L,Sreg,ext);
# info = Sirr1F2info1(L,r,Sreg,x,ext).
help1F2Solutions1:= proc(L,R,info,x,t,T,ext)
           local Dx,n1,ext1,ext2,i,H,H1;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H1:= find1F2Rat(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easy1F2(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           elif R[3][3]<>[] then
              H1:= find1F2ln(L,R,info,x,t,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easy1F2(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           elif R[3][2]<>[] then
              H1:= find1F2Rat(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easy1F2(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           else
              ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
              ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
              n1:=nops(R[2][1]);
              if n1 > 1 then
                 for i from 2 to n1 do
                    ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                    ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                    ext1:= ext1 intersect ext2;
                 od;
              fi;
              if ext1 <> {} then
                 H1:= find1F2Irr(L,R,info,x,t,ext);
                 if H1=NULL then
                    if info[3]>info[2] then
                       H:= easy1F2(L,R,info,x,t,ext);
                    else
                       return NULL
                    fi;
                 else
                    H:= H1
                 fi;
              else
                 H1:= find1F2Rat(L,R,info,x,t,T,ext);
                 if H1=NULL then
                    if info[3]>info[2] then
                       H:= easy1F2(L,R,info,x,t,ext);
                    else
                       return NULL
                    fi;
                 else
                    H:= H1
                 fi;
              fi;
           fi:
           if H=NULL then
              return NULL
           else
              H:= {op(map(s-> op([s,subs(s[-1]=-s[-1],s)]),H))};
              return [op(H)]
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsing1F2(L,t,G,ext);
# a = element of help1F2Solutions1(L,r,RSreg,info,ext,x).
help1F2Solutions2:= proc(L,a,A,x,t,ext)
              local Dx,J,j,n2,C1,M,L22,k,R,i,e1,e2,e3,L12,T;
              Dx,x:=op(_Envdiffopdomain);
              n2:= nops(a);
              C1:= sumdiffeq(hyperterm([e1],[e2,e3],x,k),k,J(x));
              L12:= de2diffop(C1,J(x));
              for i from 1 to n2-1 do
                 for j in a[i][1] do
                    L22:=subs({e1=j,e2=a[i][2][1],e3=a[i][2][2]},L12);
                    M:= ChangeOfVariables(L22,a[n2]);
                    R:=EquivExpgauge1F2(M,L,A,x,t,T,ext);
                    if R<>NULL then
                       return [{[[j],a[i][2],R]},a[n2]]
                    fi;
                 od;
              od;
              return NULL
end:

############################################################################################################
############################################################################################################

help1F2Solutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

help1F2Solutions4:= proc(a,b)
            local Dx,x,n1,n2;
            Dx,x:=op(_Envdiffopdomain);
            if a[3]<>b[3] then
               return [false,b]
            elif Matchlistmod1(a[1],b[1])=false or Matchlistmod2(a[2],b[2])=false then
               return [false,b]
            else
               n1:= length(a[4]);
               n2:= length(b[4]);
               if n2<n1 then
                  return [true,b]
               else
                  return [true,a]
               fi;
            fi;
end:

############################################################################################################
############################################################################################################

help1F2Solutions5:= proc(G,b)
            local Dx,x,n1,n2,n,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n1:= nops(G);
            n2:= 0;
            for i from 1 to n1 do
               E:= help1F2Solutions4(G[i],b);
               if E[1]=false then
                  n2:= n2+1
               else
                  n:= i;
                  break
               fi;
            od;
            if n2=n1 then
              return [false,b]
            else
               return [true,[E[2],n]]
            fi;
end:

############################################################################################################
############################################################################################################

help1F2Solutions6:= proc(H)
            local Dx,x,n,G,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n:= nops(H);
            G:= table();
            G[1]:= H[1];
            if n>1 then
               for i from 2 to n do
                  E:= help1F2Solutions5(G,H[i]);
                  if E[1]=true then
                     G[E[2][2]]:= E[2][1]
                  else
                     G[i]:= E[2]
                  fi;
               od;
            fi;
            G:= convert(G,'set');
            return G
end:

############################################################################################################
############################################################################################################

Find1F2Solutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsing1F2(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= Sirr1F2info1(L,R[1],R[2],x,t,ext);
           if F=NULL then
              return NULL
           fi;
           C:= table();
           H:= help1F2Solutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= help1F2Solutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

Hyp1F2Solutions:= proc(L)
           local Dx,x,s,k,F,E,H1,H2,G,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:= table();
           H1:= Find1F2Solutions(L);
           if H1=NULL then
              return "No 1F2 type solutions"
           fi;
           F:= table();
           G:= map(s->s[2],H1);
           for i in G do
              H2:= [[op(map(s->`if`(s[2]=i,op(s[1]),NULL),H1))],i] ;
              F[i]:= H2
           od;
           F:= convert(F,'set');
           for C in F do
              E:= map(s->subs(s[3]=map(k->help1F2Solutions3(k),s[3]),s),C[1]);
              E:= map(s->seq([s[1],s[2],op(s[3][i])],i=1..nops(s[3])),E);
              E:= help1F2Solutions6(E);
              H[C]:= [E,C[2]]
           od;
           H:= convert(H,'set');
           return H
end:



############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################
########################################### 0F2 TYPE SOLUTIONS #############################################
############################################################################################################
############################################################################################################
############################################################################################################
############################################################################################################



# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing0F2(L2,t,G,ext);      ####### L2 can be our L operator;
# C1 = Exptest0F2equiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# i = index of an element in C2[5][1];     ####### C2[5][1] = set of non-apparent singularities of L2.
towopegenexpdiff0F2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,A1,A2,A3,extp,m,d1,d2,d,n,E1,E2,E3,E,k,j,s,H,H1,c1,c;
                  Dx,x:=op(_Envdiffopdomain);
                  d:= evala(Reduce((C1[2][i][1]+C1[2][i][2]+C1[2][i][3]
                                    -C2[5][2][i][1]-C2[5][2][i][2]-C2[5][2][i][3])/3));
                  m:= max(C1[5][i]);
                  d1:= {seq(d-s/(3*m),s=0..3*m-1)};
                  E1:={};
                  A1:= table();
                  for j in d1 do
                     A2:= table();
                     n:=0;
                     for k from 1 to 3 do
                        E:= normal([seq(evala(Reduce(m*(C1[2][i][k]-C2[5][2][i][s]-j))),s=1..3)]);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),[seq(k,k=1..3)]);
                        if E=[] then
                           break
                        fi;
                        n:=n+1;
                        A2[k]:=[k,E]
                     od;
                     if n=3 then
                        E1:= E1 union {j};
                        A1[j]:= convert(A2,'list');
                     fi;
                  od;
                  if E1 = {} then
                     return NULL
                  else
                     A1:= convert(A1,'list');
                     A3:= equiva0F22(A1,x,t,ext);
                     if A3=NULL then
                        return NULL
                     fi;
                     d2:=map(s-> op([evala(Reduce(C2[5][2][i][s[1][2]]-C1[2][i][s[1][1]])),
                                     evala(Reduce(C2[5][2][i][s[2][2]]-C1[2][i][s[2][1]])),
                                     evala(Reduce(C2[5][2][i][s[3][2]]-C1[2][i][s[3][1]]))]),A3);
                     if C2[5][1][i][1]=infinity then
                        H:={};
                        if member(t,indets(d)) then
                           for c in d2 do
                              c1:= convert(series(subs(t=1/x,c),x=0,100),polynom);
                              c1:= eval(subs(csgn=1,simplify(c1) assuming(x>0)));
                              c1:= convert(series((c1-subs(x=0,c1))/x,x=0,100),polynom);
                              c1:=convert(eval(subs(csgn=1,simplify(c1) assuming(x>0))),polynom);
                              H:= H union {-c1}
                           od;
                        else
                           H:= {0}
                        fi;
                     else
                        E2:=[op(d2)];
                        H1:=map(s-> subs(t=`if`(degree(C2[5][1][i][1],x) > 1,x-C2[5][1][i][2],C2[5][1][i][1]),
                                         convert(series(evala(E2[s])/t,t=0,100),polynom)),[seq(k,k=1..nops(E2))]);
                        H1:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(x>0))),polynom),H1);
                        extp:=map(s-> equiva0F21(C2,s,i,x,t,ext),H1);
                        H1:= map(s-> evala(Trace(H1[s],extp[s],extp[s]),x),[seq(k,k=1..nops(E2))]);
                        H:= {op(H1)};
                     fi;
                     return {op(H),0}
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# R2 = IrrRegAppsing0F2(L2,t,G,ext);       ####### L2 can be our L operator.
Exptest0F2equiv1:=proc(L1,R2,x,t,ext)
                  local Dx,d1,A1,A2,A3,A4,A5,G,i,H;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:= table();
                  A2:= table();
                  A3:= table();
                  A4:= table();
                  A5:= table();
                  for i in R2[5][1] do
                     G:= genexpdiff(L1,i,t,ext);
                     if member(t,indets(G[2][1])) or member(t,indets(G[2][2])) or
                        member(t,indets(G[2][3])) then
                        d1:= Condi0F2irr(G,t,ext);
                        if d1 = NULL then
                          return NULL
                        fi;
                        A1[i]:=i;
                        A2[i]:=G[1];
                        A3[i]:=d1[1];
                        A4[i]:=d1[2];
                        A5[i]:=G[3];
                     else
                        A1[i]:= i;
                        A2[i]:= G[1];
                        A3[i]:= G[2];
                        A4[i]:= G[4];
                        A5[i]:= G[3];
                     fi;
                  od;
                  A1:= convert(A1,'list');
                  A2:= convert(A2,'list');
                  A3:= convert(A3,'list');
                  A4:= convert(A4,'list');
                  A5:= convert(A5,'list');
                  H:= table();
                  H[1]:= A1;
                  H[2]:= A2;
                  H[3]:= A3;
                  H[4]:= A4;
                  H[5]:= A5;
                  return convert(H,'list');
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing0F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest0F2equiv2(L1,C2,x,t,ext);       ####### L1 can be an operator which is gauge equivalent to L2=L;
# i = index of an element in C2[5][1].
Exptest0F2equiv2:=proc(L1,L2,C1,C2,i,x,t,ext)
                  local Dx,m,A1,A2,E,k,j,s,C;
                  Dx,x:=op(_Envdiffopdomain);
                  m:= max(C1[5][i]);
                  A1:=[[C1[2][i][1],C1[2][i][2],C1[2][i][3]],[C1[2][i][1],C1[2][i][3],C1[2][i][2]],
                       [C1[2][i][2],C1[2][i][1],C1[2][i][3]],[C1[2][i][2],C1[2][i][3],C1[2][i][1]],
                       [C1[2][i][3],C1[2][i][1],C1[2][i][2]],[C1[2][i][3],C1[2][i][2],C1[2][i][1]]];
                  A2:=[[C2[5][2][i][1],C2[5][2][i][2],C2[5][2][i][3]],[C2[5][2][i][1],C2[5][2][i][3],C2[5][2][i][2]],
                       [C2[5][2][i][2],C2[5][2][i][1],C2[5][2][i][3]],[C2[5][2][i][2],C2[5][2][i][3],C2[5][2][i][1]],
                       [C2[5][2][i][3],C2[5][2][i][1],C2[5][2][i][2]],[C2[5][2][i][3],C2[5][2][i][2],C2[5][2][i][1]]];
                  C:=[seq(j,j=1..3)];
                  for j in A1 do
                     for k in A2 do
                        E:= map(s-> evala(Reduce(m*(j[s]-k[s]))),C);
                        E:= map(s-> `if`(type(E[s],integer),s,NULL),C);
                        if nops(E)=3 then
                           return true
                        fi;
                     od;
                  od;
                  return false
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing0F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest0F2equiv2(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# i = index of an element in C2[5][1].
equiva0F21:=proc(C2,A,i,x,t,ext)
               local Dx,extp0,extp1,extp2,extp3,extp4,extp5,extp6,extp7,s,v1,v;
               Dx,x:=op(_Envdiffopdomain);
               extp0:= indets(ReplirrRoot(C2[5][1][i][2],{}),{RootOf,name});
               extp1:= indets(A,{RootOf,name}) minus {x};
               extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
               extp2:=indets({op(ext),ReplirrRoot(C2[5][1][i][2],{})},{RootOf,name});
               extp3:= extp1 minus extp2; # New names or Rootof which appear.
               if extp0 subset ext then
                  return extp3 union ext
               fi;
               extp4:= extp1 minus extp0; # All the names or RootOf
                                          # different from RootOf(C2[5][1][i][1],x) in the expression of A.
               extp6:= ext minus extp0;
               extp7:= extp4 union extp6;
               v1:={};
               for v in extp7 do
                  if extp0 intersect indets(v,{RootOf,name}) <> {} then
                     v1:= v1 union {v}
                  fi;
               od:
               return extp7 minus v1;
end:

############################################################################################################
############################################################################################################

equiva0F22:=proc(E,x,t,ext)
                  local Dx,A,A1,A2,s1,s2,s3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  A1:=table();
                  for A in E do
                     A2:=table();
                     for s1 in A[1][2] do
                        if {op(A[2][2])} minus {s1}={} then
                           next
                        else
                           for s2 in {op(A[2][2])} minus {s1} do
                              if {op(A[3][2])} minus {s1,s2}={} then
                                 next
                              else
                                 C:={op(A[3][2])} minus {s1,s2};
                                 s3:= C[1];
                                 A2[s1,s2,s3]:= [[A[1][1],s1],[A[2][1],s2],[A[3][1],s3]];
                              fi;
                           od;
                        fi;
                     od;
                     A2:= convert(A2,'list');
                     if A2<>[] then
                        A1[A]:=op(A2)
                     fi;
                  od;
                  A1:= convert(A1,'set');
                  if A1<>{} then
                     return A1
                  else
                     return NULL
                  fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing0F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest0F2equiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff0F2(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge0F21:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s,RS;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            RS:=table();
            for i in E do
               M:= ExpProduct(L1,i);
               if (collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0) and
                   member([[factor(i)],[1]],convert(RS,'set'))=false then
                  RS[i]:= [[factor(i)],[1]];
               else
                  C3:=ExptestBesSqRootequiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(ExptestBesSqRootequiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] and member([[factor(i)],E2],convert(RS,'set'))=false   then
                     RS[i]:= [[factor(i)],E2];
                  fi;
               fi;
            od:
            RS:= convert(RS,'set');
            if RS<>{} then
               return RS
            else
               return NULL
            fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext).
EquivExpgauge0F21:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest0F2equiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff0F2(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge0F21(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# L1 = M;
# L2 = L;
#
# ext = indets(L2,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L2,extppp);
# C2 = IrrRegAppsing0F2(L2,t,G,ext);       ####### L2 can be our L operator;
# C1 = Exptest0F2equiv1(L1,C2,x,t,ext);       ####### L1 can be our M operator;
# A = map(s-> towopegenexpdiff0F2(L1,L2,C1,C2,s,x,t,ext),[seq(k,k=1..n)]);
# n = nops(C2[5][1]).
helpEquivExpgauge0F2:=proc(L1,L2,C2,A,n,x,t,ext)
            local Dx,C3,C4,k,M,a,b,E,j,H1,E1,E2,i,s;
            Dx,x:=op(_Envdiffopdomain);
            # For a singularity p=RootOf(i,x) of L1 and L2, let e[i] be the difference of the generalized
            # exponent
            # of L1 and L2 at p. Then r=sum(e[i]/(x-p))-e*[infinity]/x where
            # e*[infinity]=e[infinity]-const(e[infinity]).
            # If e[i]<> e[infinity] then e[i] don't have the constant part. e*[infinity] doesn't have constant
            # part.
            E:= A[1];
            if n>1 then
               for j from 2 to n do
                  E1:= table();
                  H1:= A[j];
                  for a in H1 do
                     for b in E do
                        E1[a,b]:=  evala(a+b)
                     od;
                  od;
                  E1:= convert(E1,'set');
                  E:= E1;
               od;
            fi;
            E:=map(s-> factor(s),E);
            if E<>{0} then
               E:= map(s-> `if`(indets(s,{RootOf,name}) minus {x}={} and s<>0,s,NULL),E);
            fi;
            if E={} then
               return NULL
            fi;
            # E is the set of candidates for exp-Product transformations parameters r.
            # Here ExpProduct(L1,r) give us the differential operator M obtained from L1 by the
            # exp-Product transformation with parameter r. Homomorphisms(M,L2) is a command which give us the
            # basis of parameters for the gauge transformation between M and L2.
            for i in E do
               M:= ExpProduct(L1,i);
               if collect(evala(Reduce(L2-M)),Dx,factor)=0 or collect(evala(Rem(L2,M,Dx)),Dx,factor)=0 then
                  return {[[factor(i)],[1]]};
               else
                  C3:=Exptest0F2equiv1(M,C2,x,t,ext);
                  C4:= map(s-> `if`(Exptest0F2equiv2(M,L2,C3,C2,s,x,t,ext),s,NULL),[seq(k,k=1..n)]);
                  if nops(C4)<>n then
                     next
                  fi;
                  E2:= Homomorphisms(M, L2);
                  if E2<> [] then
                     return {[[factor(i)],E2]};
                  fi;
               fi;
            od:
            return NULL
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext).
EquivExpgauge0F2:= proc(M,L,R,x,t,T,ext)
                 local Dx,k,A1,A2,A21,n,H;
                 Dx,x:=op(_Envdiffopdomain);
                 A1:=Exptest0F2equiv1(M,R,x,t,ext);
                 n:=nops(R[5][1]);
                 A2:= table();
                 for k from 1 to n do
                    A21:= towopegenexpdiff0F2(M,L,A1,R,k,x,t,ext);
                    if A21 = NULL then
                       break
                    fi;
                    A2[k]:= A21;
                 od;
                 A2:= convert(A2,'list');
                 if nops(A2) <> n then
                    return NULL
                 fi;
                 if collect(evala(Reduce(L-M)),Dx,factor)=0 or
                    collect(evala(Rem(L,M,Dx)),Dx,factor)=0 then
                    H:={[[0],[1]]}
                 else
                    H:= helpEquivExpgauge0F2(M,L,R,A2,n,x,t,ext);
                 fi;
                 if H<>NULL then
                    return H
                 else
                    return NULL
                 fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 0F2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi0F2irr:=proc(E,t,ext)
              local Dx,x,A,s,B,C,C1,C2,l,R,R1,i;
              Dx,x:=op(_Envdiffopdomain);
              C1:=map(s-> indets(s,{RootOf,name}),E[2]);
              C2:=map(s-> `if`(member(t,s)=true,s,NULL),C1);
              # Here all the exponent differences have the variable t.
              if nops(C2) <> 3 then
                 return NULL
              fi;
              l:= max(E[3][1],E[3][2],E[3][3]);
              C:= map(s-> convert(series(subs(t=1/t^l,s),t=0,100),polynom),E[2]);
              C:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),C);
              A:=map(s-> evala(s-coeff(s,t,0)),C);
              B:=map(s-> evala(coeff(s,t,0)),C);
              R:=[solve(x^3+1 = 0, x)];
              R1:=[[R[1],R[2],R[3]],[R[1],R[3],R[2]],[R[2],R[1],R[3]],[R[2],R[3],R[1]],
                   [R[3],R[1],R[2]],[R[3],R[2],R[1]]];
              if type(evala((B[1]-B[2])*3),integer) and type(evala((B[1]-B[3])*3),integer) and
                 type(evala((B[2]-B[3])*3),integer) then
                 for i in R1 do
                    if evala(((i[2])^(-1)-(i[1])^(-1))*A[1]-((i[3])^(-1)-(i[1])^(-1))*A[2])=0 and
                       evala(((i[2])^(-1)-(i[1])^(-1))*A[1]-((i[3])^(-1)-(i[2])^(-1))*A[3])=0 then
                       return [E[2],E[4],A,B]
                    fi;
                 od;
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the non-apparant singularities of an operator L in two sets: one containing all the irregular
# singularities of L and the other its regular singularities. This algorithm "irrsing0F2".
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# H = NotAppSing(L,G,ext).
irrsing0F2:=proc(L,t,H,ext)
              local Dx,x,n,Z,d,A,B,C,F,F1,F2,F3,i,m,G,G1,E,H1;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              Z:=table();
              G:=table();
              G1:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              n:=table();
              for i in H do
                 A[i[1]]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i[1]][2][1])) or member(t,indets(A[i[1]][2][2])) or
                    member(t,indets(A[i[1]][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d:= Condi0F2irr(A[i[1]],t,ext);
                    if d = NULL then
                       return NULL
                    fi;
                    B[i]:= i;
                    C[i]:= A[i[1]][1];
                    F[i]:= d[1];
                    F3[i]:=d[4];
                    F1[i]:=d[2];
                    F2[i]:= d[3];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n[i]:=max(A[i[1]][3][1],A[i[1]][3][2],A[i[1]][3][3]);
                    E[i[1]]:=[seq(convert(series(subs(t=1/t^(n[i]),A[i[1]][2][j]),t=0,100),polynom),j=1..3)];
                    E[i[1]]:=map(s-> convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E[i[1]]);
                    m[i[1]]:=map(s->degree(s,t),E[i[1]]);
                    Z[i]:= max(m[i[1]])/(n[i]);
                 else
                    G[i]:= i;
                    G1[i]:= A[i[1]];
                 fi;
              od;
              B:= convert(B,'list');
              G:= convert(G,'list');
              G1:= convert(G1,'list');
              if B=[] then
                 return [[],[G,G1]]
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 Z:= convert(Z,'list');
                 F1:= convert(F1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 n:= convert(n,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=Z;
                 H1[5]:=n;
                 H1[6]:=F1;
                 H1[7]:=F2;
                 H1[8]:=F3;
                 H1[9]:=[G,G1];
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# That is a condition in the 0F2 case (only in this case) related to the exponent differences at a
# irregular singular point of an third order irreducible linear differential operator comming from the Bessel
# square operator by:
#      -change of variables and exp-Product transformations or
#      -change of variables and gauge transformations or
#      -change of variables, exp-Product and gauge transformations.
#
#
#
# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# i = [s,pol] where s is the minimal polynomial associated to an irregular singularity p=pol;
# E = genexpdiff(L,i,t,ext).
Condi0F2reg:=proc(E,t,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if type(evala(2*(E[2][2]-E[2][1]-E[2][3])),integer) then
                 return [[E[2][1],E[2][2],E[2][3]],[E[4][1],E[4][2],E[4][3]]]
              elif type(evala(2*(E[2][1]-E[2][2]-E[2][3])),integer) then
                 return [[E[2][2],E[2][1],E[2][3]],[E[4][2],E[4][1],E[4][3]]]
              elif type(evala(2*(E[2][3]-E[2][1]-E[2][2])),integer) then
                 return [[E[2][1],E[2][3],E[2][2]],[E[4][1],E[4][3],E[4][2]]]
              else
                 return NULL
              fi;
end:

############################################################################################################
############################################################################################################

# Compute the regular singularities "regsingtrue" of L: they belong to the set of the zeroes of f. But
# we don't know if they represent all the zeroes of f. The output is in the form [[[pol,a]],[[e]],[[Delta]]] where
# a = the regular singularity;
# pol = minimal polynomial of a;
# e = generalized exponents at a;
# Delta = exponent differences at a.
#
#
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# H = irrsing0F2(L,t,F,ext)[-1] or IrrRegAppsing0F2(L,t,G,ext)[1][-1].
regsingtrue0F2:=proc(L,t,H,ext)
              local Dx,x,B,C,F,i,H1,E,n,G;
              Dx,x:=op(_Envdiffopdomain);
              B:=table();
              C:=table();
              F:=table();
              G:=table();
              n:= nops(H[1]);
              for i from 1 to n do
                 # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                 # must not contain the parameter t.
                 if ((H[2][i][2][1]=1 and H[2][i][2][2]=2 and H[2][i][2][3]=1) or
                    (H[2][i][2][1]=2 and H[2][i][2][2]=4 and H[2][i][2][3]=2)) and
                    formal_sol(L,`has logarithm?`,x=H[1][i][2])= false then
                    next
                 else
                    E:= Condi0F2reg(H[2][i],t,ext);
                    if E <> NULL then
                       B[i]:= H[1][i];
                       C[i]:= H[2][i][1];
                       F[i]:= E[1];
                       G[i]:= E[2]
                    else
                       return NULL
                    fi;
                 fi;
              od;
              B:= convert(B,'list');
              if B=[] then
                 return []
              else
                 C:= convert(C,'list');
                 F:= convert(F,'list');
                 G:= convert(G,'list');
                 H1:= table();
                 H1[1]:=B;
                 H1[2]:=C;
                 H1[3]:=F;
                 H1[4]:=G;
                 H1:= convert(H1,'list');
                 return H1
              fi;
end:

############################################################################################################
############################################################################################################

# We divide the regular singularities of L in three parts:
#       - non-removable singularities with non-logarithmic solutions (at least one exponent differences is not
#                                                 an integer),
#       - removable singularities with integer exponent differences,
#       - regular singularities with logarithmic solutions.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# Sirr = irrsing0F2(L,t,F,ext) or IrrRegAppsing0F2(L,t,G,ext)[1];
# H = regsingtrue0F2(L,t,Sirr[-1],ext) or IrrRegAppsing0F2(L,t,G,ext)[2].
Sregseptrue0F2:=proc(L,H,ext)
              local Dx,x,A1,A2,A3,A31,A32,B1,B2,B3,B31,B32,A,B,n,C1,C2,C3,C31,C32,C,k,i,H1,m;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              B1:=table();
              B2:=table();
              B3:=table();
              C1:=table();
              C2:=table();
              C3:=table();
              if H=[] then
                 return [[],[],[]]
              fi;
              m:=nops(H[1]);
              for i from 1 to m do
                 if formal_sol(L,`has logarithm?`,x=H[1][i][2])= true then
                    C1[i]:= H[1][i];
                    C2[i]:= H[2][i];
                    C31:= table();
                    C32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          C31[n]:= k;
                       else
                          C32[n]:= k;
                       fi;
                    od;
                    C31:= convert(C31,'list');
                    C32:= convert(C32,'list');
                    C3[i]:= [C32,C31];
                 elif type(H[3][i][1],integer)=true and type(H[3][i][2],integer)=true and
                      type(H[3][i][3],integer)=true then
                    B1[i]:= H[1][i];
                    B2[i]:= H[2][i];
                    B3[i]:= [[],H[3][i]];
                 else
                    A1[i]:= H[1][i];
                    A2[i]:= H[2][i];
                    A31:= table();
                    A32:= table();
                    n:=0;
                    for k in H[3][i] do
                       n:= n+1;
                       if type(k,integer) then
                          A31[n]:= k;
                       else
                          A32[n]:= k;
                       fi;
                    od;
                    A31:= convert(A31,'list');
                    A32:= convert(A32,'list');
                    A3[i]:= [A32,A31];
                 fi;
              od;
              A1:= convert(A1,'list');
              A2:= convert(A2,'list');
              A3:= convert(A3,'list');
              B1:= convert(B1,'list');
              B2:= convert(B2,'list');
              B3:= convert(B3,'list');
              C1:= convert(C1,'list');
              C2:= convert(C2,'list');
              C3:= convert(C3,'list');
              A:=[A1,A2,A3];
              B:=[B1,B2,B3];
              C:=[C1,C2,C3];
              if A1=[] then
                 A:=[]
              fi;
              if B1=[] then
                 B:=[]
              fi;
              if C1=[] then
                 C:=[]
              fi;
              H1:= table();
              H1[1]:=A;
              H1[2]:=B;
              H1[3]:=C;
              H1:= convert(H1,'list');
              return H1
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# Sirr = irrsing0F2(L,t,F,ext) or IrrRegAppsing0F2(L,t,H,ext)[1];
# Sreg = regsingtrue0F2(L,t,Sirr[-1],ext) or IrrRegAppsing0F2(L,t,H,ext)[2];
# RSreg = Sregseptrue0F2(L,Sreg,ext) or IrrRegAppsing0F2(L,t,H,ext)[3].
UnionSregseptrue0F2:=proc(RSreg)
                  local n1,B,C1,C2,C;
                  if RSreg[1]<>[] and RSreg[2]<>[] then
                     n1:= nops(RSreg[1][1]);
                     B:=[op(RSreg[1][1]),op(RSreg[2][1])];
                     C1:= map(s->s[1],RSreg[1][3]);
                     C2:= map(s->s[2],RSreg[2][3]);
                     C:=[op(C1),op(C2)];
                  elif RSreg[1]<>[] and RSreg[2]=[] then
                     n1:= nops(RSreg[1][1]);
                     B:= RSreg[1][1];
                     C:= map(s->s[1],RSreg[1][3]);
                  elif RSreg[1]=[] and RSreg[2]<>[] then
                     n1:= 0;
                     B:= RSreg[2][1];
                     C:= map(s->s[2],RSreg[2][3]);
                  else       ####### RSreg[1]=[] and RSreg[2]=[] then
                     n1:= 0;
                     B:= []
                  fi;
                  return [B,n1,C]
end:

############################################################################################################
############################################################################################################

# Delta = exponent differences of the singularity.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# G = Singular(L,extppp);
# F = NotAppSing(L,G,ext);
# Sirr = irrsing0F2(L,t,F,ext) or IrrRegAppsing0F2(L,t,G,ext)[1];
# Sreg = regsingtrue0F2(L,t,Sirr[-1],ext) or IrrRegAppsing0F2(L,t,G,ext)[2].
AppSing:=proc(L,Sreg,Sirr,G,ext)
              local Dx,x,i,v,H,E,F,A,T;
              Dx,x:=op(_Envdiffopdomain);
              A:= {op(G)} minus {op(Sreg[1]),op(Sirr[1])};
              E:=table();
              H:=table();
              F:=table();
              if A <> {} then
                 for i in A do
                    v[i[1]]:=DEtools['gen_exp'](L,T,x=i[2],'groundfield'=ext);
                    E[i]:= i;
                    H[i]:= [v[i[1]][1][1],v[i[1]][1][2],v[i[1]][1][3]];
                    F[i]:= [v[i[1]][1][2]-v[i[1]][1][1],v[i[1]][1][3]-v[i[1]][1][1],
                          v[i[1]][1][3]-v[i[1]][1][2]];
                 od;
              fi;
              E:= convert(E,'list');
              H:= convert(H,'list');
              F:= convert(F,'list');
              return [E,H,F]
end:

############################################################################################################
############################################################################################################

# ext1 = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# R = Singular(L,extppp).
IrrRegAppsing0F2:= proc(L,t,R,ext)
              local Dx,x,A,A1,A2,A3,A4,A5,B1,B2,C1,C2,d1,E,E1,F1,F2,F3,F4,G,G1,G2,G3,G31,G32,H,H1,H2,H3,i,k,H5,
                    H4,M,M1,M2,M3,M31,M32,m,m1,N,N1,N2,N3,n,n1,R3,V2,V3,W1,W2,W3,W4,Z1;
              Dx,x:=op(_Envdiffopdomain);
              A1:=table();
              A2:=table();
              A3:=table();
              A4:=table();
              A5:=table();
              B1:=table();
              B2:=table();
              C1:=table();
              C2:=table();
              Z1:=table();
              G:=table();
              F1:=table();
              F2:=table();
              F3:=table();
              F4:=table();
              G1:=table();
              G2:=table();
              G3:=table();
              M1:=table();
              M2:=table();
              M3:=table();
              N1:=table();
              N2:=table();
              N3:=table();
              n1:=table();
              V2:=table();
              V3:=table();
              W1:=table();
              W2:=table();
              W3:=table();
              W4:=table();
              for i in R do
                 A[i]:=genexpdiff(L,i,t,ext);
                 # The condition for a point p=RootOf(i[1],x) to be an irregular singular point is that: at least
                 # one of its exponent differences contains the parameter t.
                 if member(t,indets(A[i][2][1])) or member(t,indets(A[i][2][2])) or
                    member(t,indets(A[i][2][3])) then
                    # We are here in the 1F2 case which has more other conditions for a point to be an irregular
                    # singularty of L
                    d1:= Condi0F2irr(A[i],t,ext);
                    if d1 = NULL then
                       return NULL
                    fi;
                    B1[i]:= i;
                    C1[i]:= A[i][1];
                    F1[i]:= d1[1];
                    F2[i]:=d1[2];
                    F3[i]:= d1[3];
                    F4[i]:=d1[4];
                    # We compute a list of the candidates for the multiplicity order of the irregular singular
                    # point p:=RootOf(i[1],x) using the exponent differences and the ramification index at this
                    # point.
                    n1[i]:=max(A[i][3][1],A[i][3][2],A[i][3][3]);
                    E1[i]:=[seq(convert(series(subs(t=1/t^(n1[i]),A[i][2][j]),t=0,100),polynom),j=1..3)];
                    E1[i]:=map(s->convert(eval(subs(csgn=1,simplify(s) assuming(t>0))),polynom),E1[i]);
                    m1[i]:=map(s->degree(s,t),E1[i]);
                    Z1[i]:= max(m1[i])/(n1[i]);
                    A1[i]:= i;
                    A2[i]:= A[i][1];
                    A3[i]:= d1[1];
                    A4[i]:=d1[2];
                    A5[i]:= A[i][3];
                 else
                    # For a point p=i[2]=RootOf(i[1],x) to be a regular singular point, its exponent differences
                    # must not contain the parameter t.
                    if A[i][5]=4 then
                       # We test now the last condition for a point to be an apparent singularity: all the
                       # generalized exponents have to be positive integers.
                       if (A[i][1][1]>=0 and A[i][1][2]>=0 and A[i][1][3]>=0) and
                          (type(A[i][1][1],integer) and type(A[i][1][2],integer) and
                          type(A[i][1][3],integer)) and formal_sol(L,`has logarithm?`,x=i[2])= false  then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          A1[i]:= i;
                          A2[i]:= A[i][1];
                          A3[i]:= A[i][2];
                          A4[i]:=A[i][4];
                          A5[i]:= A[i][3];
                          if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                             (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                             formal_sol(L,`has logarithm?`,x=i[2])= false then
                             W1[i]:= i;
                             W2[i]:=A[i][1];
                             W3[i]:=A[i][2];
                             W4[i]:=A[i][4];
                          else
                             E:= Condi0F2reg(A[i],t,ext);
                             if E <> NULL then
                                B2[i]:= i;
                                C2[i]:= A[i][1];
                                V2[i]:= E[1];
                                V3[i]:= E[2];
                                if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                   G1[i]:= i;
                                   G2[i]:= A[i][1];
                                   G31:= table();
                                   G32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         G31[n]:= k;
                                      else
                                         G32[n]:= k;
                                      fi;
                                   od;
                                   G31:= convert(G31,'list');
                                   G32:= convert(G32,'list');
                                   G3[i]:= [G32,G31];
                                elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                     type(A[i][2][3],integer)=true then
                                   N1[i]:= i;
                                   N2[i]:= A[i][1];
                                   N3[i]:= [[],A[i][2]];
                                else
                                   M1[i]:= i;
                                   M2[i]:= A[i][1];
                                   M31:= table();
                                   M32:= table();
                                   n:=0;
                                   for k in A[i][2] do
                                      n:= n+1;
                                      if type(k,integer) then
                                         M31[n]:= k;
                                      else
                                         M32[n]:= k;
                                      fi;
                                   od;
                                   M31:= convert(M31,'list');
                                   M32:= convert(M32,'list');
                                   M3[i]:= [M32,M31];
                                fi;
                             else
                                return NULL
                             fi;
                          fi;
                       fi;
                    else
                       A1[i]:= i;
                       A2[i]:= A[i][1];
                       A3[i]:= A[i][2];
                       A4[i]:=A[i][4];
                       A5[i]:= A[i][3];
                       if ((A[i][2][1]=1 and A[i][2][2]=2 and A[i][2][3]=1) or
                          (A[i][2][1]=2 and A[i][2][2]=4 and A[i][2][3]=2)) and
                          formal_sol(L,`has logarithm?`,x=i[2])= false then
                          W1[i]:= i;
                          W2[i]:=A[i][1];
                          W3[i]:=A[i][2];
                          W4[i]:=A[i][4];
                       else
                          E:= Condi0F2reg(A[i],t,ext);
                          if E <> NULL then
                             B2[i]:= i;
                             C2[i]:= A[i][1];
                             V2[i]:= E[1];
                             V3[i]:= E[2];
                             if formal_sol(L,`has logarithm?`,x=i[2])= true then
                                G1[i]:= i;
                                G2[i]:= A[i][1];
                                G31:= table();
                                G32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      G31[n]:= k;
                                   else
                                      G32[n]:= k;
                                   fi;
                                od;
                                G31:= convert(G31,'list');
                                G32:= convert(G32,'list');
                                G3[i]:= [G32,G31];
                             elif type(A[i][2][1],integer)=true and type(A[i][2][2],integer)=true and
                                  type(A[i][2][3],integer)=true then
                                N1[i]:= i;
                                N2[i]:= A[i][1];
                                N3[i]:= [[],A[i][2]];
                             else
                                M1[i]:= i;
                                M2[i]:= A[i][1];
                                M31:= table();
                                M32:= table();
                                n:=0;
                                for k in A[i][2] do
                                   n:= n+1;
                                   if type(k,integer) then
                                      M31[n]:= k;
                                   else
                                      M32[n]:= k;
                                   fi;
                                od;
                                M31:= convert(M31,'list');
                                M32:= convert(M32,'list');
                                M3[i]:= [M32,M31];
                             fi;
                          else
                             return NULL
                          fi;
                       fi;
                    fi;
                 fi;
              od;
              H1:= table();
              B1:= convert(B1,'list');
              if B1=[] then
                 H1:= convert(H1,'list');
              else
                 C1:= convert(C1,'list');
                 F1:= convert(F1,'list');
                 Z1:= convert(Z1,'list');
                 F2:= convert(F2,'list');
                 F3:= convert(F3,'list');
                 F4:= convert(F4,'list');
                 n1:= convert(n1,'list');
                 H1:= table();
                 H1[1]:=B1;
                 H1[2]:=C1;
                 H1[3]:=F1;
                 H1[4]:=Z1;
                 H1[5]:=n1;
                 H1[6]:=F2;
                 H1[7]:=F3;
                 H1[8]:=F4;
                 H1:= convert(H1,'list');
              fi;
              H2:= table();
              B2:= convert(B2,'list');
              if B2=[] then
                 H2:= convert(H2,'list');
                 H3:= [[],[],[]];
              else
                 C2:= convert(C2,'list');
                 V2:= convert(V2,'list');
                 V3:= convert(V3,'list');
                 H2[1]:=B2;
                 H2[2]:=C2;
                 H2[3]:=V2;
                 H2[4]:=V3;
                 H2:= convert(H2,'list');
                 R3:= table();
                 M1:= convert(M1,'list');
                 M2:= convert(M2,'list');
                 M3:= convert(M3,'list');
                 N1:= convert(N1,'list');
                 N2:= convert(N2,'list');
                 N3:= convert(N3,'list');
                 G1:= convert(G1,'list');
                 G2:= convert(G2,'list');
                 G3:= convert(G3,'list');
                 M:=[M1,M2,M3];
                 N:=[N1,N2,N3];
                 G:=[G1,G2,G3];
                 if M1=[] then
                    M:=[]
                 fi;
                 if N1=[] then
                    N:=[]
                 fi;
                 if G1=[] then
                    G:=[]
                 fi;
                 R3[1]:=M;
                 R3[2]:=N;
                 R3[3]:=G;
                 H3:= convert(R3,'list');
              fi;
              H4:= table();
              W1:= convert(W1,'list');
              if W1=[] then
                 H4:= convert(H4,'list');
              else
                 W2:= convert(W2,'list');
                 W3:= convert(W3,'list');
                 W4:= convert(W4,'list');
                 H4[1]:=W1;
                 H4[2]:=W2;
                 H4[3]:=W3;
                 H4[4]:=W4;
                 H4:= convert(H4,'list');
              fi;
              H5:= table();
              A1:= convert(A1,'list');
              if A1=[] then
                 H5:= convert(H5,'list');
              else
                 A2:= convert(A2,'list');
                 A3:= convert(A3,'list');
                 A4:= convert(A4,'list');
                 A5:= convert(A5,'list');
                 H5[1]:=A1;
                 H5[2]:=A2;
                 H5[3]:=A3;
                 H5[4]:=A4;
                 H5[5]:=A5;
                 H5:= convert(H5,'list');
              fi;
              H:= table();
              H[1]:=H1;
              H[2]:=H2;
              H[3]:=H3;
              H[4]:=H4;
              H[5]:=H5;
              H:= convert(H,'list');
              return H
end:

############################################################################################################
############################################################################################################
####### The ramification index at all irregular singular points p of L (p=RootOf(irrsingstruc(L,t,ext)[1][i],x),
#######                      with i=position of the minimal polynom of p) is 3.
############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g;
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,ext).
Hyperpossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:= table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     C[k]:= h;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# Input: a list F=[f1,...fn] of all the polar parts of g associated to its poles. n is the number of minimal
# polynomials associated to the poles of g;
# Output: the sum +-f1+-...+-fn, where the k-th possibility of signs is choosen (1<=k<=2^n).
# Here n = nops(A[1]) where A = irrsingBesSq(L,t,x,ext).
Bespossibility := proc(F)
               local Dx,x,C,n,k,i,e,c,d,h;
               Dx,x:=op(_Envdiffopdomain);
               C:=table();
               n := nops(F);
               for k from 1 to 2^n do
                     c := k-1;
                     for i from n by -1 to 1 do
                           e:=c;
                           c := iquo(c,2);
                           d[i]:=e-c*2;
                     od;
                     h:=add((-1)^d[i]*F[i],i=1..n);
                     if member(-h,{op(C)})=false and member(h,{op(C)})=false then
                           C[k]:= h;
                     fi;
               od:
               C:= convert(C,'set');
               return [op(C)]
end:

############################################################################################################
############################################################################################################

# We compute the product of (x-p)^(2*m) where p is an irregular singularity (pole) of f and m is its multiplicative
# order. That will give us the denominator of g=f^2.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# C = irrsing0F2(L,t,F,ext) or IrrRegAppsing0F2(L,t,H,ext)[1].
PolirrProd0F2:=proc(L,C)
            local Dx,x,b,i,s;
            Dx,x:=op(_Envdiffopdomain);
            b:=1;
            i:=0;
            for s in C[1] do
               i:=i+1;
               if s[1] <> infinity then
                  b:=b*(s[1])^(3*C[4][i])
               fi;
            od;
            return b
end:

############################################################################################################
############################################################################################################

degpol:=proc(a,x,t,ext)
          local Dx,l1,d;
          Dx,x:=op(_Envdiffopdomain);
          l1:=series(subs(t=1/x,a),x=0,100);
          l1:=eval(subs(csgn=1,simplify(l1) assuming(x>0)));
          d:= degree(convert(l1,polynom), x);
          if d=FAIL then
             return false
          else
             return d
          fi;
end:

############################################################################################################
############################################################################################################

# We take as input A = irrsingBesSq(L,t,ext) and l the minimal integer m such that f belong to k(x^(1/m)). We return a
# list of list in the form [[pol,s,d,acc],dA,neq,B,nonpole] where [pol,s,d,acc] is related to s with:
# s = an irregular singularity of L (pole of f);
# pol = polynomial over K with zero s;
# d = truncated series of g=f^3 at an irregular singularity s of L;
# acc = smallest integer greater than or equal to m_p/3;
# dA = degree bound for the numerator of f^3   (A = numer(f^3));
# neq = number of linear equations for the coefficients of A = numer(f^3);
# B = denominator of f^3;
# nonpole = denomintor of f but with integer power (the smallest integer greater than or equal to m_p/3);
# Here n_e=3.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = subs(x = _Z, indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name}) minus {x});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# A = irrsing0F2(L,t,F,ext) or IrrRegAppsing0F2(L,t,H,ext)[1];
# G = regsingtrue0F2(L,t,A[-1],ext) or IrrRegAppsing0F2(L,t,H,ext)[2].
#
# We want to use here Delta1. Since the generalized exponents are not orderned then we will take Delta1 to be
# the first, after the second and after the third exponent differences.
Sirr0F2info1:=proc(L,A,G,x,t,ext)
          local Dx,m,a1,a2,b1,b2,b3,R,e1,e2,e3,acc2,acc1,j,s,B,d,i,e,acc,a,b,l1,k,d1,v,v1,E,neq,dA,nonpole,
                extp1,extp2,extp3,extp4,extp5,extp6,extp0,extp7,H1;
          Dx,x:=op(_Envdiffopdomain);
          # We find the coefficients e1 and e2 of (x-p)^(1/3) in the two first solutions of b*t^3 = t_p
          # Those coefficients are the two first solutions of the equation x^3-1/b = 0. Here b=-1. So
          # the equation is now x^3+1=0.
          R:=sort([solve(x^3+1 = 0, x)]);
          e1:=R[1];
          e2:=R[2];
          #e3:=R[3];
          E:= [];
          nonpole:= 1;
          if G=[] then
             neq:= 0
          else
             neq := add(`if`(s[1]=infinity,1,degree(s[1],x)), s=G[1]);
          fi;
          B:=1;
          dA:=0;
          j:=0;
          for s in A[1] do
             j:=j+1;
             m:= 3*A[4][j];
             if type(m,integer)=false then
                return NULL
             fi;
             if s[1] <> infinity then
                B:= B*(s[1])^m;
             else
                dA := m;
             fi;
             acc:= m; # degree minimal, en valeur absolue, du polynome de
                                      # la partie polaire de f en p.
             acc2:= ceil(A[4][j]); # nombre d'equations venant de p pour les coefficients de A= f*B. Dans
                                   # les normes, la formule est: ceil((1/n_e)*m_j)=ceil(A[4][j]).
                                   # Here n_e = 3.
             acc1:= floor(2*A[4][j]); # nombre de termes dans la partie polaire de f qu'on n'a pas. Dans
                                 # les normes, la formule est: floor(((n_e-1)/n_e)*m_j)= floor((n_e-1)*A[4][j]).
                                 # Here n_e = 3.
             neq:= neq + acc2*op(`if`(s[1]=infinity,1,degree(s[1],x)));
             d:= convert(series(A[3][j][1],t=0,100),polynom);
             if type(d,`+`) then
                d1:= [op(d)]
             else
                d1:=[d]
             fi;
             # We remove the constant term in d
             e:=[];
             for i in d1 do
                if member(t,indets(i)) then
                   e:=[op(e),i]
                fi;
             od;
             l1:=map(s->subs(t=t^3,s),e);
             l1:=map(s->series(subs(t=1/x,s),x=0,100),l1);
             l1:=map(s->eval(subs(csgn=1,simplify(s) assuming(x>0))),l1);
             d1:=[e,l1];
             d:= add(d1[1][i]/(((e2)^(-1)-(e1)^(-1))*ldegree(convert(d1[2][i],polynom), x)), i=1..nops(e));
             d:= eval(subs(csgn=1,simplify(d) assuming(x>0)));
             d:=evala(expand(d^3));
             d:=convert(series(d,t=0,100),polynom);
             if type(d,`+`) then
                d1:= [op(d)]
             else
                d1:=[d]
             fi;
             d:=0;
             for i in d1 do
                l1:= degpol(i,x,t,ext);
                if l1<>false and l1<=m then
                   d:=d+i
                fi;
             od;
             if member(t,indets(d,{name}))=false then
                return NULL
             fi;
             if s[1]=infinity then
                b:=subs(t=1/x,d);
                a1:=convert(series(numer(b),x=0,100),polynom);
                a2:=expand(convert(series(denom(b),x=0,100),polynom));
                b:= factor(a1/a2);
                b:=convert(series(b,x=0,100),polynom);
                b:= eval(subs(csgn=1,simplify(b) assuming(x>0)));
                extp1:= indets(A[3][j][1],{RootOf,name}) minus {t};
                extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name}); # All the names or RootOf
                                        # different from RootOf(s,x) in the expression of A[3][j][1].
                extp2:=indets(ext,{RootOf,name});
                extp3:= extp1 minus extp2; # New names or Rootof which appear.
                extp5:= extp1 union ext;
                E:=[op(E),[s[2],1/x,[convert(evala(b),polynom)],acc,extp3,extp5]];
                nonpole:= nonpole*(1)^acc1; # we take for infinity [1/x,pol]=[1,1].
             else
                b:=`if`(degree(s[1],x) > 1,subs(t=x-s[2],d),subs(t=s[1],d));
                a1:=convert(series(numer(b),x=s[2],100),polynom);
                a2:=expand(convert(series(denom(b),x=s[2],100),polynom));
                b:= factor(a1/a2);
                b:= convert(series(b,x=s[2],100),polynom);
                b:= eval(subs(csgn=1,simplify(b) assuming(x>0)));
                extp0:= indets(ReplirrRoot(s[2],{}),{RootOf,name});
                extp1:= indets(A[3][j][1],{RootOf,name}) minus {t};
                extp1:= indets(map(s-> ReplirrRoot(s,{}),extp1),{RootOf,name});
                extp2:=indets({op(ext),ReplirrRoot(s[2],{})},{RootOf,name});
                extp3:= extp1 minus extp2; # New names or Rootof which appear.
                extp4:= extp1 minus extp0; # All the names or RootOf
                                           # different from RootOf(s,x) in the expression of A[3][j][1].
                extp6:= ext minus extp0;
                extp7:= extp4 union extp6;
                v1:={};
                for v in extp7 do
                   if extp0 intersect indets(v,{RootOf,name}) <> {} then
                      v1:= v1 union {v}
                   fi;
                od:
                extp5:= extp7 minus v1;
                E:=[op(E),[s[2],s[1],[convert(evala(b),polynom)],acc,extp3,extp5]];
                nonpole:= nonpole*(s[1])^acc1;
             fi;
          od;
          dA:= degree(B,x)+dA;
          H1:= table();
          H1[1]:=E;
          H1[2]:=dA;
          H1[3]:=neq;
          H1[4]:=B;
          H1[5]:=nonpole;
          H1:= convert(H1,'list');
          return H1
end:

############################################################################################################
############################################################################################################

####### That is the condition that any candidate [b1,b2] for the lower parameter of 0F2 must satisfy.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext = indets(map(s-> ReplirrRoot(ReplRoot2(s,{})),ext1),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,H,ext);
# f = candidate for the change of variable parameter;
# B1 = Zeroesmultordpol0F2(f,x);
# n = nops(B1);
# RS = newregsing0F2(L,t,x,B1,R[2],n,ext);
# i = index of an element in RS[1][1];
# E1 = RS[1][4][i];
# E2 = RS[1][3][i];
# m = multiplicity order of a i^th zero of f;
# (b1,b2) = candidate for the lower parameters.
Condibiexp0F2:= proc(E1,E2,m,b1,b2,ext)
              local Dx,x;
              Dx,x:=op(_Envdiffopdomain);
              if (type(evala(m*(1-b1)-E2[1]),integer) and type(evala(m*(1-b2)-E2[2]),integer) and
                 type(evala(m*(b1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b1)),integer) and
                 type(evala(E1[2][1]-m*(1-b2)),integer)) then
                 return true
              elif (type(evala(m*(1-b2)-E2[1]),integer) and type(evala(m*(1-b1)-E2[2]),integer) and
                 type(evala(-m*(b1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][1]-m*(1-b2)),integer) and
                 type(evala(E1[2][1]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(1-b1)-E2[1]),integer) and type(evala(m*(b1-b2)-E2[2]),integer) and
                 type(evala(m*(1-b2)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b1)),integer) and
                 type(evala(E1[2][1]-m*(1-b2)),integer)) then
                 return true
              elif (type(evala(m*(b1-b2)-E2[1]),integer) and type(evala(-m*(1-b1)-E2[2]),integer) and
                 type(evala(-m*(1-b2)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-m*(1-b2)),integer) and
                 type(evala(E1[1][2]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(1-b2)-E2[1]),integer) and type(evala(-m*(b1-b2)-E2[2]),integer) and
                 type(evala(m*(1-b1)-E2[3]),integer)) and (type(evala(E1[1][2]-E1[2][2]),integer) and
                 type(evala(E1[2][1]-E1[3][1]),integer) and type(evala(E1[1][2]-m*(1-b2)),integer) and
                 type(evala(E1[2][1]-m*(1-b1)),integer)) then
                 return true
              elif (type(evala(-m*(b1-b2)-E2[1]),integer) and type(evala(-m*(1-b2)-E2[2]),integer) and
                 type(evala(-m*(1-b1)-E2[3]),integer)) and (type(evala(E1[1][1]-E1[3][2]),integer) and
                 type(evala(E1[1][2]-E1[2][2]),integer) and type(evala(E1[1][1]-m*(1-b1)),integer) and
                 type(evala(E1[1][2]-m*(1-b2)),integer)) then
                 return true
              else
                 return false
              fi;
end:

############################################################################################################
############################################################################################################

# For a rational function f, we return the roots of its numerator associated with their minimal polynoms.
Zeroesmultordpol0F2:= proc(f,x)
                  local Dx,B1,B2,B3,B4,m,g,A,A1,j,s;
                  Dx, x := op(_Envdiffopdomain);
                  g:= numer(f);
                  A:=[coeffs(collect(g,[x],'distributed'),[x])];
                  A1:=map(s-> `if`(type(s,rational)=false,NULL,s),A);
                  if nops(A)<>nops(A1) then
                     return NULL
                  fi;
                  B1:= factdecomp(g);
                  B2:=map(s->`if`(member(x,indets(s,{name})),s,NULL),B1);
                  B3:= map(s->[[s[1],RootOf(s[1],x)],s[2]],B2);
                  B4:= map(s-> [[s[1][1],ConvirrRoot(s[1][2])[1]],s[2]],B3);
                  m:= degree(denom(f),x)-degree(numer(f),x);
                  if m > 0 then
                     return [[[infinity,infinity],m],op(B4)]
                  else
                     return B4
                  fi;
end:

############################################################################################################
############################################################################################################

# We find the set of all the zeroes of f, their generalized exponents, their exponent differences, their pairs of
# generalized exponents, and their multiplicity order as zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# Sreg = IrrRegAppsing0F2(L,t,H,ext)[2];
# B = Zeroesmultordpol0F2(f,x);
# n = nops(B).
newregsing0F2:= proc(L,t,x,B,Sreg,n,ext)
                  local Dx,i,j,M1,M,n1,G,A1,A2,A3,A31,A32,A4,A5,G1,R;
                  Dx, x := op(_Envdiffopdomain);
                  G:=Sreg;
                  if G<>[] then
                     n1:= nops(G[1]);
                     if n1=n then
                        G1:={op(B)};
                        M:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M:= convert(M,'list');
                        return [G,M]
                     else ####### n>n1
                        A1:= map(s-> `if`(member(s[1],{op(G[1])}),NULL,s),B);
                        A2:= [map(s->s[1],B),map(s->s[2],B),map(s->s[1],A1)];
                        A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[3]);
                        A31:= map(s-> [s[1],s[2][1],Condi0F2reg(s[2],t,ext)],A3);
                        A32:= {op(map(s-> nops(s),A31))};
                        if nops(A32)<>1 then
                           return NULL
                        fi;
                        A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                        A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                        R:= subs({G[1]=[op(G[1]),op(A5[1])],G[2]=[op(G[2]),op(A5[2])],
                                  G[3]=[op(G[3]),op(A5[3])],G[4]=[op(G[4]),op(A5[4])]},G);
                        G1:= {op(B)} minus {op(A1)};
                        M1:= table();
                        for i from 1 to n1 do
                           for j in G1 do
                              if G[1][i]=j[1] then
                                 M1[i]:= j[2];
                                 G1:=G1 minus {j};
                                 break
                              fi;
                           od;
                        od;
                        M1:= convert(M1,'list');
                        M:= [op(M1),op(A2[2])];
                        return [R,M]
                     fi;
                  else ####### G=NULL
                     A2:= [map(s->s[1],B),map(s->s[2],B)];
                     A3:= map(s-> [s,genexpdiff(L,s,t,ext)],A2[1]);
                     A31:= map(s-> [s[1],s[2][1],Condi0F2reg(s[2],t,ext)],A3);
                     A32:= {op(map(s-> nops(s),A31))};
                     if nops(A32)<>1 then
                        return NULL
                     fi;
                     A4:= map(s-> [s[1],s[2],s[3][1],s[3][2]],A31);
                     A5:= [map(s-> s[1],A4),map(s-> s[2],A4),map(s-> s[3],A4),map(s-> s[4],A4)];
                     return [A5,A2[2]]
                  fi;
end:

############################################################################################################
############################################################################################################
#################################################   EASY CASE   ############################################
############################################################################################################
############################################################################################################

# We compute the candiates for a1,a2,b1,b2 and f: [[[b1,b2],...],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext).
easy0F2f:= proc(L,R,info,x,t,ext)
                  local Dx,a,f,fn,p,i,j,linearr,temp,d,d1,nonpole,und,B1,n,SR,g,
                        g1,A1,A2,G,H1,H,c1,c2,c3,C;
                  Dx,x:=op(_Envdiffopdomain);
                  ####### info[4] gives us the denominator of f.
                  ####### degree(A,x)=info[2].
                  H:= table();
                  f := add(a[i]*x^i, i = 0 .. info[2])/info[4];
                  und := indets(f) minus {x};
                  C:=map(s->s[1],R[2][1]);       ####### In easy case, Sreg<>[]
                  if has({op(C)}, infinity) then
                     f:=subs(a[info[2]]=0,f);
                     und:=subs(a[info[2]]=NULL,und);
                  fi;
                  fn := numer(f);
                  for p in R[2][1] do
                     if p[1]=infinity then
                        p[1]:= 1/x;
                     fi;
                     fn := fn/(p[1]);
                  od;
                  linearr:=convert(collect(eval(subs(csgn=1,simplify(evala(Rem(numer(fn),denom(fn),x))) assuming(x>0))),x),polynom);
                  linearr := coeffs(linearr, x);
                  temp := SolveTools:-Linear({linearr},und);
                  f := subs(op(temp), f);
                  H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
                  j := 0; # use j indicate that if infinity is an irregular singularity
                  for p in info[1] do
                     if p[1] = infinity then
                        j := ceil(p[4]/3);
                        break;
                     fi;
                  od;
                  G := Hyperpossibility(H1);
                  for d in G do
                     c1:= convert(series(numer(f-d),x=0,100),polynom);
                     c2:= convert(series(denom(f-d),x=0,100),polynom);
                     c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                     nonpole := Normalizer(c3*info[5]);
                     A1:=convert(series(numer(nonpole),x=0,100),polynom);
                     A2:=convert(series(denom(nonpole),x=0,100),polynom);
                     temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                     temp := collect(temp, x, Normalizer);
                     linearr := coeffs(temp, x);
                     if j<>0 then # the equations at infinity
                        temp := evala(Quo(A1, A2,x));
                        temp:=convert(series(numer(temp),x=0,100),polynom);
                        temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                        temp := collect(temp, x, Normalizer);
                        d1:=degree(temp, x);
                        temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                        linearr := linearr, temp;
                     fi;
                     temp := SolveTools:-Linear({linearr},und);
                     if temp=NULL then
                        next
                     fi;
                     temp := op(temp);
                     g := subs(temp, f);
                     g := factor(op(simplef({g})));
                     B1:= Zeroesmultordpol0F2(g,x);
                     if B1=NULL then
                        next
                     fi;
                     n:= nops(B1);
                     SR:= newregsing0F2(L,t,x,B1,R[2],n,ext);
                     H[d]:= [g,SR[2],SR[1]];
                  od;
                  H:= convert(H,'set');
                  if H = {} then
                     return NULL
                  else
                     return H
                 fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for a1,a2,b1,b2 and f: [{{a1,...}},[b1,b2],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext).
easy0F2:= proc(L,R,info,x,t,ext)
          local B,ab,f,C,H,E,n1,n,ext1,ext2,ext3,ext4,i,n2,s;
          B:= easy0F2f(L,R,info,x,t,ext);
          if B = NULL then
             return NULL
          fi;
          ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
          ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
          n1:=nops(R[2][1]);
          if n1 > 1 then
             for i from 2 to n1 do
                ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                ext1:= ext1 intersect ext2;
             od;
          fi;
          H:= table();
          for f in B do
             n:= nops(f[3][1]);
             if R[3][3] <> [] then
                C:= findbi0F2ln(L,f,x,t,ext,ext1,n);
             else
                C:= findbi0F2(L,f,x,t,ext,ext1,n);
             fi;
             if C<>NULL then
                H[f]:= [C[1],f[1]]
             fi;
          od;
          H:= convert(H,'list');
          if H = [] then
             return NULL
          else
             return H
          fi;
end:

############################################################################################################
############################################################################################################
############################################   LOGARITHMIC CASE   ##########################################
############################################################################################################
############################################################################################################

# g = f^2 = (C.A1.A2^b)/B. We want to find candidates for A1. Sreg=regsingtrue(L,t,ext)[1] is the set of the
# zeros of A1.
# input: H and d, where H=subs(Sreg[4]=NULL,Sreg) and d is the degree of A1. We do a combinatorial distribution
# of d to the elements of H[1];
# return: candidates for A1 = \prod_{p\in Sreg[1]}(p^{k_p}) where
# p = an element of Sreg[1] (minimal polynomial);
# d = \sum_{p\in Sreg[1]}(k_p*deg(p)) = degree of A1.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# H = IrrRegAppsing0F2(L,t,H,ext)[2].
searchKnLog :=proc(H,d)
              local res,temp,s,deg,t,k;
              res:=NULL;
              if nops(H[1])=1 then
                 deg:=`if`(H[1][1][1]=infinity,1,degree(H[1][1][1],x));
                 if d mod deg = 0 then
                    if H[1][1][1]= infinity then
                       res:=res,1;
                    else
                       res:= res, H[1][1][1]^(d/degree(H[1][1][1],x));
                    fi
                 fi;
              else
                 deg := `if`(H[1][1][1]=infinity,1,degree(H[1][1][1],x));
                 for k from 1 to iquo(d,deg) do
                    temp := searchKnLog([seq([seq(H[i][j],j=2..nops(H[i]))],i=1..nops(H))],d-k*deg);
                    if temp<>NULL then
                       if H[1][1][1] <> infinity then
                          for t in temp do
                             res:= res,H[1][1][1]^k*t;
                          od;
                       else
                          for t in temp do
                             res:=res,t;
                          od;
                       fi;
                    fi;
                 od;
              fi;
              return [res];
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# H = Singular(L,extppp);
# F = NotAppSing(L,H,ext);
# Sirr = IrrRegAppsing0F2(L,t,H,ext)[1];
# Sreg = IrrRegAppsing0F2(L,t,H,ext)[2];
# info = Sirr0F2info1(L,Sirr,l,x,ext).
find0F2Logf := proc(L,Sreg,info,x,t,ext) ####### Rapide
                 local Dx,mult1,M,p,res,knl,kn,g,i,j,linearr,temp,d,d1,nonpole,B,B1,n,SR,dA,k,e,G,A1,A2,h,h1,H,c1,c2,c3;
                 Dx,x:=op(_Envdiffopdomain);
                 res:= {};
                 dA:= info[2];
                 B:= info[4];
                 knl := searchKnLog(Sreg,dA);
                 H:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
                 j := 0; # use j indicate that if infinity is an irregular singularity
                 k:=0;
                 for p in info[1] do
                    k:=k+1;
                    if p[1] = infinity then
                       j := ceil(info[1][k][4]/3);
                       break;
                    fi;
                 od;
                 G := Hyperpossibility(H);
                 for kn in knl do
                    for d in G do
                       c1:= convert(series(numer(c*kn/B-d),x=0,100),polynom);
                       c2:= convert(series(denom(c*kn/B-d),x=0,100),polynom);
                       c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                       nonpole := Normalizer(c3*info[5]);
                       A1:=convert(series(numer(nonpole),x=0,100),polynom);
                       A2:=convert(series(denom(nonpole),x=0,100),polynom);
                       temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                       temp := collect(temp, x, Normalizer);
                       linearr := coeffs(temp, x);
                       if j<>0 then # the equations at infinity
                          temp := evala(Quo(A1, A2,x));
                          temp:=convert(series(numer(temp),x=0,100),polynom);
                          temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                          temp := collect(temp, x, Normalizer);
                          d1:=degree(temp, x);
                          temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                          linearr := linearr, temp;
                       fi;
                       temp := SolveTools:-Linear({linearr},{c});
                       if temp=NULL then
                          next
                       fi;
                       temp := op(temp);
                       h:= subs(temp, c*kn/B);
                       B1:= Zeroesmultordpol0F2(h,x);
                       if B1=NULL then
                          next
                       fi;
                       n:= nops(B1);
                       SR:= newregsing0F2(L,t,x,B1,Sreg,n,ext);
                       res := res union {[h,SR[2],SR[1]]};
                    od;
                 od;
                 if res = {} then
                    return NULL
                 else
                    return res
                 fi;
end:

############################################################################################################
############################################################################################################

####### b11, b12 and b13 are sets of candidates for 1-b1, 1-b2 and b1-b2 respectively. We compute here the candidates
####### for [b1,b2] with the rational part of b1 and b2 in ]0,1], related to a zero of the change of variable parameter.
#######
####### m = multiplicity order of the element as a zero of f;
####### e = [Delta1,Delta2,Delta3];
####### a = [[e,e],[e,e],[e,e]].
casesbiLog0F2:= proc(b11,b21,b31,e,a,m,ext,ext1)
          local k,s,b1,b2,b3,B,B1,B2,B3,d1,d2,i,j;
          b1:=map(s->1-s,b11);
          b2:=map(s->1-s,b21);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          b2:= [seq(`if`(MatchInt2([op(b2[1..k-1])],b2[k]), NULL, b2[k]),k=1..nops(b2))];
          b3:= [seq(`if`(MatchInt2([op(b31[1..k-1])],b31[k]), NULL, b31[k]),k=1..nops(b31))];
          ####### We compute the candidates for b1, b2 and b1-b2 such that their rational parts are in [0,1]
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          B2:={op(map(s-> Ratiopartmodint(s,ext1),b2))};
          B3:={op(map(s-> Ratiopartmodint(s,ext1),b3))};
          B3:=map(s-> [s[1],s[2],s[3]],B3);
          B1:=map(s->`if`(s[1]=0,[1,1,s[3],s[4]],s),B1);
          B2:=map(s->`if`(s[1]=0,[1,1,s[3],s[4]],s),B2);
          B:= table();
          for i in B1 do
             for j in B2 do
                ####### the condition for the logarithmic case
                if type(i[1],integer) or type(j[1],integer) or type(i[1]-j[1],integer) or (type(i[1]*2,integer)
                   and type(j[1]*2,integer) and i[1]*j[1]< 0) then
                   d1:= Ratiopartmodint(i[1]-j[1],ext1);
                   d2:= Ratiopartmodint(j[1]-i[1],ext1);
                   d1:= [d1[1],d1[2],d1[3]];
                   d2:= [d2[1],d2[2],d2[3]];
                   if {d1,d2} intersect B3 <> {} then
                      ####### the condition that a candidate for (b1,b2) must satisfy
                      if Condibiexp0F2(a,e,m,i[1],j[1],ext)=true or Condibiexp0F2(a,e,m,j[1],i[1],ext)=true then
                         B[i,j]:= {i[1],j[1]}
                      fi;
                   fi;
                fi;
             od;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameters [b1,b2] related to a zero of the change of variable
# parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext);
# B = find0F2Logf(L,info,R[2],ext,x,t);
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiffLog0F2:= proc(j,G,m,ext,ext1)
          local d,i,ext2,b11,b21,b31,B,H1,H2,H3,C1,C,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          B:= table();
          C:=[[G[3][j][1],G[3][j][2],G[3][j][3]],[G[3][j][1],G[3][j][3],G[3][j][2]],[G[3][j][2],G[3][j][3],G[3][j][1]]];
          for i in C do
             ####### b11 is the list of candidates for +-(1-b1) modulo an integer
             b11:=[seq(i[1]+s,s=-m+1..m-1),seq(-i[1]+s,s=-m+1..m-1)]/m;
             ####### b21 is the list of candidates for +-(1-b2) modulo an integer
             b21:= [seq(i[2]+s,s=-m+1..m-1),seq(-i[2]+s,s=-m+1..m-1)]/m;
             ####### b31 is the list of candidates for +-(b1-b2) modulo an integer
             b31:= [seq(i[3]+s,s=-m+1..m-1),seq(-i[3]+s,s=-m+1..m-1)]/m;
             C1:=casesbiLog0F2(b11,b21,b31,G[3][j],G[4][j],m,ext,ext1);
             if C1 <> NULL then
                B[i]:=op(C1)
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters b1 and b2 related to a candidate of f: [[[b1,b2],[b1,b2]...],f,M],
# where M is the list of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext);
# B1 = find0F2Logf(L,info,R[2],ext,x,t);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter;
# f[2] list of multiplicity orders of zeroes of f[1];
# f[3][1] list of zeroes of the change of variable parameter.
findbi0F2ln:=proc(L,f,x,t,ext,ext1,n1)
                  local Dx,H,B,j,s,i;
                  Dx,x:=op(_Envdiffopdomain);
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiffLog0F2(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> `if`(nops(s)=1,[s[1],s[1]],[s[1],s[2]]),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters (b1,b2), also and f: [[[b1,b2],[b1,b2],...],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext).
find0F2ln:=proc(L,R,info,x,t,ext)
                  local Dx,B,ab,f,C,H,E,ext1,n1,n,ext2,i,s;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= find0F2Logf(L,R[2],info,x,t,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                  ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                  n1:=nops(R[2][1]);
                  if n1 > 1 then
                     for i from 2 to n1 do
                        ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                        ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                        ext1:= ext1 intersect ext2;
                     od;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi0F2ln(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        H[f]:= [C[1],f[1]]
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##############################################   IRRATIONAL CASE   ##########################################
############################################################################################################
############################################################################################################

leadcoeffRootnameinter:=proc(A,x,ext1)
                    local Dx,l,l1,ira,ira1,ira2,sum1,sum2;
                    Dx,x:=op(_Envdiffopdomain);
                    l:=A;
                    l1:=ext1;
                    while l1<>{} and (l1 intersect ext1 <> {}) do
                       ira1:=ChooseRootOflistinter(l,l1)[1];
                       if ira1<>{} then
                          ira:=ira1[1];
                       else
                          ira2:= l1 intersect ext1:
                          ira:=ira2[1];
                       fi;
                       sum1:=convert(collect(l,ira,evala),polynom);
                       sum2:= `if`(ldegree(sum1,ira)<0,convert(subs(ira=1/ira,sum1),polynom),sum1);
                       l:= lcoeff(sum2,ira);
                       l1:= indets(l,{RootOf,name});
                    od;
                    return l
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,ext);
# R = IrrRegAppsing0F2(L,t,E,ext);
# info = Sirr0F2info1(L,R[1],R[2],l,x,ext).
find0F2Irrf := proc(L,R,info,x,t,ext)
             local M,res,ira,dA,B,B1,n1,SR,ext1,ext2,A,m,i,r,sum,a,j,h,H1,s,p,G,d,A1,A2,d1,c1,c2,c3,c,linearr,
                   temp,nonpole,n,b1,v,V,g,ira1,ira2,E,H,C,l1;
             res := {};
             M:=[];
             dA:= info[2];
             B:= info[4];
             n:=nops(R[2][1]);
             ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
             ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
             if n > 1 then
                for i from 2 to n do
                   ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                   ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                   ext1:= ext1 intersect ext2;
                od;
             fi;
             ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
             ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
             ####### irrational case.
             r[1]:=1;
             for i from 1 to n do
                ira:=0;
                for j from 1 to 3 do
                   C:=ReplirrRoot(R[2][3][i][j],{});
                   l1:= leadcoeffRootnameinter(C,x,ext1);
                   if type(l1,rational)=false or l1=0 then
                      next
                   else
                      H[i]:=C;
                      ira:=1;
                      break
                   fi;
                od;
                if ira=0 then
                   return NULL
                fi;
                a[i]:= l1*(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));
                if i>1 then
                   r[i] := Normalizer(abs(a[i]/a[1])); #   r[i]=(m_i*degree(p_i,x))/(m_1*degree(p_1,x))
                   if not type(r[i],rational) then
                      return NULL
                   fi;
                fi;
             od;
             sum:=add(r[i],i=1..n);
             b1:=dA/sum; # b1 = m_1*degree(p_1,x)
             if type(b1, integer) then
                A:=c;
                for i from 1 to n do
                   m[i]:= (r[i]*b1)/(`if`(R[2][1][i][1]=infinity,1,degree(R[2][1][i][1],x)));   # m[i]=m_i
                   A:=A*(`if`(R[2][1][i][1]=infinity,1,(R[2][1][i][1])^(m[i])));
                   M:=[op(M),m[i]];
                od;
             else
                return NULL;
             fi;
             H1:= map(s-> `if`(s[1]=infinity,op(s[3]),evala(Trace(s[3][1], s[6], s[6]), x)),info[1]);
             j := 0; # use j indicate that if infinity is irregular singularity
             for p in info[1] do
                if p[1] = infinity then
                   j := ceil(p[4]/3);
                   break;
                fi;
             od;
             G := Hyperpossibility(H1);
             for d in G do
                c1:= convert(series(numer(A/B-d),x=0,100),polynom);
                c2:= convert(series(denom(A/B-d),x=0,100),polynom);
                c3:= factor(eval(subs(csgn=1,simplify(c1/c2) assuming(x>0))));
                nonpole := Normalizer(c3*info[5]);
                A1:=convert(series(numer(nonpole),x=0,100),polynom);
                A2:=convert(series(denom(nonpole),x=0,100),polynom);
                temp := convert(eval(subs(csgn=1,simplify(evala(Rem(A1, A2, x))) assuming(x>0))),polynom);
                temp := collect(temp, x, Normalizer);
                linearr := coeffs(temp, x);
                if j<>0 then # the equations at infinity
                   temp := evala(Quo(A1, A2,x));
                   temp:=convert(series(numer(temp),x=0,100),polynom);
                   temp := convert(eval(subs(csgn=1,simplify(temp) assuming(x>0))),polynom);
                   temp := collect(temp, x, Normalizer);
                   d1:=degree(temp, x);
                   temp := seq(coeff(temp, x, i), i = d1-j+1 .. d1);
                   linearr := linearr, temp;
                fi;
                temp := SolveTools:-Linear({linearr},{c});
                if temp=NULL then
                   next
                fi;
                temp := op(temp);
                h:= subs(temp, A/B);
                B1:= Zeroesmultordpol0F2(h,x);
                if B1=NULL then
                   next
                fi;
                n1:= nops(B1);
                SR:= newregsing0F2(L,t,x,B1,R[2],n1,ext);
                res := res union {[h,SR[2],SR[1]]};
             od;
             if res = {} then
                return NULL
             else
                return [res,n,ext1]
             fi;
end:

############################################################################################################
############################################################################################################

####### b11, b12 and b13 are sets of candidates for 1-b1, 1-b2 and b1-b2 respectively. We compute here the candidates
####### for [b1,b2] with the rational part of b1 and b2 in ]0,1], related to a zero of the change of variable parameter.
#######
####### m = multiplicity order of the element as a zero of f;
####### e = [Delta1,Delta2,Delta3];
####### a = [[e,e],[e,e],[e,e]].
casesbi0F2:= proc(b11,b21,b31,e,a,m,ext,ext1)
          local k,s,b1,b2,b3,B,B1,B2,B3,d1,d2,i,j;
          b1:=map(s->1-s,b11);
          b2:=map(s->1-s,b21);
          b1:= [seq(`if`(MatchInt2([op(b1[1..k-1])],b1[k]), NULL, b1[k]),k=1..nops(b1))];
          b2:= [seq(`if`(MatchInt2([op(b2[1..k-1])],b2[k]), NULL, b2[k]),k=1..nops(b2))];
          b3:= [seq(`if`(MatchInt2([op(b31[1..k-1])],b31[k]), NULL, b31[k]),k=1..nops(b31))];
          ####### We compute the candidates for b1, b2 and b1-b2 such that their rational parts are in [0,1]
          B1:={op(map(s-> Ratiopartmodint(s,ext1),b1))};
          B2:={op(map(s-> Ratiopartmodint(s,ext1),b2))};
          B3:={op(map(s-> Ratiopartmodint(s,ext1),b3))};
          B3:=map(s-> [s[1],s[2],s[3]],B3);
          B1:=map(s->`if`(s[1]=0,NULL,s),B1);
          B2:=map(s->`if`(s[1]=0,NULL,s),B2);
          if B1={} or B2={} then
             return NULL
          fi;
          B:= table();
          for i in B1 do
             for j in B2 do
                ####### the condition for the non-logarithmic case
                if type(i[1]-j[1],integer)=false and (type(i[1]*2,integer)=false or type(j[1]*2,integer)=false or
                   i[1]*j[1]>= 0) then
                   d1:= Ratiopartmodint(i[1]-j[1],ext1);
                   d2:= Ratiopartmodint(j[1]-i[1],ext1);
                   d1:= [d1[1],d1[2],d1[3]];
                   d2:= [d2[1],d2[2],d2[3]];
                   if {d1,d2} intersect B3 <> {} then
                      ####### the condition that a candidate for (b1,b2) must satisfy
                      if Condibiexp0F2(a,e,m,i[1],j[1],ext)=true or Condibiexp0F2(a,e,m,j[1],i[1],ext)=true then
                         B[i,j]:= {i[1],j[1]}
                      fi;
                   fi;
                fi;
             od;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

######### We compute the candiates for the lower parameters [b1,b2] related to a zero of the change of variable
# parameter.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext);
# B = find0F2Irrf(L,info,R[3][1],ext,x,t);
# f = element of B;
# G = f[3];
# j = index of an element in f[3][1];
# m = f[2][j] = list of the multiplicity order of the j^th element of f[3][1] as zero of the change of variable
#               parameter.
binotintexpdiff0F2:= proc(j,G,m,ext,ext1)
          local d,i,ext2,b11,b21,b31,B,H1,H2,H3,C1,C,s;
          ####### Let k be the field of constants that we want to work on such that L is in k(x)[\partial].
          ####### If ext1 is a subset of k then we are in the basefield case. if not we are in the
          ####### irrational case.
          B:= table();
          C:=[[G[3][j][1],G[3][j][2],G[3][j][3]],[G[3][j][1],G[3][j][3],G[3][j][2]],[G[3][j][2],G[3][j][3],G[3][j][1]]];
          for i in C do
             ####### b11 is the list of candidates for +-(1-b1) modulo an integer
             b11:=[seq(i[1]+s,s=-m+1..m-1),seq(-i[1]+s,s=-m+1..m-1)]/m;
             ####### b21 is the list of candidates for +-(1-b2) modulo an integer
             b21:= [seq(i[2]+s,s=-m+1..m-1),seq(-i[2]+s,s=-m+1..m-1)]/m;
             ####### b31 is the list of candidates for +-(b1-b2) modulo an integer
             b31:= [seq(i[3]+s,s=-m+1..m-1),seq(-i[3]+s,s=-m+1..m-1)]/m;
             C1:=casesbi0F2(b11,b21,b31,G[3][j],G[4][j],m,ext,ext1);
             if C1 <> NULL then
                B[i]:=op(C1)
             fi;
          od;
          B:= convert(B,'set');
          if B = {} then
             return NULL
          else
             return B
          fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters b1 and b2 related to a candidate of f: [[[b1,b2],[b1,b2]...],f,M],
# where M is the list of the multiplicity order of the zeroes of f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext);
# B1 = find0F2Irrf(L,info,R[3][1],ext,x,t);
# f = element of B1;
# n1 = nops(f[3][1]);
# f[1] a candidate for the change of variables parameter;
# f[2] list of multiplicity orders of zeroes of f[1];
# f[3][1] list of zeroes of the change of variable parameter.
findbi0F2:=proc(L,f,x,t,ext,ext1,n1)
                  local Dx,H,B,j,s,i;
                  Dx,x:=op(_Envdiffopdomain);
                  H:= {};
                  for j from 1 to n1 do
                     B[j]:=binotintexpdiff0F2(j,f[3],f[2][j],ext,ext1);
                     if B[j]=NULL then
                        return NULL
                     fi;
                     if j = 1 then
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j])
                     else
                        H:= map(s-> map(i-> ReplRoot2(i,{}),s),B[j]) intersect H
                     fi;
                     if H = {} then
                        return NULL
                     fi;
                  od;
                  H:= map(s-> `if`(nops(s)=1,[s[1],s[1]],[s[1],s[2]]),H);
                  return [[op(H)],f[1],f[2]]
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters (b1,b2), also and f: [[[b1,b2],[b1,b2],...],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = NotAppSing(L,G,ext);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext).
find0F2Irr:=proc(L,R,info,x,t,ext)
                  local Dx,B,n,f,C,H;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= find0F2Irrf(L,R,info,x,t,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  H:= table();
                  for f in B[1] do
                     n:= nops(f[3][1]);
                     C:= findbi0F2(L,f,x,t,ext,B[3],n);
                     if C<>NULL then
                        H[f]:= [C[1],f[1]]
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################
##########################################   RATIONAL CASE   ###############################################
############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute all the candidates for A1 associated with
# the degree of A2. The output is a list with element in the form [s,k] where s=A1 and k=degree(A2).
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# G = R[3][1];
# dr = degree(A1);
# k = degree(A2);
# dA = degree(cA1.A2^d) = dr+k*d = degree of the numerator of f^2;
# s = A1.
search0F2A1k:=proc(G,dA,d,x)
           local Dx,s,dr,k,res,temp;
           Dx,x:=op(_Envdiffopdomain);
           res :=NULL;
           if G=[] then
              if irem(dA,d)<>0 then
                 return {};
              else
                 return {[1,dA/d]};
              fi;
           fi;
           for k from 0 to iquo(dA,d) do
              dr :=dA-k*d;
              temp :=search0F2A1(G,dr,d,x);
              for s in temp do
                 res := res, [s,k]
              od;
           od;
           {res};
end:

############################################################################################################
############################################################################################################

# Let f^2=(cA1.A2^d)/B where B is the denominator of f^2. We compute here all the candidates for A1 having the same
# degree.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# G = R[3][1];
# de = degree(A1);
# k = degree(A2);
# dA = degree(cA1.A2^d) = de+k*d = degree of the numerator of f^2.
search0F2A1 :=proc(G,de,d,x)
           local Dx,d1,res,i,temp,k,t,E,n,j,s,m;
           Dx,x:=op(_Envdiffopdomain);
           res :=NULL;
           if nops(G[1])=0 then
              if de=0 then
                 return [1];
              else
                 return NULL;
              fi;
           fi;
           s :=G[1][1][1];
           d1:=denom(G[3][1][1][1]);
           i :=abs(d/d1);
           if type(i,integer)=false then
              return NULL
           fi;
           if (de-i*(`if`(s=infinity,1,degree(s,x))))<0 then
              return NULL;
           fi;
           E:= [seq([seq(G[n][j],j=2..nops(G[n]))],n=1..nops(G))];
           m:=add(`if`(j[1]=infinity,1,degree(j[1],x)),j=E[1]);
           for k from 1 to iquo(de-m,i*(`if`(s=infinity,1,degree(s,x)))) do
              if i*k>=d then
                 break
              fi;
              temp := search0F2A1(E,de-k*i*(`if`(s=infinity,1,degree(s,x))),d,x);
              if temp <> NULL then
                 for t in temp do
                    res := res,(`if`(s=infinity,1,s))^(i*k)*t;
                 od;
              fi;
           od;
           [res];
end:

############################################################################################################
############################################################################################################

poly_to_powerseries := proc(P, x, p, T, acc)
                       local d, l, i, S,a;
                       a:=convert(P,polynom);
                       if a=0 then
                          # error "only treat non-zero powseries"
                          return NULL
                       fi;
                       if p = infinity then
                          d, l:= degree(a,x), lcoeff(a,x);
                          [l, -d,add(evala(coeff(a,x,d-i)/l)*T^i, i=0..acc-1), acc]
                       else
                          S := collect(eval(a, x = T + p), T, evala);
                          d, l := ldegree(S,T), tcoeff(S,T);
                          # Note: the evala in the line S := .. is needed to ensure that d is correct.
                          [l, d,add(evala(coeff(S,T,d+i)/l)*T^i, i=0..acc-1), acc]
                       fi;
end:

############################################################################################################
############################################################################################################

pow_reciprocal := proc(L1,tp)
                  [Normalizer(1/L1[1]),-L1[2],convert(series(1/L1[3],tp=0,L1[4]),polynom),L1[4]]
end:

############################################################################################################
############################################################################################################

pow_time := proc(L1,L2,tp)
            local acc,f,ind;
            acc := min(L1[4],L2[4]);
            f := expand(L1[3]*L2[3]);
            ind := [op(indets(f))];
            if nops(ind) = 2 then
               f := subs(ind[1] = tp, f);
               f := subs(ind[2] = tp, f)
            fi;
            [evala(L1[1]*L2[1]),L1[2]+L2[2],convert(series(f,tp=0,acc),polynom),acc]
end:

############################################################################################################
############################################################################################################

dthroot_powseries := proc(L, tp, d, ext)
                     local X,v,res,i,s,l;
                     l:=L[1];
                     try
                        v:=evala(Factors(X^d -l,ext));
                     catch :
                        v:=evala(Factors(numer(normal(X^d - l)),{op(ext)}))
                     end try;
                     v:=select(has,v[2],X);
                     res := NULL;
                     for i in v do
                        if degree(i[1],X) = 1 then
                           res := res, -coeff(i[1],X,0)/coeff(i[1],X,1)
                        fi;
                     od;
                     v := series(L[3]^(1/d), tp=0, L[4]);
                     v := collect(convert(v,polynom),tp,evala);
                     res :={seq([i, L[2]/d, v, L[4]],  i = [res])};
                     res;
end:

############################################################################################################
############################################################################################################

simplef := proc(flist)
           local f,res;
           res := {};
           for f in flist do
              f := evala(Normal(f, expanded));
              f := evala(Sqrfree(f,x));
              f := f[1]*mul(i[1]^i[2],i=f[2]);
              res := res union {f};
           od;
           res;
end:

############################################################################################################
############################################################################################################

# Let f=(cA1.A2^d)/B where B is the denominator of f. We compute here a candidates for f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,E,ext);
# l = 1;
# dA = degree(A);
# d = denom(b1) or denom(b2) or denom(b1-b2);
# knl = search0F2A1k(R[3][1],dA,d,x);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext);
# B = info[4];
# kn element of knl.
helpcompute0F2f := proc(info,B,d,kn,x,t,T,ext)
                      local Dx,a,l,acc,ep,B1, Bs, temp, ser,lp, uk, und,i, j, lr,f, extp,ukn,uki,p;
                      Dx,x:=op(_Envdiffopdomain);
                      for i from 1 to nops(info[1]) do
                         acc := info[1][i][4];
                         p:= info[1][i][1];
                         ep := poly_to_powerseries(kn[1],x,p,T,acc);
                         ep := pow_reciprocal(ep,T);
                         Bs := poly_to_powerseries(B,x,p,T,acc);
                         temp := pow_time(ep, Bs, T);
                         ep := poly_to_powerseries(info[1][i][3][1],x,p,T,acc);
                         ser[i] := pow_time(temp, ep,T);
                         lp := ser[i][1];
                         und := indets(lp,RootOf) minus ext;
                         if und={} then
                            if not assigned(l) then
                               l:=lp;
                            else
                               if length(lp)<length(l) then
                                  l := lp;
                               fi;
                            fi;
                         fi;
                      od;
                      if not assigned(l) then
                         return {};
                      fi;
                      uk := add(a[i]*x^i,i=0..kn[2]);
                      und := indets(uk) minus {x};
                      for i from 1 to nops(info[1]) do
                         acc := info[1][i][4];
                         p:= info[1][i][1];
                         temp := ser[i];
                         if temp[2] <> 0 and p <> infinity then
                            error "we made an error"
                         fi;
                         temp[1] := temp[1]/l;
                         extp := indets(p,RootOf) union ext;
                         ep := dthroot_powseries(temp,T,d,extp);
                         ukn := NULL;
                         for uki in {uk} do
                            temp := poly_to_powerseries(uki,x,p,T,acc);
                            if temp = NULL then
                               next
                            fi;
                            for j in ep do
                               lr := temp[3]-j[3];
                               lr := coeffs(lr,T);
                               lr := {lr,temp[1]-j[1]};
                               lr := map(numer,lr);
                               if type(p,RootOf) and not member(p,ext) then
                                  lr := map(evala@Expand, lr);
                                  lr := map(coeffs, lr,p)
                               fi;
                               lr := map(s-> `if`(indets(s) intersect und={},NULL,s),lr);
                               lr := SolveTools:-Linear(lr,und);
                               if lr <> NULL then
                                  f:= subs(op(lr),uki);
                                  ukn := ukn,f;
                               fi;
                            od;
                         od;
                         uk := ukn;
                         if uk = NULL then
                            break
                         fi;
                      od;
                      if uk <> NULL then
                         return {seq(evala(l*i^d*kn[1]/B),i=[uk])}
                      else
                         return NULL
                      fi;
end:

############################################################################################################
############################################################################################################

# Let f=(cA1.A2^d)/B where B is the denominator of f. We compute here all the candidates for f.
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,E,ext);
# dA = degree(A);
# d = denom(b1) or denom(b2) or denom(b1-b2);
# knl = search1F2A1k(R[3][1],dA,d,x);
# info = Sirr1F2info1(L,R[1],R[2],x,t,ext).
help0F2f := proc(L,info,d,knl,x,t,T,ext)
                local Dx,res,kn,uk,B;
                Dx,x:=op(_Envdiffopdomain);
                res := {};
                if knl <> {} then
                   B:= info[4];
                   for kn in knl do
                      uk:= helpcompute0F2f(info,B,d,kn,x,t,T,ext);
                      if uk <> NULL then
                         res:= res union uk;
                      fi;
                   od;
                fi;
                res;
end:

############################################################################################################
############################################################################################################

Goodexpbi1:=proc(A,x,t,ext)
          local Dx,n,B,i;
          Dx,x:=op(_Envdiffopdomain);
          B:=table();
          n:=nops(A);
          for i from 1 to n do
             if nops(A[i][1])=3 then
                if abs(denom(A[i][1][1])*denom(A[i][1][2]))=abs(denom(A[i][1][3])) or
                   abs(denom(A[i][1][1]))=abs(denom(A[i][1][2])) then
                   B[i]:= A[i]
                elif abs(denom(A[i][1][1])*denom(A[i][1][3]))=abs(denom(A[i][1][2])) or
                     abs(denom(A[i][1][1]))=abs(denom(A[i][1][3])) then
                   B[i]:= [[A[i][1][1],A[i][1][3],A[i][1][2]],[]]
                else # abs(denom(A[i][1][1])*denom(A[i][1][2]))=abs(denom(A[i][1][3])) or
                     # abs(denom(A[i][1][1]))=abs(denom(A[i][1][3]))
                   B[i]:= [[A[i][1][2],A[i][1][3],A[i][1][1]],[]]
                fi;
             else
                B[i]:= A[i]
             fi;
          od;
          return convert(B,'list');
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing1F2(L,t,G,ext).
Goodexpbi2:=proc(R,x,t,ext)
          local Dx,l,n,A,B,s,i,j,e,C;
          Dx,x:=op(_Envdiffopdomain);
          n:=nops(R[3][1][1]);
          if n=1 then
             return {R[3][1]}
          else
             A:= Goodexpbi1(R[3][1][3],x,t,ext);
             B:= map(s-> `if`(nops(s[1])=3 and s[1][1]<>s[1][2],{s,[[s[1][2],s[1][1],s[1][3]],[]]},{s}),A);
             C:={[seq(e[i],i=1..n)]};
             for i from 1 to n do
                C:=map(s-> seq(subs(e[i]=B[i][j],s),j=1..nops(B[i])),C)
             od:
             return {seq(subs({seq(R[3][1][3][i]=l[i],i=1..n)},R[3][1]),l in C)}
         fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# E = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,E,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,ext).
find0F2Ratf:= proc(L,R,info,x,t,T,ext)
           local  Dx,Newp,s, l, dnu,dv, d1,d2, res, i, knl, j, k,f,ff,B,B1,n,SR,dA,Sreg,h,g,H;
           Dx,x:=op(_Envdiffopdomain);
           res:={};
           Newp:=Goodexpbi2(R,x,t,ext);
           for Sreg in Newp do
              dA:= info[2];
              if Sreg=[] then
                 dnu := {};
                 for i from 2 to floor(dA/2) do
                    if irem(dA,i)=0 then
                       dnu:= dnu union {i}
                    fi;
                 od;
                 dnu := dnu union {dA};
              else
                 dnu := {};
                 dv := denom(Sreg[3][1][1][1]);
                 for i from 1 to nops(Sreg[1]) do
                    dv:=lcm(dv,denom(Sreg[3][i][1][1]));
                 od;
                 if dv >= dA then
                    for i from 2 to dA*dv do
                       if irem(i,dv)=0 then
                          dnu := dnu union {i};
                       fi;
                    od;
                 else
                    for i from 1 to iquo(dA,dv) do
                       d1:=0;
                       d2:=0;
                       for s from 1 to nops(Sreg[1]) do
                          d1:=d1+i*dv/denom(Sreg[3][s][1][1])*(`if`(Sreg[1][s][1]=infinity,1,
                                                               degree(Sreg[1][s][1],x)));
                          d2 := gcd(d2, i*dv/denom(Sreg[3][s][1][1]));
                       od;
                       if d1<=dA and irem(dA,d2)=0 then
                          dnu := dnu union {i*dv};
                       fi;
                    od;
                 fi;
              fi;
              B:= info[4];
              for i in dnu do
                 knl := search0F2A1k(Sreg,dA,i,x);
                 if knl={} then
                    next
                 fi;
                 ff := help0F2f(L,info,i,knl,x,t,T,ext);
                 ff := simplef(ff);
                 if assigned(_Env_original_field) then
                    f := {};
                    for k in ff do
                       if indets(k, RootOf) minus _Env_original_field ={} then
                          f:=f union {k};
                       fi;
                    od;
                 else
                    f:= map(j->factor(j),ff);
                 fi;
                 res := res union f
              od;
           od;
           if res = {} then
              return NULL
           else
              H:= table();
              for g in res do
                 B1:= Zeroesmultordpol0F2(g,x);
                 if B1=NULL then
                    next
                 fi;
                 n:= nops(B1);
                 SR:= newregsing0F2(L,t,x,B1,R[2],n,ext);
                 H[g] := [g,SR[2],SR[1]];
              od;
              H:= convert(H,'set');
              return H
           fi;
end:

############################################################################################################
############################################################################################################

# We compute the candiates for the lower parameters (b1,b2), also and f: [[[b1,b2],[b1,b2],...],f].
#
# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# R = IrrRegAppsing0F2(L,t,G,ext);
# info = Sirr0F2info1(L,R[1],R[2],x,t,ext).
find0F2Rat:=proc(L,R,info,x,t,T,ext)
                  local Dx,B,n,ab,f,C,H,E,i,n1,s,G1,ext1,ext2;
                  Dx,x:=op(_Envdiffopdomain);
                  B:= find0F2Ratf(L,R,info,x,t,T,ext);
                  if B = NULL then
                     return NULL
                  fi;
                  if R[2]=[] then
                     ext1:= {}
                  else
                     ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
                     ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
                     n1:=nops(R[2][1]);
                     if n1 > 1 then
                        for i from 2 to n1 do
                           ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                           ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                           ext1:= ext1 intersect ext2;
                        od;
                     fi;
                  fi;
                  H:= table();
                  for f in B do
                     n:= nops(f[3][1]);
                     C:= findbi0F2(L,f,x,t,ext,ext1,n);
                     if C<>NULL then
                        H[f]:= [C[1],f[1]]
                     fi;
                  od;
                  H:= convert(H,'list');
                  if H = [] then
                     return NULL
                  else
                     return H
                  fi;
end:

############################################################################################################
############################################################################################################

Pcurvature1:= proc(Lp, n, p, c, T)
               local Dx, x, A, F, R, j, i, l, G, NewLp;
               Dx, x := op(_Envdiffopdomain);
               l := lcoeff(Lp, Dx);
               if `mod`(Gcdex(l, x^p-c, x, 's', 'e'), p) <> 1 then ### error "selected wrong value for c"
                  return NULL
               fi;
               NewLp := `mod`(Rem(s*Lp, x^p-c, x), p);
               R[1] := Dx;
               for i from 2 to p+n-1 do
                  R[i] := rightdivision(mult(Dx, R[i-1]), NewLp)[2];
                  R[i] := `mod`(evala(Rem(R[i], x^p-c, x)), p)
               od;
               A := Matrix(n, n, [seq([seq(coeff(R[p+j], Dx, i), i = 0 .. n-1)], j = 0 .. n-1)]);
               F := LinearAlgebra:-CharacteristicPolynomial(A, T);
               F := `mod`(evala(Rem(F, x^p-c, x)), p);
               if has(F, x) then ####### error "bug in Pcurvature1"
                  return NULL
               else
                  F
               fi;
end:

############################################################################################################
############################################################################################################

IsGaugeEq:= proc(F1, F2, p)
             evalb(`mod`(Normal(F1-F2), p) = 0)
end:

############################################################################################################
############################################################################################################

IsShiftOf:= proc(F1, F2, T, p)
             local s, n;
             n := degree(F1, T);
             if degree(F2, T) <> n then
                return false
             fi;
             if lcoeff(F1, T) <> 1 or lcoeff(F2, T) <> 1 or `mod`(n, p) = 0 then
                # error "did not implement this"
                return false
             fi;
             s := coeff(F1-F2, T, n-1)/n;
             evalb(`mod`(Normal(subs(T = T+s, F2)-F1), p) = 0)
end:

############################################################################################################
############################################################################################################

testzero1:= proc(l1,l2,p)
             local Dx, x, i;
             Dx, x := op(_Envdiffopdomain);
             for i from 0 to p-1 do
                if `mod`(evala(Rem(l1, x-i, x)), p) <> 0 and `mod`(evala(Rem(l2, x-i, x)), p) <> 0 then
                   return i
                fi;
             od;
             NULL
end:

############################################################################################################
############################################################################################################

helpEquivtest1:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p,c;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 or testzero1(l1,l2,p)=NULL do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               c:= testzero1(l1,l2,p);
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  c:= testzero1(l1,l2,p);
                  if l1 mod p <> 0 and l2 mod p <> 0 and c<>NULL then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p,c]
end:

############################################################################################################
############################################################################################################

helpEquivtest11:= proc(L1,L2,n1,l1,l2,p0)
               local Dx, x,p,L1p,L2p;
               Dx, x := op(_Envdiffopdomain);
               p := p0;
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while degree(L1p,Dx) < n1 or degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p := evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p := evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               return [L1p,L2p,p]

end:

############################################################################################################
############################################################################################################

Equivtest11 := proc (L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               while l1 mod p = 0 or l2 mod p = 0 do
                  p := nextprime(p);
               od;
               L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
               L1p := collect(L1p, Dx, Normal) mod p;
               L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
               L2p := collect(L2p, Dx, Normal) mod p;
               while  l1 mod p = 0 or l2 mod p = 0 or degree(L1p,Dx) < n1 or
                     degree(L2p,Dx) < n1 do
                  p := nextprime(p);
                  if l1 mod p <> 0 and l2 mod p <> 0 then
                     L1p:=evala(Primpart(numer(Algebraic:-Normal(L1) mod p),Dx)) mod p;
                     L1p := collect(L1p, Dx, Normal) mod p;
                     L2p:=evala(Primpart(numer(Algebraic:-Normal(L2) mod p),Dx)) mod p;
                     L2p := collect(L2p, Dx, Normal) mod p;
                  fi;
               od;
               c:= testzero1(l1,l2,p);
               if c=NULL then
                  return procname(L1,L2,p)
               fi;
               F1 := Pcurvature1(L1p,n1,p,c,T);
               F2 := Pcurvature1(L2p,n1,p,c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, p) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, p) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest1 := proc (L11,L22,p0)
               local Dx, x,L1,L2,n1,n2,l1,l2,p,L1p,L2p,F1,F2,T,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L11, Dx);
               n2 := degree(L22, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               L1 :=collect(evala(Primpart(L11,Dx)),Dx,factor);
               L2 :=collect(evala(Primpart(L22,Dx)),Dx,factor);
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               F1 := Pcurvature1(E[1],n1,E[3],E[4],T);
               F2 := Pcurvature1(E[2],n1,E[3],E[4],T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else ####### error "we should not reach this point"
                  return "false"
               fi;
end:

############################################################################################################
############################################################################################################

Equivtest12:= proc(L1,L2,p0)
               local Dx, x,n1,n2,l1,l2,p,F1,F2,E,c;
               #` options trace;`
               if nargs = 2 then
                  return procname(args, 1)
               fi;
               Dx, x := op(_Envdiffopdomain);
               n1 := degree(L1, Dx);
               n2 := degree(L2, Dx);
               if n1 <> n2 then
                  return "false"
               fi;
               l1 := lcoeff(L1, Dx);
               l2 := lcoeff(L2, Dx);
               p := nextprime(p0);
               E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
               c:= testzero1(l1,l2,E[3]);
               while c=NULL do
                  p:= nextprime(E[3]);
                  E:= helpEquivtest1(L1,L2,n1,l1,l2,p);
                  c:= testzero1(l1,l2,E[3]);
               od;
               F1 := Pcurvature1(E[1],n1,E[3],c,T);
               F2 := Pcurvature1(E[2],n1,E[3],c,T);
               if F1<>NULL and F2<>NULL then
                  if IsGaugeEq(F1, F2, E[3]) = true then
                     return "probably Gauge"
                  elif IsShiftOf(F1, F2, T, E[3]) = true then
                     return "probably Projectiv but certainly not Gauge"
                  else
                     return "certainly not projectively equivalent"
                  fi;
               else
                  error "we should not reach this point"
               fi;
end:

############################################################################################################
############################################################################################################

# E = Singular(L,ext);
# A = NotAppSing(L,E,ext);
# r = irrsing1F2(L,t,A,ext);
# Sreg = regsingtrue1F2(L,t,r[-1],ext);
# RSreg = Sregseptrue1F2(L,Sreg,ext);
# info = Sirr1F2info1(L,r,Sreg,x,ext).
help0F2Solutions1:= proc(L,R,info,x,t,T,ext)
           local Dx,n1,ext1,ext2,i,H,H1;
           Dx,x:=op(_Envdiffopdomain);
           if R[3] =[[],[],[]] then    ####### Here Sreg=[].
              H1:= find0F2Rat(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easy0F2(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           elif R[3][3]<>[] then
              H1:= find0F2ln(L,R,info,x,t,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easy0F2(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           elif R[3][2]<>[] then
              H1:= find0F2Rat(L,R,info,x,t,T,ext);
              if H1=NULL then
                 if info[3]>info[2] then
                    H:= easy0F2(L,R,info,x,t,ext);
                 else
                    return NULL
                 fi;
              else
                 H:= H1
              fi;
           else
              ext1:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][1],{RootOf,name}));
              ext1:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext1,{RootOf,name}));
              n1:=nops(R[2][1]);
              if n1 > 1 then
                 for i from 2 to n1 do
                    ext2:=map(s-> `if`(type(s,RootOf),ConvirrRoot(s)[1],s),indets(R[2][3][i],{RootOf,name}));
                    ext2:= map(s-> op({op(ReplRoot(s,{})),ReplRoot1(s)}),indets(ext2,{RootOf,name}));
                    ext1:= ext1 intersect ext2;
                 od;
              fi;
              if ext1 <> {} then
                 H1:= find0F2Irr(L,R,info,x,t,ext);
                 if H1=NULL then
                    if info[3]>info[2] then
                       H:= easy0F2(L,R,info,x,t,ext);
                    else
                       return NULL
                    fi;
                 else
                    H:= H1
                 fi;
              else
                 H1:= find0F2Rat(L,R,info,x,t,T,ext);
                 if H1=NULL then
                    if info[3]>info[2] then
                       H:= easy0F2(L,R,info,x,t,ext);
                    else
                       return NULL
                    fi;
                 else
                    H:= H1
                 fi;
              fi;
           fi:
           if H=NULL then
              return NULL
           else
              H:= {op(map(s-> op([s,subs(s[-1]=-s[-1],s)]),H))};
              return [op(H)]
           fi;
end:

############################################################################################################
############################################################################################################

# ext = indets(L,{RootOf,name}) minus {x,Dx};
# ext: = indets(map(s-> ReplirrRoot(s),ext),{RootOf,name});
# extppp = {};
# G = Singular(L,extppp);
# A = IrrRegAppsing0F2(L,t,G,ext);
# a = element of help0F2Solutions1(L,r,RSreg,info,ext,x).
help0F2Solutions2:= proc(L,a,A,x,t,ext)
           local Dx,J,n1,C1,M,L02,L22,k,R,i,e1,e2,T;
           Dx,x:=op(_Envdiffopdomain);
              n1:= nops(a[1]);
              C1:= sumdiffeq(hyperterm([],[e1,e2],x,k),k,J(x));
              L02:= de2diffop(C1,J(x));
              for i from 1 to n1 do
                 L22:=subs({e1=a[1][i][1],e2=a[1][i][2]},L02);
                 M:= ChangeOfVariables(L22,a[2]);
                 R:=EquivExpgauge0F2(M,L,A,x,t,T,ext);
                 if R<>NULL then
                    return [{[a[1][i],R]},a[2]]
                 fi;
              od;
              return NULL
end:

############################################################################################################
############################################################################################################

help0F2Solutions3:= proc(G)
            local Dx,x,m,n,n1,i,s;
            Dx,x:=op(_Envdiffopdomain);
            m:= nops(G[2]);
            n:= length(G[2][1]);
            s:= 1;
            if m>1 then
               for i from 2 to m do
                  n1:= length(G[2][i]);
                  if n1<n then
                     n:= n1;
                     s:= i;
                  fi;
               od;
            fi;
            return [G[1],[G[2][s]]];
end:

############################################################################################################
############################################################################################################

help0F2Solutions4:= proc(a,b)
            local Dx,x,n1,n2;
            Dx,x:=op(_Envdiffopdomain);
            if a[2]<>b[2] then
               return [false,b]
            elif Matchlistmod2(a[1],b[1])=false then
               return [false,b]
            else
               n1:= length(a[3]);
               n2:= length(b[3]);
               if n2<n1 then
                  return [true,b]
               else
                  return [true,a]
               fi;
            fi;
end:

############################################################################################################
############################################################################################################

help0F2Solutions5:= proc(G,b)
            local Dx,x,n1,n2,n,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n1:= nops(G);
            n2:= 0;
            for i from 1 to n1 do
               E:= help0F2Solutions4(G[i],b);
               if E[1]=false then
                  n2:= n2+1
               else
                  n:= i;
                  break
               fi;
            od;
            if n2=n1 then
              return [false,b]
            else
               return [true,[E[2],n]]
            fi;
end:

############################################################################################################
############################################################################################################

help0F2Solutions6:= proc(H)
            local Dx,x,n,G,E,i;
            Dx,x:=op(_Envdiffopdomain);
            n:= nops(H);
            G:= table();
            G[1]:= H[1];
            if n>1 then
               for i from 2 to n do
                  E:= help0F2Solutions5(G,H[i]);
                  if E[1]=true then
                     G[E[2][2]]:= E[2][1]
                  else
                     G[i]:= E[2]
                  fi;
               od;
            fi;
            G:= convert(G,'set');
            return G
end:

############################################################################################################
############################################################################################################

Find0F2Solutions:= proc(L)
           local Dx,x,a,R,A,ext,extppp,F,E,C2,H,C,t,T;
           Dx,x:=op(_Envdiffopdomain);
           ext:=indets(L,{RootOf,name}) minus {x,Dx};
           ext:= indets(map(s-> ReplirrRoot(s,{}),ext),{RootOf,name});
           extppp:={};
           E:= Singular(L,extppp);
           R:=IrrRegAppsing0F2(L,t,E,ext);
           if R=NULL then
              return NULL
           fi;
           F:= Sirr0F2info1(L,R[1],R[2],x,t,ext);
           if F=NULL then
              return NULL
           fi;
           C:= table();
           H:= help0F2Solutions1(L,R,F,x,t,T,ext);
           if H=NULL then
              return NULL
           fi;
           for a in H do
              C2:= help0F2Solutions2(L,a,R,x,t,ext);
              if C2<>NULL then
                 C[a]:= C2
              fi;
           od;
           C:= convert(C,'set');
           if C<>{} then
              return C
           else
              return NULL
           fi;
end:

############################################################################################################
############################################################################################################

Hyp0F2Solutions:= proc(L)
           local Dx,x,s,k,F,E,H1,H2,G,i,H,C;
           Dx,x:=op(_Envdiffopdomain);
           H:= table();
           H1:= Find0F2Solutions(L);
           if H1=NULL then
              return "No 0F2 type solutions"
           fi;
           F:= table();
           G:= map(s->s[2],H1);
           for i in G do
              H2:= [[op(map(s->`if`(s[2]=i,op(s[1]),NULL),H1))],i] ;
              F[i]:= H2
           od;
           F:= convert(F,'set');
           for C in F do
              E:= map(s->subs(s[2]=map(k->help0F2Solutions3(k),s[2]),s),C[1]);
              E:= map(s->seq([s[1],op(s[2][i])],i=1..nops(s[2])),E);
              E:= help0F2Solutions6(E);
              H[C]:= [E,C[2]]
           od;
           H:= convert(H,'set');
           return H
end:
